/*
 * Copyright (c) 2014 Patrick P. Frey
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
#include "normalizerCharConv.hpp"
#include "strus/errorBufferInterface.hpp"
#include "strus/base/string_conv.hpp"
#include "textwolf/charset_utf8.hpp"
#include "textwolf/cstringiterator.hpp"
#include "private/errorUtils.hpp"
#include "private/internationalization.hpp"
#include <cstring>

using namespace strus;

class CharMap
{
public:
	enum ConvType {Diacritical, Lowercase, Uppercase};

	CharMap(){}
	explicit CharMap( ConvType type)
	{
		load( type);
	}

	void load( ConvType type);

	typedef const char* (*ExceptionsF)( unsigned int chr);
	std::string rewrite( const char* src, std::size_t srcsize, ExceptionsF exceptions, ErrorBufferInterface* errorhnd) const;

private:
	void set( unsigned int chr, const char* value);
	void set( unsigned int chr, unsigned int mapchr);

	void buildMapDiacritical();
	void buildMapTolower();
	void buildMapToupper();

private:
	std::map<unsigned int,std::size_t> m_map;
	std::string m_strings;
};

class CharMapNormalizerInstance
	:public NormalizerFunctionInstanceInterface
{
public:
	CharMapNormalizerInstance( CharMap::ConvType maptype_, CharMap::ExceptionsF exceptions_, ErrorBufferInterface* errorhnd_)
		:m_map(maptype_),m_exceptions(exceptions_),m_errorhnd(errorhnd_){}

	virtual std::string normalize(
			const char* src,
			std::size_t srcsize) const
	{
		return m_map.rewrite( src, srcsize, m_exceptions, m_errorhnd);
	}

private:
	CharMap m_map;
	CharMap::ExceptionsF m_exceptions;
	ErrorBufferInterface* m_errorhnd;
};


void CharMap::set( unsigned int chr, const char* value)
{
	m_map[ chr] = m_strings.size();
	m_strings.append( value);
	m_strings.push_back( '\0');
}

void CharMap::set( unsigned int chr, unsigned int mapchr)
{
	textwolf::charset::UTF8 utf8;
	std::string buf;
	utf8.print( mapchr, buf);
	m_map[ chr] = m_strings.size();
	m_strings.append( buf);
	m_strings.push_back( '\0');
}

void CharMap::buildMapDiacritical()
{
	set( 0xa0, " "); /* ' ' -> ' ' */
	set( 0xa8, " "); /* '¨' -> ' ' */
	set( 0xaa, "a"); /* 'ª' -> 'a' */
	set( 0xaf, " "); /* '¯' -> ' ' */
	set( 0xb2, "2"); /* '²' -> '2' */
	set( 0xb3, "3"); /* '³' -> '3' */
	set( 0xb4, " "); /* '´' -> ' ' */
	set( 0xb8, " "); /* '¸' -> ' ' */
	set( 0xb9, "1"); /* '¹' -> '1' */
	set( 0xba, "o"); /* 'º' -> 'o' */
	set( 0xbc, "1/4"); /* '¼' -> '1/4' */
	set( 0xbd, "1/2"); /* '½' -> '1/2' */
	set( 0xbe, "3/4"); /* '¾' -> '3/4' */
	set( 0xc0, "A"); /* 'À' -> 'A' */
	set( 0xc1, "A"); /* 'Á' -> 'A' */
	set( 0xc2, "A"); /* 'Â' -> 'A' */
	set( 0xc3, "A"); /* 'Ã' -> 'A' */
	set( 0xc4, "A"); /* 'Ä' -> 'A' */
	set( 0xc5, "A"); /* 'Å' -> 'A' */
	set( 0xc7, "C"); /* 'Ç' -> 'C' */
	set( 0xc8, "E"); /* 'È' -> 'E' */
	set( 0xc9, "E"); /* 'É' -> 'E' */
	set( 0xca, "E"); /* 'Ê' -> 'E' */
	set( 0xcb, "E"); /* 'Ë' -> 'E' */
	set( 0xcc, "I"); /* 'Ì' -> 'I' */
	set( 0xcd, "I"); /* 'Í' -> 'I' */
	set( 0xce, "I"); /* 'Î' -> 'I' */
	set( 0xcf, "I"); /* 'Ï' -> 'I' */
	set( 0xd1, "N"); /* 'Ñ' -> 'N' */
	set( 0xd2, "O"); /* 'Ò' -> 'O' */
	set( 0xd3, "O"); /* 'Ó' -> 'O' */
	set( 0xd4, "O"); /* 'Ô' -> 'O' */
	set( 0xd5, "O"); /* 'Õ' -> 'O' */
	set( 0xd6, "O"); /* 'Ö' -> 'O' */
	set( 0xd9, "U"); /* 'Ù' -> 'U' */
	set( 0xda, "U"); /* 'Ú' -> 'U' */
	set( 0xdb, "U"); /* 'Û' -> 'U' */
	set( 0xdc, "U"); /* 'Ü' -> 'U' */
	set( 0xdd, "Y"); /* 'Ý' -> 'Y' */
	set( 0xe0, "a"); /* 'à' -> 'a' */
	set( 0xe1, "a"); /* 'á' -> 'a' */
	set( 0xe2, "a"); /* 'â' -> 'a' */
	set( 0xe3, "a"); /* 'ã' -> 'a' */
	set( 0xe4, "a"); /* 'ä' -> 'a' */
	set( 0xe5, "a"); /* 'å' -> 'a' */
	set( 0xe7, "c"); /* 'ç' -> 'c' */
	set( 0xe8, "e"); /* 'è' -> 'e' */
	set( 0xe9, "e"); /* 'é' -> 'e' */
	set( 0xea, "e"); /* 'ê' -> 'e' */
	set( 0xeb, "e"); /* 'ë' -> 'e' */
	set( 0xec, "i"); /* 'ì' -> 'i' */
	set( 0xed, "i"); /* 'í' -> 'i' */
	set( 0xee, "i"); /* 'î' -> 'i' */
	set( 0xef, "i"); /* 'ï' -> 'i' */
	set( 0xf1, "n"); /* 'ñ' -> 'n' */
	set( 0xf2, "o"); /* 'ò' -> 'o' */
	set( 0xf3, "o"); /* 'ó' -> 'o' */
	set( 0xf4, "o"); /* 'ô' -> 'o' */
	set( 0xf5, "o"); /* 'õ' -> 'o' */
	set( 0xf6, "o"); /* 'ö' -> 'o' */
	set( 0xf9, "u"); /* 'ù' -> 'u' */
	set( 0xfa, "u"); /* 'ú' -> 'u' */
	set( 0xfb, "u"); /* 'û' -> 'u' */
	set( 0xfc, "u"); /* 'ü' -> 'u' */
	set( 0xfd, "y"); /* 'ý' -> 'y' */
	set( 0xff, "y"); /* 'ÿ' -> 'y' */
	set( 0x100, "A"); /* 'Ā' -> 'A' */
	set( 0x101, "a"); /* 'ā' -> 'a' */
	set( 0x102, "A"); /* 'Ă' -> 'A' */
	set( 0x103, "a"); /* 'ă' -> 'a' */
	set( 0x104, "A"); /* 'Ą' -> 'A' */
	set( 0x105, "a"); /* 'ą' -> 'a' */
	set( 0x106, "C"); /* 'Ć' -> 'C' */
	set( 0x107, "c"); /* 'ć' -> 'c' */
	set( 0x108, "C"); /* 'Ĉ' -> 'C' */
	set( 0x109, "c"); /* 'ĉ' -> 'c' */
	set( 0x10a, "C"); /* 'Ċ' -> 'C' */
	set( 0x10b, "c"); /* 'ċ' -> 'c' */
	set( 0x10c, "C"); /* 'Č' -> 'C' */
	set( 0x10d, "c"); /* 'č' -> 'c' */
	set( 0x10e, "D"); /* 'Ď' -> 'D' */
	set( 0x10f, "d"); /* 'ď' -> 'd' */
	set( 0x110, "D"); /* 'Đ' -> 'D' */
	set( 0x111, "d"); /* 'đ' -> 'd' */
	set( 0x112, "E"); /* 'Ē' -> 'E' */
	set( 0x113, "e"); /* 'ē' -> 'e' */
	set( 0x114, "E"); /* 'Ĕ' -> 'E' */
	set( 0x115, "e"); /* 'ĕ' -> 'e' */
	set( 0x116, "E"); /* 'Ė' -> 'E' */
	set( 0x117, "e"); /* 'ė' -> 'e' */
	set( 0x118, "E"); /* 'Ę' -> 'E' */
	set( 0x119, "e"); /* 'ę' -> 'e' */
	set( 0x11a, "E"); /* 'Ě' -> 'E' */
	set( 0x11b, "e"); /* 'ě' -> 'e' */
	set( 0x11c, "G"); /* 'Ĝ' -> 'G' */
	set( 0x11d, "g"); /* 'ĝ' -> 'g' */
	set( 0x11e, "G"); /* 'Ğ' -> 'G' */
	set( 0x11f, "g"); /* 'ğ' -> 'g' */
	set( 0x120, "G"); /* 'Ġ' -> 'G' */
	set( 0x121, "g"); /* 'ġ' -> 'g' */
	set( 0x122, "G"); /* 'Ģ' -> 'G' */
	set( 0x123, "g"); /* 'ģ' -> 'g' */
	set( 0x124, "H"); /* 'Ĥ' -> 'H' */
	set( 0x125, "h"); /* 'ĥ' -> 'h' */
	set( 0x128, "I"); /* 'Ĩ' -> 'I' */
	set( 0x129, "i"); /* 'ĩ' -> 'i' */
	set( 0x12a, "I"); /* 'Ī' -> 'I' */
	set( 0x12b, "i"); /* 'ī' -> 'i' */
	set( 0x12c, "I"); /* 'Ĭ' -> 'I' */
	set( 0x12d, "i"); /* 'ĭ' -> 'i' */
	set( 0x12e, "I"); /* 'Į' -> 'I' */
	set( 0x12f, "i"); /* 'į' -> 'i' */
	set( 0x130, "I"); /* 'İ' -> 'I' */
	set( 0x132, "IJ"); /* 'Ĳ' -> 'IJ' */
	set( 0x133, "ij"); /* 'ĳ' -> 'ij' */
	set( 0x134, "J"); /* 'Ĵ' -> 'J' */
	set( 0x135, "j"); /* 'ĵ' -> 'j' */
	set( 0x136, "K"); /* 'Ķ' -> 'K' */
	set( 0x137, "k"); /* 'ķ' -> 'k' */
	set( 0x139, "L"); /* 'Ĺ' -> 'L' */
	set( 0x13a, "l"); /* 'ĺ' -> 'l' */
	set( 0x13b, "L"); /* 'Ļ' -> 'L' */
	set( 0x13c, "l"); /* 'ļ' -> 'l' */
	set( 0x13d, "L"); /* 'Ľ' -> 'L' */
	set( 0x13e, "l"); /* 'ľ' -> 'l' */
	set( 0x13f, "L"); /* 'Ŀ' -> 'L' */
	set( 0x140, "l"); /* 'ŀ' -> 'l' */
	set( 0x143, "N"); /* 'Ń' -> 'N' */
	set( 0x144, "n"); /* 'ń' -> 'n' */
	set( 0x145, "N"); /* 'Ņ' -> 'N' */
	set( 0x146, "n"); /* 'ņ' -> 'n' */
	set( 0x147, "N"); /* 'Ň' -> 'N' */
	set( 0x148, "n"); /* 'ň' -> 'n' */
	set( 0x149, "n"); /* 'ŉ' -> 'n' */
	set( 0x14a, "N"); /* 'Ŋ' -> 'N' */
	set( 0x14b, "n"); /* 'ŋ' -> 'n' */
	set( 0x14c, "O"); /* 'Ō' -> 'O' */
	set( 0x14d, "o"); /* 'ō' -> 'o' */
	set( 0x14e, "O"); /* 'Ŏ' -> 'O' */
	set( 0x14f, "o"); /* 'ŏ' -> 'o' */
	set( 0x150, "O"); /* 'Ő' -> 'O' */
	set( 0x151, "o"); /* 'ő' -> 'o' */
	set( 0x154, "R"); /* 'Ŕ' -> 'R' */
	set( 0x155, "r"); /* 'ŕ' -> 'r' */
	set( 0x156, "R"); /* 'Ŗ' -> 'R' */
	set( 0x157, "r"); /* 'ŗ' -> 'r' */
	set( 0x158, "R"); /* 'Ř' -> 'R' */
	set( 0x159, "r"); /* 'ř' -> 'r' */
	set( 0x15a, "S"); /* 'Ś' -> 'S' */
	set( 0x15b, "s"); /* 'ś' -> 's' */
	set( 0x15c, "S"); /* 'Ŝ' -> 'S' */
	set( 0x15d, "s"); /* 'ŝ' -> 's' */
	set( 0x15e, "S"); /* 'Ş' -> 'S' */
	set( 0x15f, "s"); /* 'ş' -> 's' */
	set( 0x160, "S"); /* 'Š' -> 'S' */
	set( 0x161, "s"); /* 'š' -> 's' */
	set( 0x162, "T"); /* 'Ţ' -> 'T' */
	set( 0x163, "t"); /* 'ţ' -> 't' */
	set( 0x164, "T"); /* 'Ť' -> 'T' */
	set( 0x165, "t"); /* 'ť' -> 't' */
	set( 0x168, "U"); /* 'Ũ' -> 'U' */
	set( 0x169, "u"); /* 'ũ' -> 'u' */
	set( 0x16a, "U"); /* 'Ū' -> 'U' */
	set( 0x16b, "u"); /* 'ū' -> 'u' */
	set( 0x16c, "U"); /* 'Ŭ' -> 'U' */
	set( 0x16d, "u"); /* 'ŭ' -> 'u' */
	set( 0x16e, "U"); /* 'Ů' -> 'U' */
	set( 0x16f, "u"); /* 'ů' -> 'u' */
	set( 0x170, "U"); /* 'Ű' -> 'U' */
	set( 0x171, "u"); /* 'ű' -> 'u' */
	set( 0x172, "U"); /* 'Ų' -> 'U' */
	set( 0x173, "u"); /* 'ų' -> 'u' */
	set( 0x174, "W"); /* 'Ŵ' -> 'W' */
	set( 0x175, "w"); /* 'ŵ' -> 'w' */
	set( 0x176, "Y"); /* 'Ŷ' -> 'Y' */
	set( 0x177, "y"); /* 'ŷ' -> 'y' */
	set( 0x178, "Y"); /* 'Ÿ' -> 'Y' */
	set( 0x179, "Z"); /* 'Ź' -> 'Z' */
	set( 0x17a, "z"); /* 'ź' -> 'z' */
	set( 0x17b, "Z"); /* 'Ż' -> 'Z' */
	set( 0x17c, "z"); /* 'ż' -> 'z' */
	set( 0x17d, "Z"); /* 'Ž' -> 'Z' */
	set( 0x17e, "z"); /* 'ž' -> 'z' */
	set( 0x17f, "s"); /* 'ſ' -> 's' */
	set( 0x180, "b"); /* 'ƀ' -> 'b' */
	set( 0x181, "B"); /* 'Ɓ' -> 'B' */
	set( 0x182, "b"); /* 'Ƃ' -> 'b' */
	set( 0x183, "b"); /* 'ƃ' -> 'b' */
	set( 0x184, "Ƅ"); /* 'Ƅ' -> 'b' */
	set( 0x185, "b"); /* 'ƅ' -> 'b' */
	set( 0x186, "C"); /* 'Ɔ' -> 'C' */
	set( 0x187, "C"); /* 'Ƈ' -> 'C' */
	set( 0x188, "c"); /* 'ƈ' -> 'c' */
	set( 0x189, "D"); /* 'Ɖ' -> 'D' */
	set( 0x18a, "D"); /* 'Ɗ' -> 'D' */
	set( 0x18b, "d"); /* 'Ƌ' -> 'd' */
	set( 0x18c, "d"); /* 'ƌ' -> 'Ƌ' */
	set( 0x18e, "E"); /* 'Ǝ' -> 'E' */
	set( 0x18f, "e"); /* 'Ə' -> 'e' */
	set( 0x190, "E"); /* 'Ɛ' -> 'E' */
	set( 0x191, "F"); /* 'Ƒ' -> 'F' */
	set( 0x192, "f"); /* 'ƒ' -> 'f' */
	set( 0x193, "G"); /* 'Ɠ' -> 'G' */
	set( 0x194, "g"); /* 'Ɣ' -> 'g' */
	set( 0x195, "h"); /* 'ƕ' -> 'h' */
	set( 0x196, "i"); /* 'Ɩ' -> 'i' */
	set( 0x197, "I"); /* 'Ɨ' -> 'I' */
	set( 0x198, "K"); /* 'Ƙ' -> 'K' */
	set( 0x199, "k"); /* 'ƙ' -> 'k' */
	set( 0x19a, "l"); /* 'ƚ' -> 'l' */
	set( 0x19c, "W"); /* 'Ɯ' -> 'W' */
	set( 0x19d, "N"); /* 'Ɲ' -> 'N' */
	set( 0x19e, "n"); /* 'ƞ' -> 'n' */
	set( 0x19f, "0"); /* 'Ɵ' -> 'O' */
	set( 0x1a0, "O"); /* 'Ơ' -> 'O' */
	set( 0x1a1, "o"); /* 'ơ' -> 'o' */
	set( 0x1a2, "A"); /* 'Ƣ' -> 'A' */
	set( 0x1a3, "a"); /* 'ƣ' -> 'a' */
	set( 0x1a4, "P"); /* 'Ƥ' -> 'P' */
	set( 0x1a5, "p"); /* 'ƥ' -> 'p' */
	set( 0x1a6, "R"); /* 'Ʀ' -> 'R' */
	set( 0x1a7, "S"); /* 'Ƨ' -> 'S' */
	set( 0x1a8, "s"); /* 'ƨ' -> 's' */
	set( 0x1a9, "S"); /* 'Ʃ' -> 'S' */
	set( 0x1ac, "T"); /* 'Ƭ' -> 'T' */
	set( 0x1ad, "t"); /* 'ƭ' -> 't' */
	set( 0x1ae, "T"); /* 'Ʈ' -> 'T' */
	set( 0x1af, "U"); /* 'Ư' -> 'U' */
	set( 0x1b0, "u"); /* 'ư' -> 'u' */
	set( 0x1b1, "V"); /* 'Ʊ' -> 'V' */
	set( 0x1b2, "V"); /* 'Ʋ' -> 'V' */
	set( 0x1b3, "Y"); /* 'Ƴ' -> 'Y' */
	set( 0x1b4, "y"); /* 'ƴ' -> 'y' */
	set( 0x1b5, "Z"); /* 'Ƶ' -> 'Z' */
	set( 0x1b6, "z"); /* 'ƶ' -> 'z' */
	set( 0x1b7, "3"); /* 'Ʒ' -> '3' */
	set( 0x1b8, "3"); /* 'Ƹ' -> '3' */
	set( 0x1b9, "3"); /* 'ƹ' -> '3' */
	set( 0x1bc, "5"); /* 'Ƽ' -> '5' */
	set( 0x1bd, "5"); /* 'ƽ' -> '5' */
	set( 0x1bf, "p"); /* 'ƿ' -> 'p' */
	set( 0x1c4, "DZ"); /* 'Ǆ' -> 'DZ' */
	set( 0x1c5, "Dz"); /* 'ǅ' -> 'Dz' */
	set( 0x1c6, "dz"); /* 'ǆ' -> 'dz' */
	set( 0x1c7, "LJ"); /* 'Ǉ' -> 'LJ' */
	set( 0x1c8, "Lj"); /* 'ǈ' -> 'Lj' */
	set( 0x1c9, "lj"); /* 'ǉ' -> 'lj' */
	set( 0x1ca, "NJ"); /* 'Ǌ' -> 'NJ' */
	set( 0x1cb, "Nj"); /* 'ǋ' -> 'Nj' */
	set( 0x1cc, "nj"); /* 'ǌ' -> 'nj' */
	set( 0x1cd, "A"); /* 'Ǎ' -> 'A' */
	set( 0x1ce, "a"); /* 'ǎ' -> 'a' */
	set( 0x1cf, "I"); /* 'Ǐ' -> 'I' */
	set( 0x1d0, "i"); /* 'ǐ' -> 'i' */
	set( 0x1d1, "O"); /* 'Ǒ' -> 'O' */
	set( 0x1d2, "o"); /* 'ǒ' -> 'o' */
	set( 0x1d3, "U"); /* 'Ǔ' -> 'U' */
	set( 0x1d4, "u"); /* 'ǔ' -> 'u' */
	set( 0x1d5, "U"); /* 'Ǖ' -> 'U' */
	set( 0x1d6, "u"); /* 'ǖ' -> 'u' */
	set( 0x1d7, "U"); /* 'Ǘ' -> 'U' */
	set( 0x1d8, "u"); /* 'ǘ' -> 'u' */
	set( 0x1d9, "U"); /* 'Ǚ' -> 'U' */
	set( 0x1da, "u"); /* 'ǚ' -> 'u' */
	set( 0x1db, "U"); /* 'Ǜ' -> 'U' */
	set( 0x1dc, "u"); /* 'ǜ' -> 'u' */
	set( 0x1de, "A"); /* 'Ǟ' -> 'A' */
	set( 0x1df, "a"); /* 'ǟ' -> 'a' */
	set( 0x1e0, "A"); /* 'Ǡ' -> 'A' */
	set( 0x1e1, "a"); /* 'ǡ' -> 'a' */
	set( 0x1e2, "AE"); /* 'Ǣ' -> 'AE' */
	set( 0x1e4, "G"); /* 'Ǥ' -> 'G' */
	set( 0x1e5, "g"); /* 'ǥ' -> 'g' */
	set( 0x1e6, "G"); /* 'Ǧ' -> 'G' */
	set( 0x1e7, "g"); /* 'ǧ' -> 'g' */
	set( 0x1e8, "K"); /* 'Ǩ' -> 'K' */
	set( 0x1e9, "k"); /* 'ǩ' -> 'k' */
	set( 0x1ea, "O"); /* 'Ǫ' -> 'O' */
	set( 0x1eb, "o"); /* 'ǫ' -> 'o' */
	set( 0x1ec, "O"); /* 'Ǭ' -> 'O' */
	set( 0x1ed, "o"); /* 'ǭ' -> 'o' */
	set( 0x1ee, "3"); /* 'Ǯ' -> '3' */
	set( 0x1ef, "3"); /* 'ǯ' -> '3' */
	set( 0x1f0, "j"); /* 'ǰ' -> 'j' */
	set( 0x1f1, "DZ"); /* 'Ǳ' -> 'DZ' */
	set( 0x1f2, "Dz"); /* 'ǲ' -> 'Dz' */
	set( 0x1f3, "dz"); /* 'ǳ' -> 'dz' */
	set( 0x1f4, "G"); /* 'Ǵ' -> 'G' */
	set( 0x1f5, "g"); /* 'ǵ' -> 'g' */
	set( 0x1f8, "N"); /* 'Ǹ' -> 'N' */
	set( 0x1f9, "n"); /* 'ǹ' -> 'n' */
	set( 0x1fa, "A"); /* 'Ǻ' -> 'A' */
	set( 0x1fb, "a"); /* 'ǻ' -> 'a' */
	set( 0x1fc, "AE"); /* 'Ǽ' -> 'AE' */
	set( 0x1fd, "ae"); /* 'ǽ' -> 'ae' */
	set( 0x1fe, "O"); /* 'Ǿ' -> 'O' */
	set( 0x1ff, "o"); /* 'ǿ' -> 'o' */
	set( 0x200, "A"); /* 'Ȁ' -> 'A' */
	set( 0x201, "a"); /* 'ȁ' -> 'a' */
	set( 0x202, "A"); /* 'Ȃ' -> 'A' */
	set( 0x203, "a"); /* 'ȃ' -> 'a' */
	set( 0x204, "E"); /* 'Ȅ' -> 'E' */
	set( 0x205, "e"); /* 'ȅ' -> 'e' */
	set( 0x206, "E"); /* 'Ȇ' -> 'E' */
	set( 0x207, "e"); /* 'ȇ' -> 'e' */
	set( 0x208, "I"); /* 'Ȉ' -> 'I' */
	set( 0x209, "i"); /* 'ȉ' -> 'i' */
	set( 0x20a, "I"); /* 'Ȋ' -> 'I' */
	set( 0x20b, "i"); /* 'ȋ' -> 'i' */
	set( 0x20c, "O"); /* 'Ȍ' -> 'O' */
	set( 0x20d, "o"); /* 'ȍ' -> 'o' */
	set( 0x20e, "O"); /* 'Ȏ' -> 'O' */
	set( 0x20f, "o"); /* 'ȏ' -> 'o' */
	set( 0x210, "R"); /* 'Ȑ' -> 'R' */
	set( 0x211, "r"); /* 'ȑ' -> 'r' */
	set( 0x212, "R"); /* 'Ȓ' -> 'R' */
	set( 0x213, "r"); /* 'ȓ' -> 'r' */
	set( 0x214, "U"); /* 'Ȕ' -> 'U' */
	set( 0x215, "u"); /* 'ȕ' -> 'u' */
	set( 0x216, "U"); /* 'Ȗ' -> 'U' */
	set( 0x217, "u"); /* 'ȗ' -> 'u' */
	set( 0x218, "S"); /* 'Ș' -> 'S' */
	set( 0x219, "s"); /* 'ș' -> 's' */
	set( 0x21a, "T"); /* 'Ț' -> 'T' */
	set( 0x21b, "t"); /* 'ț' -> 't' */
	set( 0x21e, "H"); /* 'Ȟ' -> 'H' */
	set( 0x21f, "h"); /* 'ȟ' -> 'h' */
	set( 0x220, "N"); /* 'Ƞ' -> 'ƞ' */
	set( 0x222, "8"); /* 'Ȣ' -> '8' */
	set( 0x223, "8"); /* 'ȣ' -> '8' */
	set( 0x224, "Z"); /* 'Ȥ' -> 'Z' */
	set( 0x225, "z"); /* 'ȥ' -> 'z' */
	set( 0x226, "A"); /* 'Ȧ' -> 'A' */
	set( 0x227, "a"); /* 'ȧ' -> 'a' */
	set( 0x228, "E"); /* 'Ȩ' -> 'E' */
	set( 0x229, "e"); /* 'ȩ' -> 'e' */
	set( 0x22a, "O"); /* 'Ȫ' -> 'O' */
	set( 0x22b, "o"); /* 'ȫ' -> 'o' */
	set( 0x22c, "O"); /* 'Ȭ' -> 'O' */
	set( 0x22d, "o"); /* 'ȭ' -> 'o' */
	set( 0x22e, "O"); /* 'Ȯ' -> 'O' */
	set( 0x22f, "o"); /* 'ȯ' -> 'o' */
	set( 0x230, "O"); /* 'Ȱ' -> 'O' */
	set( 0x231, "o"); /* 'ȱ' -> 'o' */
	set( 0x232, "Y"); /* 'Ȳ' -> 'Y' */
	set( 0x233, "y"); /* 'ȳ' -> 'y' */
	set( 0x23a, "A"); /* 'Ⱥ' -> 'A' */
	set( 0x23b, "C"); /* 'Ȼ' -> 'C' */
	set( 0x23c, "c"); /* 'ȼ' -> 'c' */
	set( 0x23d, "L"); /* 'Ƚ' -> 'L' */
	set( 0x23e, "T"); /* 'Ⱦ' -> 'T' */
	set( 0x23f, "s"); /* 'ȿ' -> 's' */
	set( 0x240, "z"); /* 'ɀ' -> 'z' */
	set( 0x241, "z"); /* 'Ɂ' -> 'z' */
	set( 0x242, "z"); /* 'ɂ' -> 'z' */
	set( 0x243, "B"); /* 'Ƀ' -> 'B' */
	set( 0x244, "A"); /* 'Ʉ' -> 'A' */
	set( 0x245, "A"); /* 'Ʌ' -> 'A' */
	set( 0x246, "E"); /* 'Ɇ' -> 'E' */
	set( 0x247, "e"); /* 'ɇ' -> 'e' */
	set( 0x248, "J"); /* 'Ɉ' -> 'J' */
	set( 0x249, "j"); /* 'ɉ' -> 'j' */
	set( 0x24a, "Q"); /* 'Ɋ' -> 'Q' */
	set( 0x24b, "q"); /* 'ɋ' -> 'q' */
	set( 0x24c, "R"); /* 'Ɍ' -> 'R' */
	set( 0x24d, "r"); /* 'ɍ' -> 'r' */
	set( 0x24e, "Y"); /* 'Ɏ' -> 'Y' */
	set( 0x24f, "y"); /* 'ɏ' -> 'y' */
	set( 0x250, "a"); /* 'ɐ' -> 'a' */
	set( 0x251, "a"); /* 'ɑ' -> 'a' */
	set( 0x252, "d"); /* 'ɒ' -> 'd' */
	set( 0x253, "b"); /* 'ɓ' -> 'b' */
	set( 0x254, "c"); /* 'ɔ' -> 'c' */
	set( 0x256, "d"); /* 'ɖ' -> 'd' */
	set( 0x257, "d"); /* 'ɗ' -> 'd' */
	set( 0x259, "e"); /* 'ə' -> 'e' */
	set( 0x25b, "e"); /* 'ɛ' -> 'e' */
	set( 0x260, "g"); /* 'ɠ' -> 'g' */
	set( 0x263, "g"); /* 'ɣ' -> 'g' */
	set( 0x268, "i"); /* 'ɨ' -> 'i' */
	set( 0x269, "t"); /* 'ɩ' -> 't' */
	set( 0x26b, "l"); /* 'ɫ' -> 'l' */
	set( 0x26f, "w"); /* 'ɯ' -> 'w' */
	set( 0x271, "m"); /* 'ɱ' -> 'm' */
	set( 0x272, "n"); /* 'ɲ' -> 'n' */
	set( 0x275, "o"); /* 'ɵ' -> 'o' */
	set( 0x27d, "r"); /* 'ɽ' -> 'r' */
	set( 0x280, "R"); /* 'ʀ' -> 'R' */
	set( 0x283, "s"); /* 'ʃ' -> 's' */
	set( 0x288, "t"); /* 'ʈ' -> 't' */
	set( 0x289, "u"); /* 'ʉ' -> 'u' */
	set( 0x28a, "v"); /* 'ʊ' -> 'v' */
	set( 0x28b, "v"); /* 'ʋ' -> 'v' */
	set( 0x28c, "v"); /* 'ʌ' -> 'v' */
	set( 0x292, "3"); /* 'ʒ' -> 'Ʒ' */
	set( 0x2b0, "h"); /* 'ʰ' -> 'h' */
	set( 0x2b2, "j"); /* 'ʲ' -> 'j' */
	set( 0x2b3, "r"); /* 'ʳ' -> 'r' */
	set( 0x2b7, "w"); /* 'ʷ' -> 'w' */
	set( 0x2b8, "y"); /* 'ʸ' -> 'y' */
	set( 0x2d8, " "); /* '˘' -> ' ' */
	set( 0x2d9, " "); /* '˙' -> ' ' */
	set( 0x2da, " "); /* '˚' -> ' ' */
	set( 0x2db, " "); /* '˛' -> ' ' */
	set( 0x2dc, " "); /* '˜' -> ' ' */
	set( 0x2dd, " "); /* '˝' -> ' ' */
	set( 0x2e1, "l"); /* 'ˡ' -> 'l' */
	set( 0x2e2, "s"); /* 'ˢ' -> 's' */
	set( 0x2e3, "x"); /* 'ˣ' -> 'x' */
	set( 0x37a, " "); /* 'ͺ' -> ' ' */
	set( 0x37e, ";"); /* ';' -> ';' */
	set( 0x384, " "); /* '΄' -> ' ' */
	set( 0x385, " "); /* '΅' -> ' ' */
	set( 0x3f0, "k"); /* 'ϰ' -> 'k' */
	set( 0x3f1, "p"); /* 'ϱ' -> 'p' */
	set( 0x3f2, "c"); /* 'ϲ' -> 'c' */
	set( 0x3f4, "o"); /* 'ϴ' -> 'o' */
	set( 0x3f5, "e"); /* 'ϵ' -> 'e' */
	set( 0x3f7, "p"); /* 'Ϸ' -> 'p' */
	set( 0x3f8, "p"); /* 'ϸ' -> 'p' */
	set( 0x3f9, "C"); /* 'Ϲ' -> 'C' */
	set( 0x3fa, "M"); /* 'Ϻ' -> 'M' */
	set( 0x3fb, "m"); /* 'ϻ' -> 'm' */
	set( 0x3fd, "C"); /* 'Ͻ' -> 'C' */
	set( 0x3fe, "C"); /* 'Ͼ' -> 'C' */
	set( 0x3ff, "C"); /* 'Ͽ' -> 'C' */
	set( 0x400, "E"); /* 'Ѐ' -> 'E' */
	set( 0x401, "E"); /* 'Ё' -> 'E' */
	set( 0x402, "h"); /* 'Ђ' -> 'h' */
	set( 0x403, "R"); /* 'Ѓ' -> 'R' */
	set( 0x404, "e"); /* 'Є' -> 'e' */
	set( 0x405, "S"); /* 'Ѕ' -> 'S' */
	set( 0x406, "I"); /* 'І' -> 'I' */
	set( 0x407, "I"); /* 'Ї' -> 'I' */
	set( 0x408, "J"); /* 'Ј' -> 'J' */
	set( 0x409, "Jb"); /* 'Љ' -> 'Jb' */
	set( 0x40a, "Hb"); /* 'Њ' -> 'Hb' */
	set( 0x40b, "h"); /* 'Ћ' -> 'K' */
	set( 0x40c, "K"); /* 'Ќ' -> 'ќ' */
	set( 0x40d, "N"); /* 'Ѝ' -> 'N' */
	set( 0x40e, "y"); /* 'Ў' -> 'y' */
	set( 0x40f, "U"); /* 'Џ' -> 'U' */
	set( 0x1d2c, "A"); /* 'ᴬ' -> 'A' */
	set( 0x1d2e, "B"); /* 'ᴮ' -> 'B' */
	set( 0x1d30, "D"); /* 'ᴰ' -> 'D' */
	set( 0x1d31, "E"); /* 'ᴱ' -> 'E' */
	set( 0x1d33, "G"); /* 'ᴳ' -> 'G' */
	set( 0x1d34, "H"); /* 'ᴴ' -> 'H' */
	set( 0x1d35, "I"); /* 'ᴵ' -> 'I' */
	set( 0x1d36, "J"); /* 'ᴶ' -> 'J' */
	set( 0x1d37, "K"); /* 'ᴷ' -> 'K' */
	set( 0x1d38, "L"); /* 'ᴸ' -> 'L' */
	set( 0x1d39, "M"); /* 'ᴹ' -> 'M' */
	set( 0x1d3a, "N"); /* 'ᴺ' -> 'N' */
	set( 0x1d3c, "O"); /* 'ᴼ' -> 'O' */
	set( 0x1d3e, "P"); /* 'ᴾ' -> 'P' */
	set( 0x1d3f, "R"); /* 'ᴿ' -> 'R' */
	set( 0x1d40, "T"); /* 'ᵀ' -> 'T' */
	set( 0x1d41, "U"); /* 'ᵁ' -> 'U' */
	set( 0x1d42, "W"); /* 'ᵂ' -> 'W' */
	set( 0x1d43, "a"); /* 'ᵃ' -> 'a' */
	set( 0x1d47, "b"); /* 'ᵇ' -> 'b' */
	set( 0x1d48, "d"); /* 'ᵈ' -> 'd' */
	set( 0x1d49, "e"); /* 'ᵉ' -> 'e' */
	set( 0x1d4d, "g"); /* 'ᵍ' -> 'g' */
	set( 0x1d4f, "k"); /* 'ᵏ' -> 'k' */
	set( 0x1d50, "m"); /* 'ᵐ' -> 'm' */
	set( 0x1d52, "o"); /* 'ᵒ' -> 'o' */
	set( 0x1d56, "p"); /* 'ᵖ' -> 'p' */
	set( 0x1d57, "t"); /* 'ᵗ' -> 't' */
	set( 0x1d58, "u"); /* 'ᵘ' -> 'u' */
	set( 0x1d5b, "v"); /* 'ᵛ' -> 'v' */
	set( 0x1d62, "i"); /* 'ᵢ' -> 'i' */
	set( 0x1d63, "r"); /* 'ᵣ' -> 'r' */
	set( 0x1d64, "u"); /* 'ᵤ' -> 'u' */
	set( 0x1d65, "v"); /* 'ᵥ' -> 'v' */
	set( 0x1d9c, "c"); /* 'ᶜ' -> 'c' */
	set( 0x1da0, "f"); /* 'ᶠ' -> 'f' */
	set( 0x1dbb, "z"); /* 'ᶻ' -> 'z' */
	set( 0x1e00, "A"); /* 'Ḁ' -> 'A' */
	set( 0x1e01, "a"); /* 'ḁ' -> 'a' */
	set( 0x1e02, "B"); /* 'Ḃ' -> 'B' */
	set( 0x1e03, "b"); /* 'ḃ' -> 'b' */
	set( 0x1e04, "B"); /* 'Ḅ' -> 'B' */
	set( 0x1e05, "b"); /* 'ḅ' -> 'b' */
	set( 0x1e06, "B"); /* 'Ḇ' -> 'B' */
	set( 0x1e07, "b"); /* 'ḇ' -> 'b' */
	set( 0x1e08, "C"); /* 'Ḉ' -> 'C' */
	set( 0x1e09, "c"); /* 'ḉ' -> 'c' */
	set( 0x1e0a, "D"); /* 'Ḋ' -> 'D' */
	set( 0x1e0b, "d"); /* 'ḋ' -> 'd' */
	set( 0x1e0c, "D"); /* 'Ḍ' -> 'D' */
	set( 0x1e0d, "d"); /* 'ḍ' -> 'd' */
	set( 0x1e0e, "D"); /* 'Ḏ' -> 'D' */
	set( 0x1e0f, "d"); /* 'ḏ' -> 'd' */
	set( 0x1e10, "D"); /* 'Ḑ' -> 'D' */
	set( 0x1e11, "d"); /* 'ḑ' -> 'd' */
	set( 0x1e12, "D"); /* 'Ḓ' -> 'D' */
	set( 0x1e13, "d"); /* 'ḓ' -> 'd' */
	set( 0x1e14, "E"); /* 'Ḕ' -> 'E' */
	set( 0x1e15, "e"); /* 'ḕ' -> 'e' */
	set( 0x1e16, "E"); /* 'Ḗ' -> 'E' */
	set( 0x1e17, "e"); /* 'ḗ' -> 'e' */
	set( 0x1e18, "E"); /* 'Ḙ' -> 'E' */
	set( 0x1e19, "e"); /* 'ḙ' -> 'e' */
	set( 0x1e1a, "E"); /* 'Ḛ' -> 'E' */
	set( 0x1e1b, "e"); /* 'ḛ' -> 'e' */
	set( 0x1e1c, "E"); /* 'Ḝ' -> 'E' */
	set( 0x1e1d, "e"); /* 'ḝ' -> 'e' */
	set( 0x1e1e, "F"); /* 'Ḟ' -> 'F' */
	set( 0x1e1f, "f"); /* 'ḟ' -> 'f' */
	set( 0x1e20, "G"); /* 'Ḡ' -> 'G' */
	set( 0x1e21, "g"); /* 'ḡ' -> 'g' */
	set( 0x1e22, "H"); /* 'Ḣ' -> 'H' */
	set( 0x1e23, "h"); /* 'ḣ' -> 'h' */
	set( 0x1e24, "H"); /* 'Ḥ' -> 'H' */
	set( 0x1e25, "h"); /* 'ḥ' -> 'h' */
	set( 0x1e26, "H"); /* 'Ḧ' -> 'H' */
	set( 0x1e27, "h"); /* 'ḧ' -> 'h' */
	set( 0x1e28, "H"); /* 'Ḩ' -> 'H' */
	set( 0x1e29, "h"); /* 'ḩ' -> 'h' */
	set( 0x1e2a, "H"); /* 'Ḫ' -> 'H' */
	set( 0x1e2b, "h"); /* 'ḫ' -> 'h' */
	set( 0x1e2c, "I"); /* 'Ḭ' -> 'I' */
	set( 0x1e2d, "i"); /* 'ḭ' -> 'i' */
	set( 0x1e2e, "I"); /* 'Ḯ' -> 'I' */
	set( 0x1e2f, "i"); /* 'ḯ' -> 'i' */
	set( 0x1e30, "K"); /* 'Ḱ' -> 'K' */
	set( 0x1e31, "k"); /* 'ḱ' -> 'k' */
	set( 0x1e32, "K"); /* 'Ḳ' -> 'K' */
	set( 0x1e33, "k"); /* 'ḳ' -> 'k' */
	set( 0x1e34, "K"); /* 'Ḵ' -> 'K' */
	set( 0x1e35, "k"); /* 'ḵ' -> 'k' */
	set( 0x1e36, "L"); /* 'Ḷ' -> 'L' */
	set( 0x1e37, "l"); /* 'ḷ' -> 'l' */
	set( 0x1e38, "L"); /* 'Ḹ' -> 'L' */
	set( 0x1e39, "l"); /* 'ḹ' -> 'l' */
	set( 0x1e3a, "L"); /* 'Ḻ' -> 'L' */
	set( 0x1e3b, "l"); /* 'ḻ' -> 'l' */
	set( 0x1e3c, "L"); /* 'Ḽ' -> 'L' */
	set( 0x1e3d, "l"); /* 'ḽ' -> 'l' */
	set( 0x1e3e, "M"); /* 'Ḿ' -> 'M' */
	set( 0x1e3f, "m"); /* 'ḿ' -> 'm' */
	set( 0x1e40, "M"); /* 'Ṁ' -> 'M' */
	set( 0x1e41, "m"); /* 'ṁ' -> 'm' */
	set( 0x1e42, "M"); /* 'Ṃ' -> 'M' */
	set( 0x1e43, "m"); /* 'ṃ' -> 'm' */
	set( 0x1e44, "N"); /* 'Ṅ' -> 'N' */
	set( 0x1e45, "n"); /* 'ṅ' -> 'n' */
	set( 0x1e46, "N"); /* 'Ṇ' -> 'N' */
	set( 0x1e47, "n"); /* 'ṇ' -> 'n' */
	set( 0x1e48, "N"); /* 'Ṉ' -> 'N' */
	set( 0x1e49, "n"); /* 'ṉ' -> 'n' */
	set( 0x1e4a, "N"); /* 'Ṋ' -> 'N' */
	set( 0x1e4b, "n"); /* 'ṋ' -> 'n' */
	set( 0x1e4c, "O"); /* 'Ṍ' -> 'O' */
	set( 0x1e4d, "o"); /* 'ṍ' -> 'o' */
	set( 0x1e4e, "O"); /* 'Ṏ' -> 'O' */
	set( 0x1e4f, "o"); /* 'ṏ' -> 'o' */
	set( 0x1e50, "O"); /* 'Ṑ' -> 'O' */
	set( 0x1e51, "o"); /* 'ṑ' -> 'o' */
	set( 0x1e52, "O"); /* 'Ṓ' -> 'O' */
	set( 0x1e53, "o"); /* 'ṓ' -> 'o' */
	set( 0x1e54, "P"); /* 'Ṕ' -> 'P' */
	set( 0x1e55, "p"); /* 'ṕ' -> 'p' */
	set( 0x1e56, "P"); /* 'Ṗ' -> 'P' */
	set( 0x1e57, "p"); /* 'ṗ' -> 'p' */
	set( 0x1e58, "R"); /* 'Ṙ' -> 'R' */
	set( 0x1e59, "r"); /* 'ṙ' -> 'r' */
	set( 0x1e5a, "R"); /* 'Ṛ' -> 'R' */
	set( 0x1e5b, "r"); /* 'ṛ' -> 'r' */
	set( 0x1e5c, "R"); /* 'Ṝ' -> 'R' */
	set( 0x1e5d, "r"); /* 'ṝ' -> 'r' */
	set( 0x1e5e, "R"); /* 'Ṟ' -> 'R' */
	set( 0x1e5f, "r"); /* 'ṟ' -> 'r' */
	set( 0x1e60, "S"); /* 'Ṡ' -> 'S' */
	set( 0x1e61, "s"); /* 'ṡ' -> 's' */
	set( 0x1e62, "S"); /* 'Ṣ' -> 'S' */
	set( 0x1e63, "s"); /* 'ṣ' -> 's' */
	set( 0x1e64, "S"); /* 'Ṥ' -> 'S' */
	set( 0x1e65, "s"); /* 'ṥ' -> 's' */
	set( 0x1e66, "S"); /* 'Ṧ' -> 'S' */
	set( 0x1e67, "s"); /* 'ṧ' -> 's' */
	set( 0x1e68, "S"); /* 'Ṩ' -> 'S' */
	set( 0x1e69, "s"); /* 'ṩ' -> 's' */
	set( 0x1e6a, "T"); /* 'Ṫ' -> 'T' */
	set( 0x1e6b, "t"); /* 'ṫ' -> 't' */
	set( 0x1e6c, "T"); /* 'Ṭ' -> 'T' */
	set( 0x1e6d, "t"); /* 'ṭ' -> 't' */
	set( 0x1e6e, "T"); /* 'Ṯ' -> 'T' */
	set( 0x1e6f, "t"); /* 'ṯ' -> 't' */
	set( 0x1e70, "T"); /* 'Ṱ' -> 'T' */
	set( 0x1e71, "t"); /* 'ṱ' -> 't' */
	set( 0x1e72, "U"); /* 'Ṳ' -> 'U' */
	set( 0x1e73, "u"); /* 'ṳ' -> 'u' */
	set( 0x1e74, "U"); /* 'Ṵ' -> 'U' */
	set( 0x1e75, "u"); /* 'ṵ' -> 'u' */
	set( 0x1e76, "U"); /* 'Ṷ' -> 'U' */
	set( 0x1e77, "u"); /* 'ṷ' -> 'u' */
	set( 0x1e78, "U"); /* 'Ṹ' -> 'U' */
	set( 0x1e79, "u"); /* 'ṹ' -> 'u' */
	set( 0x1e7a, "U"); /* 'Ṻ' -> 'U' */
	set( 0x1e7b, "u"); /* 'ṻ' -> 'u' */
	set( 0x1e7c, "V"); /* 'Ṽ' -> 'V' */
	set( 0x1e7d, "v"); /* 'ṽ' -> 'v' */
	set( 0x1e7e, "V"); /* 'Ṿ' -> 'V' */
	set( 0x1e7f, "v"); /* 'ṿ' -> 'v' */
	set( 0x1e80, "W"); /* 'Ẁ' -> 'W' */
	set( 0x1e81, "w"); /* 'ẁ' -> 'w' */
	set( 0x1e82, "W"); /* 'Ẃ' -> 'W' */
	set( 0x1e83, "w"); /* 'ẃ' -> 'w' */
	set( 0x1e84, "W"); /* 'Ẅ' -> 'W' */
	set( 0x1e85, "w"); /* 'ẅ' -> 'w' */
	set( 0x1e86, "W"); /* 'Ẇ' -> 'W' */
	set( 0x1e87, "w"); /* 'ẇ' -> 'w' */
	set( 0x1e88, "W"); /* 'Ẉ' -> 'W' */
	set( 0x1e89, "w"); /* 'ẉ' -> 'w' */
	set( 0x1e8a, "X"); /* 'Ẋ' -> 'X' */
	set( 0x1e8b, "x"); /* 'ẋ' -> 'x' */
	set( 0x1e8c, "X"); /* 'Ẍ' -> 'X' */
	set( 0x1e8d, "x"); /* 'ẍ' -> 'x' */
	set( 0x1e8e, "Y"); /* 'Ẏ' -> 'Y' */
	set( 0x1e8f, "y"); /* 'ẏ' -> 'y' */
	set( 0x1e90, "Z"); /* 'Ẑ' -> 'Z' */
	set( 0x1e91, "z"); /* 'ẑ' -> 'z' */
	set( 0x1e92, "Z"); /* 'Ẓ' -> 'Z' */
	set( 0x1e93, "z"); /* 'ẓ' -> 'z' */
	set( 0x1e94, "Z"); /* 'Ẕ' -> 'Z' */
	set( 0x1e95, "z"); /* 'ẕ' -> 'z' */
	set( 0x1e96, "h"); /* 'ẖ' -> 'h' */
	set( 0x1e97, "t"); /* 'ẗ' -> 't' */
	set( 0x1e98, "w"); /* 'ẘ' -> 'w' */
	set( 0x1e99, "y"); /* 'ẙ' -> 'y' */
	set( 0x1e9a, "a"); /* 'ẚ' -> 'a' */
	set( 0x1e9b, "s"); /* 'ẛ' -> 's' */
	set( 0x1ea0, "A"); /* 'Ạ' -> 'A' */
	set( 0x1ea1, "a"); /* 'ạ' -> 'a' */
	set( 0x1ea2, "A"); /* 'Ả' -> 'A' */
	set( 0x1ea3, "a"); /* 'ả' -> 'a' */
	set( 0x1ea4, "A"); /* 'Ấ' -> 'A' */
	set( 0x1ea5, "a"); /* 'ấ' -> 'a' */
	set( 0x1ea6, "A"); /* 'Ầ' -> 'A' */
	set( 0x1ea7, "a"); /* 'ầ' -> 'a' */
	set( 0x1ea8, "A"); /* 'Ẩ' -> 'A' */
	set( 0x1ea9, "a"); /* 'ẩ' -> 'a' */
	set( 0x1eaa, "A"); /* 'Ẫ' -> 'A' */
	set( 0x1eab, "a"); /* 'ẫ' -> 'a' */
	set( 0x1eac, "A"); /* 'Ậ' -> 'A' */
	set( 0x1ead, "a"); /* 'ậ' -> 'a' */
	set( 0x1eae, "A"); /* 'Ắ' -> 'A' */
	set( 0x1eaf, "a"); /* 'ắ' -> 'a' */
	set( 0x1eb0, "A"); /* 'Ằ' -> 'A' */
	set( 0x1eb1, "a"); /* 'ằ' -> 'a' */
	set( 0x1eb2, "A"); /* 'Ẳ' -> 'A' */
	set( 0x1eb3, "a"); /* 'ẳ' -> 'a' */
	set( 0x1eb4, "A"); /* 'Ẵ' -> 'A' */
	set( 0x1eb5, "a"); /* 'ẵ' -> 'a' */
	set( 0x1eb6, "A"); /* 'Ặ' -> 'A' */
	set( 0x1eb7, "a"); /* 'ặ' -> 'a' */
	set( 0x1eb8, "E"); /* 'Ẹ' -> 'E' */
	set( 0x1eb9, "e"); /* 'ẹ' -> 'e' */
	set( 0x1eba, "E"); /* 'Ẻ' -> 'E' */
	set( 0x1ebb, "e"); /* 'ẻ' -> 'e' */
	set( 0x1ebc, "E"); /* 'Ẽ' -> 'E' */
	set( 0x1ebd, "e"); /* 'ẽ' -> 'e' */
	set( 0x1ebe, "E"); /* 'Ế' -> 'E' */
	set( 0x1ebf, "e"); /* 'ế' -> 'e' */
	set( 0x1ec0, "E"); /* 'Ề' -> 'E' */
	set( 0x1ec1, "e"); /* 'ề' -> 'e' */
	set( 0x1ec2, "E"); /* 'Ể' -> 'E' */
	set( 0x1ec3, "e"); /* 'ể' -> 'e' */
	set( 0x1ec4, "E"); /* 'Ễ' -> 'E' */
	set( 0x1ec5, "e"); /* 'ễ' -> 'e' */
	set( 0x1ec6, "E"); /* 'Ệ' -> 'E' */
	set( 0x1ec7, "e"); /* 'ệ' -> 'e' */
	set( 0x1ec8, "I"); /* 'Ỉ' -> 'I' */
	set( 0x1ec9, "i"); /* 'ỉ' -> 'i' */
	set( 0x1eca, "I"); /* 'Ị' -> 'I' */
	set( 0x1ecb, "i"); /* 'ị' -> 'i' */
	set( 0x1ecc, "O"); /* 'Ọ' -> 'O' */
	set( 0x1ecd, "o"); /* 'ọ' -> 'o' */
	set( 0x1ece, "O"); /* 'Ỏ' -> 'O' */
	set( 0x1ecf, "o"); /* 'ỏ' -> 'o' */
	set( 0x1ed0, "O"); /* 'Ố' -> 'O' */
	set( 0x1ed1, "o"); /* 'ố' -> 'o' */
	set( 0x1ed2, "O"); /* 'Ồ' -> 'O' */
	set( 0x1ed3, "o"); /* 'ồ' -> 'o' */
	set( 0x1ed4, "O"); /* 'Ổ' -> 'O' */
	set( 0x1ed5, "o"); /* 'ổ' -> 'o' */
	set( 0x1ed6, "O"); /* 'Ỗ' -> 'O' */
	set( 0x1ed7, "o"); /* 'ỗ' -> 'o' */
	set( 0x1ed8, "O"); /* 'Ộ' -> 'O' */
	set( 0x1ed9, "o"); /* 'ộ' -> 'o' */
	set( 0x1eda, "O"); /* 'Ớ' -> 'O' */
	set( 0x1edb, "o"); /* 'ớ' -> 'o' */
	set( 0x1edc, "O"); /* 'Ờ' -> 'O' */
	set( 0x1edd, "o"); /* 'ờ' -> 'o' */
	set( 0x1ede, "O"); /* 'Ở' -> 'O' */
	set( 0x1edf, "o"); /* 'ở' -> 'o' */
	set( 0x1ee0, "O"); /* 'Ỡ' -> 'O' */
	set( 0x1ee1, "o"); /* 'ỡ' -> 'o' */
	set( 0x1ee2, "O"); /* 'Ợ' -> 'O' */
	set( 0x1ee3, "o"); /* 'ợ' -> 'o' */
	set( 0x1ee4, "U"); /* 'Ụ' -> 'U' */
	set( 0x1ee5, "u"); /* 'ụ' -> 'u' */
	set( 0x1ee6, "U"); /* 'Ủ' -> 'U' */
	set( 0x1ee7, "u"); /* 'ủ' -> 'u' */
	set( 0x1ee8, "U"); /* 'Ứ' -> 'U' */
	set( 0x1ee9, "u"); /* 'ứ' -> 'u' */
	set( 0x1eea, "U"); /* 'Ừ' -> 'U' */
	set( 0x1eeb, "u"); /* 'ừ' -> 'u' */
	set( 0x1eec, "U"); /* 'Ử' -> 'U' */
	set( 0x1eed, "u"); /* 'ử' -> 'u' */
	set( 0x1eee, "U"); /* 'Ữ' -> 'U' */
	set( 0x1eef, "u"); /* 'ữ' -> 'u' */
	set( 0x1ef0, "U"); /* 'Ự' -> 'U' */
	set( 0x1ef1, "u"); /* 'ự' -> 'u' */
	set( 0x1ef2, "Y"); /* 'Ỳ' -> 'Y' */
	set( 0x1ef3, "y"); /* 'ỳ' -> 'y' */
	set( 0x1ef4, "Y"); /* 'Ỵ' -> 'Y' */
	set( 0x1ef5, "y"); /* 'ỵ' -> 'y' */
	set( 0x1ef6, "Y"); /* 'Ỷ' -> 'Y' */
	set( 0x1ef7, "y"); /* 'ỷ' -> 'y' */
	set( 0x1ef8, "Y"); /* 'Ỹ' -> 'Y' */
	set( 0x1ef9, "y"); /* 'ỹ' -> 'y' */
	set( 0x1fbd, " "); /* '᾽' -> ' ' */
	set( 0x1fbf, " "); /* '᾿' -> ' ' */
	set( 0x1fc0, " "); /* '῀' -> ' ' */
	set( 0x1fc1, " "); /* '῁' -> ' ' */
	set( 0x1fcd, " "); /* '῍' -> ' ' */
	set( 0x1fce, " "); /* '῎' -> ' ' */
	set( 0x1fcf, " "); /* '῏' -> ' ' */
	set( 0x1fdd, " "); /* '῝' -> ' ' */
	set( 0x1fde, " "); /* '῞' -> ' ' */
	set( 0x1fdf, " "); /* '῟' -> ' ' */
	set( 0x1fed, " "); /* '῭' -> ' ' */
	set( 0x1fee, " "); /* '΅' -> ' ' */
	set( 0x1fef, "`"); /* '`' -> '`' */
	set( 0x1ffd, " "); /* '´' -> ' ' */
	set( 0x1ffe, " "); /* '῾' -> ' ' */
	set( 0x2000, " "); /* ' ' -> ' ' */
	set( 0x2001, " "); /* ' ' -> ' ' */
	set( 0x2002, " "); /* ' ' -> ' ' */
	set( 0x2003, " "); /* ' ' -> ' ' */
	set( 0x2004, " "); /* ' ' -> ' ' */
	set( 0x2005, " "); /* ' ' -> ' ' */
	set( 0x2006, " "); /* ' ' -> ' ' */
	set( 0x2007, " "); /* ' ' -> ' ' */
	set( 0x2008, " "); /* ' ' -> ' ' */
	set( 0x2009, " "); /* ' ' -> ' ' */
	set( 0x200a, " "); /* ' ' -> ' ' */
	set( 0x2017, " "); /* '‗' -> ' ' */
	set( 0x2024, "."); /* '․' -> '.' */
	set( 0x2025, ".."); /* '‥' -> '..' */
	set( 0x2026, "..."); /* '…' -> '...' */
	set( 0x202f, " "); /* ' ' -> ' ' */
	set( 0x203c, "!!"); /* '‼' -> '!!' */
	set( 0x203e, " "); /* '‾' -> ' ' */
	set( 0x2047, "??"); /* '⁇' -> '??' */
	set( 0x2048, "?!"); /* '⁈' -> '?!' */
	set( 0x2049, "!?"); /* '⁉' -> '!?' */
	set( 0x205f, " "); /* ' ' -> ' ' */
	set( 0x2070, "0"); /* '⁰' -> '0' */
	set( 0x2071, "i"); /* 'ⁱ' -> 'i' */
	set( 0x2074, "4"); /* '⁴' -> '4' */
	set( 0x2075, "5"); /* '⁵' -> '5' */
	set( 0x2076, "6"); /* '⁶' -> '6' */
	set( 0x2077, "7"); /* '⁷' -> '7' */
	set( 0x2078, "8"); /* '⁸' -> '8' */
	set( 0x2079, "9"); /* '⁹' -> '9' */
	set( 0x207a, "+"); /* '⁺' -> '+' */
	set( 0x207c, "="); /* '⁼' -> '=' */
	set( 0x207d, "("); /* '⁽' -> '(' */
	set( 0x207e, ")"); /* '⁾' -> ')' */
	set( 0x207f, "n"); /* 'ⁿ' -> 'n' */
	set( 0x2080, "0"); /* '₀' -> '0' */
	set( 0x2081, "1"); /* '₁' -> '1' */
	set( 0x2082, "2"); /* '₂' -> '2' */
	set( 0x2083, "3"); /* '₃' -> '3' */
	set( 0x2084, "4"); /* '₄' -> '4' */
	set( 0x2085, "5"); /* '₅' -> '5' */
	set( 0x2086, "6"); /* '₆' -> '6' */
	set( 0x2087, "7"); /* '₇' -> '7' */
	set( 0x2088, "8"); /* '₈' -> '8' */
	set( 0x2089, "9"); /* '₉' -> '9' */
	set( 0x208a, "+"); /* '₊' -> '+' */
	set( 0x208c, "="); /* '₌' -> '=' */
	set( 0x208d, "("); /* '₍' -> '(' */
	set( 0x208e, ")"); /* '₎' -> ')' */
	set( 0x2090, "a"); /* 'ₐ' -> 'a' */
	set( 0x2091, "e"); /* 'ₑ' -> 'e' */
	set( 0x2092, "o"); /* 'ₒ' -> 'o' */
	set( 0x2093, "x"); /* 'ₓ' -> 'x' */
	set( 0x20a8, "Rs"); /* '₨' -> 'Rs' */
	set( 0x2100, "a/c"); /* '℀' -> 'a/c' */
	set( 0x2101, "a/s"); /* '℁' -> 'a/s' */
	set( 0x2102, "C"); /* 'ℂ' -> 'C' */
	set( 0x2103, "C"); /* '℃' -> 'C' */
	set( 0x2105, "c/o"); /* '℅' -> 'c/o' */
	set( 0x2106, "c/u"); /* '℆' -> 'c/u' */
	set( 0x2109, "F"); /* '℉' -> 'F' */
	set( 0x210a, "g"); /* 'ℊ' -> 'g' */
	set( 0x210b, "H"); /* 'ℋ' -> 'H' */
	set( 0x210c, "H"); /* 'ℌ' -> 'H' */
	set( 0x210d, "H"); /* 'ℍ' -> 'H' */
	set( 0x210e, "h"); /* 'ℎ' -> 'h' */
	set( 0x2110, "I"); /* 'ℐ' -> 'I' */
	set( 0x2111, "I"); /* 'ℑ' -> 'I' */
	set( 0x2112, "L"); /* 'ℒ' -> 'L' */
	set( 0x2113, "l"); /* 'ℓ' -> 'l' */
	set( 0x2115, "N"); /* 'ℕ' -> 'N' */
	set( 0x2116, "No"); /* '№' -> 'No' */
	set( 0x2119, "P"); /* 'ℙ' -> 'P' */
	set( 0x211a, "Q"); /* 'ℚ' -> 'Q' */
	set( 0x211b, "R"); /* 'ℛ' -> 'R' */
	set( 0x211c, "R"); /* 'ℜ' -> 'R' */
	set( 0x211d, "R"); /* 'ℝ' -> 'R' */
	set( 0x2120, "SM"); /* '℠' -> 'SM' */
	set( 0x2121, "TEL"); /* '℡' -> 'TEL' */
	set( 0x2122, "TM"); /* '™' -> 'TM' */
	set( 0x2124, "Z"); /* 'ℤ' -> 'Z' */
	set( 0x2128, "Z"); /* 'ℨ' -> 'Z' */
	set( 0x212a, "K"); /* 'K' -> 'K' */
	set( 0x212b, "A"); /* 'Å' -> 'A' */
	set( 0x212c, "B"); /* 'ℬ' -> 'B' */
	set( 0x212d, "C"); /* 'ℭ' -> 'C' */
	set( 0x212f, "e"); /* 'ℯ' -> 'e' */
	set( 0x2130, "E"); /* 'ℰ' -> 'E' */
	set( 0x2131, "F"); /* 'ℱ' -> 'F' */
	set( 0x2133, "M"); /* 'ℳ' -> 'M' */
	set( 0x2134, "o"); /* 'ℴ' -> 'o' */
	set( 0x2139, "i"); /* 'ℹ' -> 'i' */
	set( 0x213b, "FAX"); /* '℻' -> 'FAX' */
	set( 0x2145, "D"); /* 'ⅅ' -> 'D' */
	set( 0x2146, "d"); /* 'ⅆ' -> 'd' */
	set( 0x2147, "e"); /* 'ⅇ' -> 'e' */
	set( 0x2148, "i"); /* 'ⅈ' -> 'i' */
	set( 0x2149, "j"); /* 'ⅉ' -> 'j' */
	set( 0x2150, "1/7"); /* '⅐' -> '1/7' */
	set( 0x2151, "1/9"); /* '⅑' -> '1/9' */
	set( 0x2152, "1/10"); /* '⅒' -> '1/10' */
	set( 0x2153, "1/3"); /* '⅓' -> '1/3' */
	set( 0x2154, "2/3"); /* '⅔' -> '2/3' */
	set( 0x2155, "1/5"); /* '⅕' -> '1/5' */
	set( 0x2156, "2/5"); /* '⅖' -> '2/5' */
	set( 0x2157, "3/5"); /* '⅗' -> '3/5' */
	set( 0x2158, "4/5"); /* '⅘' -> '4/5' */
	set( 0x2159, "1/6"); /* '⅙' -> '1/6' */
	set( 0x215a, "5/6"); /* '⅚' -> '5/6' */
	set( 0x215b, "1/8"); /* '⅛' -> '1/8' */
	set( 0x215c, "3/8"); /* '⅜' -> '3/8' */
	set( 0x215d, "5/8"); /* '⅝' -> '5/8' */
	set( 0x215e, "7/8"); /* '⅞' -> '7/8' */
	set( 0x215f, "1"); /* '⅟' -> '1' */
	set( 0x2160, "I"); /* 'Ⅰ' -> 'I' */
	set( 0x2161, "II"); /* 'Ⅱ' -> 'II' */
	set( 0x2162, "III"); /* 'Ⅲ' -> 'III' */
	set( 0x2163, "IV"); /* 'Ⅳ' -> 'IV' */
	set( 0x2164, "V"); /* 'Ⅴ' -> 'V' */
	set( 0x2165, "VI"); /* 'Ⅵ' -> 'VI' */
	set( 0x2166, "VII"); /* 'Ⅶ' -> 'VII' */
	set( 0x2167, "VIII"); /* 'Ⅷ' -> 'VIII' */
	set( 0x2168, "IX"); /* 'Ⅸ' -> 'IX' */
	set( 0x2169, "X"); /* 'Ⅹ' -> 'X' */
	set( 0x216a, "XI"); /* 'Ⅺ' -> 'XI' */
	set( 0x216b, "XII"); /* 'Ⅻ' -> 'XII' */
	set( 0x216c, "L"); /* 'Ⅼ' -> 'L' */
	set( 0x216d, "C"); /* 'Ⅽ' -> 'C' */
	set( 0x216e, "D"); /* 'Ⅾ' -> 'D' */
	set( 0x216f, "M"); /* 'Ⅿ' -> 'M' */
	set( 0x2170, "i"); /* 'ⅰ' -> 'i' */
	set( 0x2171, "ii"); /* 'ⅱ' -> 'ii' */
	set( 0x2172, "iii"); /* 'ⅲ' -> 'iii' */
	set( 0x2173, "iv"); /* 'ⅳ' -> 'iv' */
	set( 0x2174, "v"); /* 'ⅴ' -> 'v' */
	set( 0x2175, "vi"); /* 'ⅵ' -> 'vi' */
	set( 0x2176, "vii"); /* 'ⅶ' -> 'vii' */
	set( 0x2177, "viii"); /* 'ⅷ' -> 'viii' */
	set( 0x2178, "ix"); /* 'ⅸ' -> 'ix' */
	set( 0x2179, "x"); /* 'ⅹ' -> 'x' */
	set( 0x217a, "xi"); /* 'ⅺ' -> 'xi' */
	set( 0x217b, "xii"); /* 'ⅻ' -> 'xii' */
	set( 0x217c, "l"); /* 'ⅼ' -> 'l' */
	set( 0x217d, "c"); /* 'ⅽ' -> 'c' */
	set( 0x217e, "d"); /* 'ⅾ' -> 'd' */
	set( 0x217f, "m"); /* 'ⅿ' -> 'm' */
	set( 0x2189, "0/3"); /* '↉' -> '0/3' */
	set( 0x2260, "="); /* '≠' -> '=' */
	set( 0x226e, "<"); /* '≮' -> '<' */
	set( 0x226f, ">"); /* '≯' -> '>' */
	set( 0x2460, "1"); /* '①' -> '1' */
	set( 0x2461, "2"); /* '②' -> '2' */
	set( 0x2462, "3"); /* '③' -> '3' */
	set( 0x2463, "4"); /* '④' -> '4' */
	set( 0x2464, "5"); /* '⑤' -> '5' */
	set( 0x2465, "6"); /* '⑥' -> '6' */
	set( 0x2466, "7"); /* '⑦' -> '7' */
	set( 0x2467, "8"); /* '⑧' -> '8' */
	set( 0x2468, "9"); /* '⑨' -> '9' */
	set( 0x2469, "10"); /* '⑩' -> '10' */
	set( 0x246a, "11"); /* '⑪' -> '11' */
	set( 0x246b, "12"); /* '⑫' -> '12' */
	set( 0x246c, "13"); /* '⑬' -> '13' */
	set( 0x246d, "14"); /* '⑭' -> '14' */
	set( 0x246e, "15"); /* '⑮' -> '15' */
	set( 0x246f, "16"); /* '⑯' -> '16' */
	set( 0x2470, "17"); /* '⑰' -> '17' */
	set( 0x2471, "18"); /* '⑱' -> '18' */
	set( 0x2472, "19"); /* '⑲' -> '19' */
	set( 0x2473, "20"); /* '⑳' -> '20' */
	set( 0x2474, "(1)"); /* '⑴' -> '(1)' */
	set( 0x2475, "(2)"); /* '⑵' -> '(2)' */
	set( 0x2476, "(3)"); /* '⑶' -> '(3)' */
	set( 0x2477, "(4)"); /* '⑷' -> '(4)' */
	set( 0x2478, "(5)"); /* '⑸' -> '(5)' */
	set( 0x2479, "(6)"); /* '⑹' -> '(6)' */
	set( 0x247a, "(7)"); /* '⑺' -> '(7)' */
	set( 0x247b, "(8)"); /* '⑻' -> '(8)' */
	set( 0x247c, "(9)"); /* '⑼' -> '(9)' */
	set( 0x247d, "(10)"); /* '⑽' -> '(10)' */
	set( 0x247e, "(11)"); /* '⑾' -> '(11)' */
	set( 0x247f, "(12)"); /* '⑿' -> '(12)' */
	set( 0x2480, "(13)"); /* '⒀' -> '(13)' */
	set( 0x2481, "(14)"); /* '⒁' -> '(14)' */
	set( 0x2482, "(15)"); /* '⒂' -> '(15)' */
	set( 0x2483, "(16)"); /* '⒃' -> '(16)' */
	set( 0x2484, "(17)"); /* '⒄' -> '(17)' */
	set( 0x2485, "(18)"); /* '⒅' -> '(18)' */
	set( 0x2486, "(19)"); /* '⒆' -> '(19)' */
	set( 0x2487, "(20)"); /* '⒇' -> '(20)' */
	set( 0x2488, "1."); /* '⒈' -> '1.' */
	set( 0x2489, "2."); /* '⒉' -> '2.' */
	set( 0x248a, "3."); /* '⒊' -> '3.' */
	set( 0x248b, "4."); /* '⒋' -> '4.' */
	set( 0x248c, "5."); /* '⒌' -> '5.' */
	set( 0x248d, "6."); /* '⒍' -> '6.' */
	set( 0x248e, "7."); /* '⒎' -> '7.' */
	set( 0x248f, "8."); /* '⒏' -> '8.' */
	set( 0x2490, "9."); /* '⒐' -> '9.' */
	set( 0x2491, "10."); /* '⒑' -> '10.' */
	set( 0x2492, "11."); /* '⒒' -> '11.' */
	set( 0x2493, "12."); /* '⒓' -> '12.' */
	set( 0x2494, "13."); /* '⒔' -> '13.' */
	set( 0x2495, "14."); /* '⒕' -> '14.' */
	set( 0x2496, "15."); /* '⒖' -> '15.' */
	set( 0x2497, "16."); /* '⒗' -> '16.' */
	set( 0x2498, "17."); /* '⒘' -> '17.' */
	set( 0x2499, "18."); /* '⒙' -> '18.' */
	set( 0x249a, "19."); /* '⒚' -> '19.' */
	set( 0x249b, "20."); /* '⒛' -> '20.' */
	set( 0x249c, "(a)"); /* '⒜' -> '(a)' */
	set( 0x249d, "(b)"); /* '⒝' -> '(b)' */
	set( 0x249e, "(c)"); /* '⒞' -> '(c)' */
	set( 0x249f, "(d)"); /* '⒟' -> '(d)' */
	set( 0x24a0, "(e)"); /* '⒠' -> '(e)' */
	set( 0x24a1, "(f)"); /* '⒡' -> '(f)' */
	set( 0x24a2, "(g)"); /* '⒢' -> '(g)' */
	set( 0x24a3, "(h)"); /* '⒣' -> '(h)' */
	set( 0x24a4, "(i)"); /* '⒤' -> '(i)' */
	set( 0x24a5, "(j)"); /* '⒥' -> '(j)' */
	set( 0x24a6, "(k)"); /* '⒦' -> '(k)' */
	set( 0x24a7, "(l)"); /* '⒧' -> '(l)' */
	set( 0x24a8, "(m)"); /* '⒨' -> '(m)' */
	set( 0x24a9, "(n)"); /* '⒩' -> '(n)' */
	set( 0x24aa, "(o)"); /* '⒪' -> '(o)' */
	set( 0x24ab, "(p)"); /* '⒫' -> '(p)' */
	set( 0x24ac, "(q)"); /* '⒬' -> '(q)' */
	set( 0x24ad, "(r)"); /* '⒭' -> '(r)' */
	set( 0x24ae, "(s)"); /* '⒮' -> '(s)' */
	set( 0x24af, "(t)"); /* '⒯' -> '(t)' */
	set( 0x24b0, "(u)"); /* '⒰' -> '(u)' */
	set( 0x24b1, "(v)"); /* '⒱' -> '(v)' */
	set( 0x24b2, "(w)"); /* '⒲' -> '(w)' */
	set( 0x24b3, "(x)"); /* '⒳' -> '(x)' */
	set( 0x24b4, "(y)"); /* '⒴' -> '(y)' */
	set( 0x24b5, "(z)"); /* '⒵' -> '(z)' */
	set( 0x24b6, "A"); /* 'Ⓐ' -> 'A' */
	set( 0x24b7, "B"); /* 'Ⓑ' -> 'B' */
	set( 0x24b8, "C"); /* 'Ⓒ' -> 'C' */
	set( 0x24b9, "D"); /* 'Ⓓ' -> 'D' */
	set( 0x24ba, "E"); /* 'Ⓔ' -> 'E' */
	set( 0x24bb, "F"); /* 'Ⓕ' -> 'F' */
	set( 0x24bc, "G"); /* 'Ⓖ' -> 'G' */
	set( 0x24bd, "H"); /* 'Ⓗ' -> 'H' */
	set( 0x24be, "I"); /* 'Ⓘ' -> 'I' */
	set( 0x24bf, "J"); /* 'Ⓙ' -> 'J' */
	set( 0x24c0, "K"); /* 'Ⓚ' -> 'K' */
	set( 0x24c1, "L"); /* 'Ⓛ' -> 'L' */
	set( 0x24c2, "M"); /* 'Ⓜ' -> 'M' */
	set( 0x24c3, "N"); /* 'Ⓝ' -> 'N' */
	set( 0x24c4, "O"); /* 'Ⓞ' -> 'O' */
	set( 0x24c5, "P"); /* 'Ⓟ' -> 'P' */
	set( 0x24c6, "Q"); /* 'Ⓠ' -> 'Q' */
	set( 0x24c7, "R"); /* 'Ⓡ' -> 'R' */
	set( 0x24c8, "S"); /* 'Ⓢ' -> 'S' */
	set( 0x24c9, "T"); /* 'Ⓣ' -> 'T' */
	set( 0x24ca, "U"); /* 'Ⓤ' -> 'U' */
	set( 0x24cb, "V"); /* 'Ⓥ' -> 'V' */
	set( 0x24cc, "W"); /* 'Ⓦ' -> 'W' */
	set( 0x24cd, "X"); /* 'Ⓧ' -> 'X' */
	set( 0x24ce, "Y"); /* 'Ⓨ' -> 'Y' */
	set( 0x24cf, "Z"); /* 'Ⓩ' -> 'Z' */
	set( 0x24d0, "a"); /* 'ⓐ' -> 'a' */
	set( 0x24d1, "b"); /* 'ⓑ' -> 'b' */
	set( 0x24d2, "c"); /* 'ⓒ' -> 'c' */
	set( 0x24d3, "d"); /* 'ⓓ' -> 'd' */
	set( 0x24d4, "e"); /* 'ⓔ' -> 'e' */
	set( 0x24d5, "f"); /* 'ⓕ' -> 'f' */
	set( 0x24d6, "g"); /* 'ⓖ' -> 'g' */
	set( 0x24d7, "h"); /* 'ⓗ' -> 'h' */
	set( 0x24d8, "i"); /* 'ⓘ' -> 'i' */
	set( 0x24d9, "j"); /* 'ⓙ' -> 'j' */
	set( 0x24da, "k"); /* 'ⓚ' -> 'k' */
	set( 0x24db, "l"); /* 'ⓛ' -> 'l' */
	set( 0x24dc, "m"); /* 'ⓜ' -> 'm' */
	set( 0x24dd, "n"); /* 'ⓝ' -> 'n' */
	set( 0x24de, "o"); /* 'ⓞ' -> 'o' */
	set( 0x24df, "p"); /* 'ⓟ' -> 'p' */
	set( 0x24e0, "q"); /* 'ⓠ' -> 'q' */
	set( 0x24e1, "r"); /* 'ⓡ' -> 'r' */
	set( 0x24e2, "s"); /* 'ⓢ' -> 's' */
	set( 0x24e3, "t"); /* 'ⓣ' -> 't' */
	set( 0x24e4, "u"); /* 'ⓤ' -> 'u' */
	set( 0x24e5, "v"); /* 'ⓥ' -> 'v' */
	set( 0x24e6, "w"); /* 'ⓦ' -> 'w' */
	set( 0x24e7, "x"); /* 'ⓧ' -> 'x' */
	set( 0x24e8, "y"); /* 'ⓨ' -> 'y' */
	set( 0x24e9, "z"); /* 'ⓩ' -> 'z' */
	set( 0x24ea, "0"); /* '⓪' -> '0' */
	set( 0x2a74, "::="); /* '⩴' -> '::=' */
	set( 0x2a75, "=="); /* '⩵' -> '==' */
	set( 0x2a76, "==="); /* '⩶' -> '===' */
	set( 0x2c7c, "j"); /* 'ⱼ' -> 'j' */
	set( 0x2c7d, "V"); /* 'ⱽ' -> 'V' */
	set( 0x3000, " "); /* '　' -> ' ' */
	set( 0x309b, " "); /* '゛' -> ' ' */
	set( 0x309c, " "); /* '゜' -> ' ' */
	set( 0x3250, "PTE"); /* '㉐' -> 'PTE' */
	set( 0x3251, "21"); /* '㉑' -> '21' */
	set( 0x3252, "22"); /* '㉒' -> '22' */
	set( 0x3253, "23"); /* '㉓' -> '23' */
	set( 0x3254, "24"); /* '㉔' -> '24' */
	set( 0x3255, "25"); /* '㉕' -> '25' */
	set( 0x3256, "26"); /* '㉖' -> '26' */
	set( 0x3257, "27"); /* '㉗' -> '27' */
	set( 0x3258, "28"); /* '㉘' -> '28' */
	set( 0x3259, "29"); /* '㉙' -> '29' */
	set( 0x325a, "30"); /* '㉚' -> '30' */
	set( 0x325b, "31"); /* '㉛' -> '31' */
	set( 0x325c, "32"); /* '㉜' -> '32' */
	set( 0x325d, "33"); /* '㉝' -> '33' */
	set( 0x325e, "34"); /* '㉞' -> '34' */
	set( 0x325f, "35"); /* '㉟' -> '35' */
	set( 0x32b1, "36"); /* '㊱' -> '36' */
	set( 0x32b2, "37"); /* '㊲' -> '37' */
	set( 0x32b3, "38"); /* '㊳' -> '38' */
	set( 0x32b4, "39"); /* '㊴' -> '39' */
	set( 0x32b5, "40"); /* '㊵' -> '40' */
	set( 0x32b6, "41"); /* '㊶' -> '41' */
	set( 0x32b7, "42"); /* '㊷' -> '42' */
	set( 0x32b8, "43"); /* '㊸' -> '43' */
	set( 0x32b9, "44"); /* '㊹' -> '44' */
	set( 0x32ba, "45"); /* '㊺' -> '45' */
	set( 0x32bb, "46"); /* '㊻' -> '46' */
	set( 0x32bc, "47"); /* '㊼' -> '47' */
	set( 0x32bd, "48"); /* '㊽' -> '48' */
	set( 0x32be, "49"); /* '㊾' -> '49' */
	set( 0x32bf, "50"); /* '㊿' -> '50' */
	set( 0x32c0, "1"); /* '㋀' -> '1' */
	set( 0x32c1, "2"); /* '㋁' -> '2' */
	set( 0x32c2, "3"); /* '㋂' -> '3' */
	set( 0x32c3, "4"); /* '㋃' -> '4' */
	set( 0x32c4, "5"); /* '㋄' -> '5' */
	set( 0x32c5, "6"); /* '㋅' -> '6' */
	set( 0x32c6, "7"); /* '㋆' -> '7' */
	set( 0x32c7, "8"); /* '㋇' -> '8' */
	set( 0x32c8, "9"); /* '㋈' -> '9' */
	set( 0x32c9, "10"); /* '㋉' -> '10' */
	set( 0x32ca, "11"); /* '㋊' -> '11' */
	set( 0x32cb, "12"); /* '㋋' -> '12' */
	set( 0x32cc, "Hg"); /* '㋌' -> 'Hg' */
	set( 0x32cd, "erg"); /* '㋍' -> 'erg' */
	set( 0x32ce, "eV"); /* '㋎' -> 'eV' */
	set( 0x32cf, "LTD"); /* '㋏' -> 'LTD' */
	set( 0x3358, "0"); /* '㍘' -> '0' */
	set( 0x3359, "1"); /* '㍙' -> '1' */
	set( 0x335a, "2"); /* '㍚' -> '2' */
	set( 0x335b, "3"); /* '㍛' -> '3' */
	set( 0x335c, "4"); /* '㍜' -> '4' */
	set( 0x335d, "5"); /* '㍝' -> '5' */
	set( 0x335e, "6"); /* '㍞' -> '6' */
	set( 0x335f, "7"); /* '㍟' -> '7' */
	set( 0x3360, "8"); /* '㍠' -> '8' */
	set( 0x3361, "9"); /* '㍡' -> '9' */
	set( 0x3362, "10"); /* '㍢' -> '10' */
	set( 0x3363, "11"); /* '㍣' -> '11' */
	set( 0x3364, "12"); /* '㍤' -> '12' */
	set( 0x3365, "13"); /* '㍥' -> '13' */
	set( 0x3366, "14"); /* '㍦' -> '14' */
	set( 0x3367, "15"); /* '㍧' -> '15' */
	set( 0x3368, "16"); /* '㍨' -> '16' */
	set( 0x3369, "17"); /* '㍩' -> '17' */
	set( 0x336a, "18"); /* '㍪' -> '18' */
	set( 0x336b, "19"); /* '㍫' -> '19' */
	set( 0x336c, "20"); /* '㍬' -> '20' */
	set( 0x336d, "21"); /* '㍭' -> '21' */
	set( 0x336e, "22"); /* '㍮' -> '22' */
	set( 0x336f, "23"); /* '㍯' -> '23' */
	set( 0x3370, "24"); /* '㍰' -> '24' */
	set( 0x3371, "hPa"); /* '㍱' -> 'hPa' */
	set( 0x3372, "da"); /* '㍲' -> 'da' */
	set( 0x3373, "AU"); /* '㍳' -> 'AU' */
	set( 0x3374, "bar"); /* '㍴' -> 'bar' */
	set( 0x3375, "oV"); /* '㍵' -> 'oV' */
	set( 0x3376, "pc"); /* '㍶' -> 'pc' */
	set( 0x3377, "dm"); /* '㍷' -> 'dm' */
	set( 0x3378, "dm2"); /* '㍸' -> 'dm2' */
	set( 0x3379, "dm3"); /* '㍹' -> 'dm3' */
	set( 0x337a, "IU"); /* '㍺' -> 'IU' */
	set( 0x3380, "pA"); /* '㎀' -> 'pA' */
	set( 0x3381, "nA"); /* '㎁' -> 'nA' */
	set( 0x3382, "A"); /* '㎂' -> 'A' */
	set( 0x3383, "mA"); /* '㎃' -> 'mA' */
	set( 0x3384, "kA"); /* '㎄' -> 'kA' */
	set( 0x3385, "KB"); /* '㎅' -> 'KB' */
	set( 0x3386, "MB"); /* '㎆' -> 'MB' */
	set( 0x3387, "GB"); /* '㎇' -> 'GB' */
	set( 0x3388, "cal"); /* '㎈' -> 'cal' */
	set( 0x3389, "kcal"); /* '㎉' -> 'kcal' */
	set( 0x338a, "pF"); /* '㎊' -> 'pF' */
	set( 0x338b, "nF"); /* '㎋' -> 'nF' */
	set( 0x338c, "F"); /* '㎌' -> 'F' */
	set( 0x338d, "g"); /* '㎍' -> 'g' */
	set( 0x338e, "mg"); /* '㎎' -> 'mg' */
	set( 0x338f, "kg"); /* '㎏' -> 'kg' */
	set( 0x3390, "Hz"); /* '㎐' -> 'Hz' */
	set( 0x3391, "kHz"); /* '㎑' -> 'kHz' */
	set( 0x3392, "MHz"); /* '㎒' -> 'MHz' */
	set( 0x3393, "GHz"); /* '㎓' -> 'GHz' */
	set( 0x3394, "THz"); /* '㎔' -> 'THz' */
	set( 0x3395, "l"); /* '㎕' -> 'l' */
	set( 0x3396, "ml"); /* '㎖' -> 'ml' */
	set( 0x3397, "dl"); /* '㎗' -> 'dl' */
	set( 0x3398, "kl"); /* '㎘' -> 'kl' */
	set( 0x3399, "fm"); /* '㎙' -> 'fm' */
	set( 0x339a, "nm"); /* '㎚' -> 'nm' */
	set( 0x339b, "m"); /* '㎛' -> 'm' */
	set( 0x339c, "mm"); /* '㎜' -> 'mm' */
	set( 0x339d, "cm"); /* '㎝' -> 'cm' */
	set( 0x339e, "km"); /* '㎞' -> 'km' */
	set( 0x339f, "mm2"); /* '㎟' -> 'mm2' */
	set( 0x33a0, "cm2"); /* '㎠' -> 'cm2' */
	set( 0x33a1, "m2"); /* '㎡' -> 'm2' */
	set( 0x33a2, "km2"); /* '㎢' -> 'km2' */
	set( 0x33a3, "mm3"); /* '㎣' -> 'mm3' */
	set( 0x33a4, "cm3"); /* '㎤' -> 'cm3' */
	set( 0x33a5, "m3"); /* '㎥' -> 'm3' */
	set( 0x33a6, "km3"); /* '㎦' -> 'km3' */
	set( 0x33a7, "ms"); /* '㎧' -> 'ms' */
	set( 0x33a8, "ms2"); /* '㎨' -> 'ms2' */
	set( 0x33a9, "Pa"); /* '㎩' -> 'Pa' */
	set( 0x33aa, "kPa"); /* '㎪' -> 'kPa' */
	set( 0x33ab, "MPa"); /* '㎫' -> 'MPa' */
	set( 0x33ac, "GPa"); /* '㎬' -> 'GPa' */
	set( 0x33ad, "rad"); /* '㎭' -> 'rad' */
	set( 0x33ae, "rads"); /* '㎮' -> 'rads' */
	set( 0x33af, "rads2"); /* '㎯' -> 'rads2' */
	set( 0x33b0, "ps"); /* '㎰' -> 'ps' */
	set( 0x33b1, "ns"); /* '㎱' -> 'ns' */
	set( 0x33b2, "s"); /* '㎲' -> 's' */
	set( 0x33b3, "ms"); /* '㎳' -> 'ms' */
	set( 0x33b4, "pV"); /* '㎴' -> 'pV' */
	set( 0x33b5, "nV"); /* '㎵' -> 'nV' */
	set( 0x33b6, "V"); /* '㎶' -> 'V' */
	set( 0x33b7, "mV"); /* '㎷' -> 'mV' */
	set( 0x33b8, "kV"); /* '㎸' -> 'kV' */
	set( 0x33b9, "MV"); /* '㎹' -> 'MV' */
	set( 0x33ba, "pW"); /* '㎺' -> 'pW' */
	set( 0x33bb, "nW"); /* '㎻' -> 'nW' */
	set( 0x33bc, "W"); /* '㎼' -> 'W' */
	set( 0x33bd, "mW"); /* '㎽' -> 'mW' */
	set( 0x33be, "kW"); /* '㎾' -> 'kW' */
	set( 0x33bf, "MW"); /* '㎿' -> 'MW' */
	set( 0x33c0, "k"); /* '㏀' -> 'k' */
	set( 0x33c1, "M"); /* '㏁' -> 'M' */
	set( 0x33c2, "a.m."); /* '㏂' -> 'a.m.' */
	set( 0x33c3, "Bq"); /* '㏃' -> 'Bq' */
	set( 0x33c4, "cc"); /* '㏄' -> 'cc' */
	set( 0x33c5, "cd"); /* '㏅' -> 'cd' */
	set( 0x33c6, "Ckg"); /* '㏆' -> 'Ckg' */
	set( 0x33c7, "Co."); /* '㏇' -> 'Co.' */
	set( 0x33c8, "dB"); /* '㏈' -> 'dB' */
	set( 0x33c9, "Gy"); /* '㏉' -> 'Gy' */
	set( 0x33ca, "ha"); /* '㏊' -> 'ha' */
	set( 0x33cb, "HP"); /* '㏋' -> 'HP' */
	set( 0x33cc, "in"); /* '㏌' -> 'in' */
	set( 0x33cd, "KK"); /* '㏍' -> 'KK' */
	set( 0x33ce, "KM"); /* '㏎' -> 'KM' */
	set( 0x33cf, "kt"); /* '㏏' -> 'kt' */
	set( 0x33d0, "lm"); /* '㏐' -> 'lm' */
	set( 0x33d1, "ln"); /* '㏑' -> 'ln' */
	set( 0x33d2, "log"); /* '㏒' -> 'log' */
	set( 0x33d3, "lx"); /* '㏓' -> 'lx' */
	set( 0x33d4, "mb"); /* '㏔' -> 'mb' */
	set( 0x33d5, "mil"); /* '㏕' -> 'mil' */
	set( 0x33d6, "mol"); /* '㏖' -> 'mol' */
	set( 0x33d7, "PH"); /* '㏗' -> 'PH' */
	set( 0x33d8, "p.m."); /* '㏘' -> 'p.m.' */
	set( 0x33d9, "PPM"); /* '㏙' -> 'PPM' */
	set( 0x33da, "PR"); /* '㏚' -> 'PR' */
	set( 0x33db, "sr"); /* '㏛' -> 'sr' */
	set( 0x33dc, "Sv"); /* '㏜' -> 'Sv' */
	set( 0x33dd, "Wb"); /* '㏝' -> 'Wb' */
	set( 0x33de, "Vm"); /* '㏞' -> 'Vm' */
	set( 0x33df, "Am"); /* '㏟' -> 'Am' */
	set( 0x33e0, "1"); /* '㏠' -> '1' */
	set( 0x33e1, "2"); /* '㏡' -> '2' */
	set( 0x33e2, "3"); /* '㏢' -> '3' */
	set( 0x33e3, "4"); /* '㏣' -> '4' */
	set( 0x33e4, "5"); /* '㏤' -> '5' */
	set( 0x33e5, "6"); /* '㏥' -> '6' */
	set( 0x33e6, "7"); /* '㏦' -> '7' */
	set( 0x33e7, "8"); /* '㏧' -> '8' */
	set( 0x33e8, "9"); /* '㏨' -> '9' */
	set( 0x33e9, "10"); /* '㏩' -> '10' */
	set( 0x33ea, "11"); /* '㏪' -> '11' */
	set( 0x33eb, "12"); /* '㏫' -> '12' */
	set( 0x33ec, "13"); /* '㏬' -> '13' */
	set( 0x33ed, "14"); /* '㏭' -> '14' */
	set( 0x33ee, "15"); /* '㏮' -> '15' */
	set( 0x33ef, "16"); /* '㏯' -> '16' */
	set( 0x33f0, "17"); /* '㏰' -> '17' */
	set( 0x33f1, "18"); /* '㏱' -> '18' */
	set( 0x33f2, "19"); /* '㏲' -> '19' */
	set( 0x33f3, "20"); /* '㏳' -> '20' */
	set( 0x33f4, "21"); /* '㏴' -> '21' */
	set( 0x33f5, "22"); /* '㏵' -> '22' */
	set( 0x33f6, "23"); /* '㏶' -> '23' */
	set( 0x33f7, "24"); /* '㏷' -> '24' */
	set( 0x33f8, "25"); /* '㏸' -> '25' */
	set( 0x33f9, "26"); /* '㏹' -> '26' */
	set( 0x33fa, "27"); /* '㏺' -> '27' */
	set( 0x33fb, "28"); /* '㏻' -> '28' */
	set( 0x33fc, "29"); /* '㏼' -> '29' */
	set( 0x33fd, "30"); /* '㏽' -> '30' */
	set( 0x33fe, "31"); /* '㏾' -> '31' */
	set( 0x33ff, "gal"); /* '㏿' -> 'gal' */
	set( 0xfb00, "ff"); /* 'ﬀ' -> 'ff' */
	set( 0xfb01, "fi"); /* 'ﬁ' -> 'fi' */
	set( 0xfb02, "fl"); /* 'ﬂ' -> 'fl' */
	set( 0xfb03, "ffi"); /* 'ﬃ' -> 'ffi' */
	set( 0xfb04, "ffl"); /* 'ﬄ' -> 'ffl' */
	set( 0xfb05, "st"); /* 'ﬅ' -> 'st' */
	set( 0xfb06, "st"); /* 'ﬆ' -> 'st' */
	set( 0xfb29, "+"); /* '﬩' -> '+' */
	set( 0xfc5e, " "); /* 'ﱞ' -> ' ' */
	set( 0xfc5f, " "); /* 'ﱟ' -> ' ' */
	set( 0xfc60, " "); /* 'ﱠ' -> ' ' */
	set( 0xfc61, " "); /* 'ﱡ' -> ' ' */
	set( 0xfc62, " "); /* 'ﱢ' -> ' ' */
	set( 0xfc63, " "); /* 'ﱣ' -> ' ' */
	set( 0xfdfa, "   "); /* 'ﷺ' -> '   ' */
	set( 0xfdfb, " "); /* 'ﷻ' -> ' ' */
	set( 0xfe10, ","); /* '︐' -> ',' */
	set( 0xfe13, ":"); /* '︓' -> ':' */
	set( 0xfe14, ";"); /* '︔' -> ';' */
	set( 0xfe15, "!"); /* '︕' -> '!' */
	set( 0xfe16, "?"); /* '︖' -> '?' */
	set( 0xfe19, "..."); /* '︙' -> '...' */
	set( 0xfe30, ".."); /* '︰' -> '..' */
	set( 0xfe33, "_"); /* '︳' -> '_' */
	set( 0xfe34, "_"); /* '︴' -> '_' */
	set( 0xfe35, "("); /* '︵' -> '(' */
	set( 0xfe36, ")"); /* '︶' -> ')' */
	set( 0xfe37, "{"); /* '︷' -> '{' */
	set( 0xfe38, "}"); /* '︸' -> '}' */
	set( 0xfe47, "["); /* '﹇' -> '[' */
	set( 0xfe48, "]"); /* '﹈' -> ']' */
	set( 0xfe49, " "); /* '﹉' -> ' ' */
	set( 0xfe4a, " "); /* '﹊' -> ' ' */
	set( 0xfe4b, " "); /* '﹋' -> ' ' */
	set( 0xfe4c, " "); /* '﹌' -> ' ' */
	set( 0xfe4d, "_"); /* '﹍' -> '_' */
	set( 0xfe4e, "_"); /* '﹎' -> '_' */
	set( 0xfe4f, "_"); /* '﹏' -> '_' */
	set( 0xfe50, ","); /* '﹐' -> ',' */
	set( 0xfe52, "."); /* '﹒' -> '.' */
	set( 0xfe54, ";"); /* '﹔' -> ';' */
	set( 0xfe55, ":"); /* '﹕' -> ':' */
	set( 0xfe56, "?"); /* '﹖' -> '?' */
	set( 0xfe57, "!"); /* '﹗' -> '!' */
	set( 0xfe59, "("); /* '﹙' -> '(' */
	set( 0xfe5a, ")"); /* '﹚' -> ')' */
	set( 0xfe5b, "{"); /* '﹛' -> '{' */
	set( 0xfe5c, "}"); /* '﹜' -> '}' */
	set( 0xfe5f, "#"); /* '﹟' -> '#' */
	set( 0xfe60, "&"); /* '﹠' -> '&' */
	set( 0xfe61, "*"); /* '﹡' -> '*' */
	set( 0xfe62, "+"); /* '﹢' -> '+' */
	set( 0xfe63, "-"); /* '﹣' -> '-' */
	set( 0xfe64, "<"); /* '﹤' -> '<' */
	set( 0xfe65, ">"); /* '﹥' -> '>' */
	set( 0xfe66, "="); /* '﹦' -> '=' */
	set( 0xfe68, "\\"); /* '﹨' -> '\' */
	set( 0xfe69, "$"); /* '﹩' -> '$' */
	set( 0xfe6a, "%"); /* '﹪' -> '%' */
	set( 0xfe6b, "@"); /* '﹫' -> '@' */
	set( 0xfe70, " "); /* 'ﹰ' -> ' ' */
	set( 0xfe72, " "); /* 'ﹲ' -> ' ' */
	set( 0xfe74, " "); /* 'ﹴ' -> ' ' */
	set( 0xfe76, " "); /* 'ﹶ' -> ' ' */
	set( 0xfe78, " "); /* 'ﹸ' -> ' ' */
	set( 0xfe7a, " "); /* 'ﹺ' -> ' ' */
	set( 0xfe7c, " "); /* 'ﹼ' -> ' ' */
	set( 0xfe7e, " "); /* 'ﹾ' -> ' ' */
	set( 0xff01, "!"); /* '！' -> '!' */
	set( 0xff02, "\""); /* '＂' -> '"' */
	set( 0xff03, "#"); /* '＃' -> '#' */
	set( 0xff04, "$"); /* '＄' -> '$' */
	set( 0xff05, "%"); /* '％' -> '%' */
	set( 0xff06, "&"); /* '＆' -> '&' */
	set( 0xff07, "'"); /* '＇' -> ''' */
	set( 0xff08, "("); /* '（' -> '(' */
	set( 0xff09, ")"); /* '）' -> ')' */
	set( 0xff0a, "*"); /* '＊' -> '*' */
	set( 0xff0b, "+"); /* '＋' -> '+' */
	set( 0xff0c, ","); /* '，' -> ',' */
	set( 0xff0d, "-"); /* '－' -> '-' */
	set( 0xff0e, "."); /* '．' -> '.' */
	set( 0xff0f, "/"); /* '／' -> '/' */
	set( 0xff10, "0"); /* '０' -> '0' */
	set( 0xff11, "1"); /* '１' -> '1' */
	set( 0xff12, "2"); /* '２' -> '2' */
	set( 0xff13, "3"); /* '３' -> '3' */
	set( 0xff14, "4"); /* '４' -> '4' */
	set( 0xff15, "5"); /* '５' -> '5' */
	set( 0xff16, "6"); /* '６' -> '6' */
	set( 0xff17, "7"); /* '７' -> '7' */
	set( 0xff18, "8"); /* '８' -> '8' */
	set( 0xff19, "9"); /* '９' -> '9' */
	set( 0xff1a, ":"); /* '：' -> ':' */
	set( 0xff1b, ";"); /* '；' -> ';' */
	set( 0xff1c, "<"); /* '＜' -> '<' */
	set( 0xff1d, "="); /* '＝' -> '=' */
	set( 0xff1e, ">"); /* '＞' -> '>' */
	set( 0xff1f, "?"); /* '？' -> '?' */
	set( 0xff20, "@"); /* '＠' -> '@' */
	set( 0xff21, "A"); /* 'Ａ' -> 'A' */
	set( 0xff22, "B"); /* 'Ｂ' -> 'B' */
	set( 0xff23, "C"); /* 'Ｃ' -> 'C' */
	set( 0xff24, "D"); /* 'Ｄ' -> 'D' */
	set( 0xff25, "E"); /* 'Ｅ' -> 'E' */
	set( 0xff26, "F"); /* 'Ｆ' -> 'F' */
	set( 0xff27, "G"); /* 'Ｇ' -> 'G' */
	set( 0xff28, "H"); /* 'Ｈ' -> 'H' */
	set( 0xff29, "I"); /* 'Ｉ' -> 'I' */
	set( 0xff2a, "J"); /* 'Ｊ' -> 'J' */
	set( 0xff2b, "K"); /* 'Ｋ' -> 'K' */
	set( 0xff2c, "L"); /* 'Ｌ' -> 'L' */
	set( 0xff2d, "M"); /* 'Ｍ' -> 'M' */
	set( 0xff2e, "N"); /* 'Ｎ' -> 'N' */
	set( 0xff2f, "O"); /* 'Ｏ' -> 'O' */
	set( 0xff30, "P"); /* 'Ｐ' -> 'P' */
	set( 0xff31, "Q"); /* 'Ｑ' -> 'Q' */
	set( 0xff32, "R"); /* 'Ｒ' -> 'R' */
	set( 0xff33, "S"); /* 'Ｓ' -> 'S' */
	set( 0xff34, "T"); /* 'Ｔ' -> 'T' */
	set( 0xff35, "U"); /* 'Ｕ' -> 'U' */
	set( 0xff36, "V"); /* 'Ｖ' -> 'V' */
	set( 0xff37, "W"); /* 'Ｗ' -> 'W' */
	set( 0xff38, "X"); /* 'Ｘ' -> 'X' */
	set( 0xff39, "Y"); /* 'Ｙ' -> 'Y' */
	set( 0xff3a, "Z"); /* 'Ｚ' -> 'Z' */
	set( 0xff3b, "["); /* '［' -> '[' */
	set( 0xff3c, "\\"); /* '＼' -> '\' */
	set( 0xff3d, "]"); /* '］' -> ']' */
	set( 0xff3e, "^"); /* '＾' -> '^' */
	set( 0xff3f, "_"); /* '＿' -> '_' */
	set( 0xff40, "`"); /* '｀' -> '`' */
	set( 0xff41, "a"); /* 'ａ' -> 'a' */
	set( 0xff42, "b"); /* 'ｂ' -> 'b' */
	set( 0xff43, "c"); /* 'ｃ' -> 'c' */
	set( 0xff44, "d"); /* 'ｄ' -> 'd' */
	set( 0xff45, "e"); /* 'ｅ' -> 'e' */
	set( 0xff46, "f"); /* 'ｆ' -> 'f' */
	set( 0xff47, "g"); /* 'ｇ' -> 'g' */
	set( 0xff48, "h"); /* 'ｈ' -> 'h' */
	set( 0xff49, "i"); /* 'ｉ' -> 'i' */
	set( 0xff4a, "j"); /* 'ｊ' -> 'j' */
	set( 0xff4b, "k"); /* 'ｋ' -> 'k' */
	set( 0xff4c, "l"); /* 'ｌ' -> 'l' */
	set( 0xff4d, "m"); /* 'ｍ' -> 'm' */
	set( 0xff4e, "n"); /* 'ｎ' -> 'n' */
	set( 0xff4f, "o"); /* 'ｏ' -> 'o' */
	set( 0xff50, "p"); /* 'ｐ' -> 'p' */
	set( 0xff51, "q"); /* 'ｑ' -> 'q' */
	set( 0xff52, "r"); /* 'ｒ' -> 'r' */
	set( 0xff53, "s"); /* 'ｓ' -> 's' */
	set( 0xff54, "t"); /* 'ｔ' -> 't' */
	set( 0xff55, "u"); /* 'ｕ' -> 'u' */
	set( 0xff56, "v"); /* 'ｖ' -> 'v' */
	set( 0xff57, "w"); /* 'ｗ' -> 'w' */
	set( 0xff58, "x"); /* 'ｘ' -> 'x' */
	set( 0xff59, "y"); /* 'ｙ' -> 'y' */
	set( 0xff5a, "z"); /* 'ｚ' -> 'z' */
	set( 0xff5b, "{"); /* '｛' -> '{' */
	set( 0xff5c, "|"); /* '｜' -> '|' */
	set( 0xff5d, "}"); /* '｝' -> '}' */
	set( 0xff5e, "~"); /* '～' -> '~' */
	set( 0xffe3, " "); /* '￣' -> ' ' */
	set( 0x1d400, "A"); /* '𝐀' -> 'A' */
	set( 0x1d401, "B"); /* '𝐁' -> 'B' */
	set( 0x1d402, "C"); /* '𝐂' -> 'C' */
	set( 0x1d403, "D"); /* '𝐃' -> 'D' */
	set( 0x1d404, "E"); /* '𝐄' -> 'E' */
	set( 0x1d405, "F"); /* '𝐅' -> 'F' */
	set( 0x1d406, "G"); /* '𝐆' -> 'G' */
	set( 0x1d407, "H"); /* '𝐇' -> 'H' */
	set( 0x1d408, "I"); /* '𝐈' -> 'I' */
	set( 0x1d409, "J"); /* '𝐉' -> 'J' */
	set( 0x1d40a, "K"); /* '𝐊' -> 'K' */
	set( 0x1d40b, "L"); /* '𝐋' -> 'L' */
	set( 0x1d40c, "M"); /* '𝐌' -> 'M' */
	set( 0x1d40d, "N"); /* '𝐍' -> 'N' */
	set( 0x1d40e, "O"); /* '𝐎' -> 'O' */
	set( 0x1d40f, "P"); /* '𝐏' -> 'P' */
	set( 0x1d410, "Q"); /* '𝐐' -> 'Q' */
	set( 0x1d411, "R"); /* '𝐑' -> 'R' */
	set( 0x1d412, "S"); /* '𝐒' -> 'S' */
	set( 0x1d413, "T"); /* '𝐓' -> 'T' */
	set( 0x1d414, "U"); /* '𝐔' -> 'U' */
	set( 0x1d415, "V"); /* '𝐕' -> 'V' */
	set( 0x1d416, "W"); /* '𝐖' -> 'W' */
	set( 0x1d417, "X"); /* '𝐗' -> 'X' */
	set( 0x1d418, "Y"); /* '𝐘' -> 'Y' */
	set( 0x1d419, "Z"); /* '𝐙' -> 'Z' */
	set( 0x1d41a, "a"); /* '𝐚' -> 'a' */
	set( 0x1d41b, "b"); /* '𝐛' -> 'b' */
	set( 0x1d41c, "c"); /* '𝐜' -> 'c' */
	set( 0x1d41d, "d"); /* '𝐝' -> 'd' */
	set( 0x1d41e, "e"); /* '𝐞' -> 'e' */
	set( 0x1d41f, "f"); /* '𝐟' -> 'f' */
	set( 0x1d420, "g"); /* '𝐠' -> 'g' */
	set( 0x1d421, "h"); /* '𝐡' -> 'h' */
	set( 0x1d422, "i"); /* '𝐢' -> 'i' */
	set( 0x1d423, "j"); /* '𝐣' -> 'j' */
	set( 0x1d424, "k"); /* '𝐤' -> 'k' */
	set( 0x1d425, "l"); /* '𝐥' -> 'l' */
	set( 0x1d426, "m"); /* '𝐦' -> 'm' */
	set( 0x1d427, "n"); /* '𝐧' -> 'n' */
	set( 0x1d428, "o"); /* '𝐨' -> 'o' */
	set( 0x1d429, "p"); /* '𝐩' -> 'p' */
	set( 0x1d42a, "q"); /* '𝐪' -> 'q' */
	set( 0x1d42b, "r"); /* '𝐫' -> 'r' */
	set( 0x1d42c, "s"); /* '𝐬' -> 's' */
	set( 0x1d42d, "t"); /* '𝐭' -> 't' */
	set( 0x1d42e, "u"); /* '𝐮' -> 'u' */
	set( 0x1d42f, "v"); /* '𝐯' -> 'v' */
	set( 0x1d430, "w"); /* '𝐰' -> 'w' */
	set( 0x1d431, "x"); /* '𝐱' -> 'x' */
	set( 0x1d432, "y"); /* '𝐲' -> 'y' */
	set( 0x1d433, "z"); /* '𝐳' -> 'z' */
	set( 0x1d434, "A"); /* '𝐴' -> 'A' */
	set( 0x1d435, "B"); /* '𝐵' -> 'B' */
	set( 0x1d436, "C"); /* '𝐶' -> 'C' */
	set( 0x1d437, "D"); /* '𝐷' -> 'D' */
	set( 0x1d438, "E"); /* '𝐸' -> 'E' */
	set( 0x1d439, "F"); /* '𝐹' -> 'F' */
	set( 0x1d43a, "G"); /* '𝐺' -> 'G' */
	set( 0x1d43b, "H"); /* '𝐻' -> 'H' */
	set( 0x1d43c, "I"); /* '𝐼' -> 'I' */
	set( 0x1d43d, "J"); /* '𝐽' -> 'J' */
	set( 0x1d43e, "K"); /* '𝐾' -> 'K' */
	set( 0x1d43f, "L"); /* '𝐿' -> 'L' */
	set( 0x1d440, "M"); /* '𝑀' -> 'M' */
	set( 0x1d441, "N"); /* '𝑁' -> 'N' */
	set( 0x1d442, "O"); /* '𝑂' -> 'O' */
	set( 0x1d443, "P"); /* '𝑃' -> 'P' */
	set( 0x1d444, "Q"); /* '𝑄' -> 'Q' */
	set( 0x1d445, "R"); /* '𝑅' -> 'R' */
	set( 0x1d446, "S"); /* '𝑆' -> 'S' */
	set( 0x1d447, "T"); /* '𝑇' -> 'T' */
	set( 0x1d448, "U"); /* '𝑈' -> 'U' */
	set( 0x1d449, "V"); /* '𝑉' -> 'V' */
	set( 0x1d44a, "W"); /* '𝑊' -> 'W' */
	set( 0x1d44b, "X"); /* '𝑋' -> 'X' */
	set( 0x1d44c, "Y"); /* '𝑌' -> 'Y' */
	set( 0x1d44d, "Z"); /* '𝑍' -> 'Z' */
	set( 0x1d44e, "a"); /* '𝑎' -> 'a' */
	set( 0x1d44f, "b"); /* '𝑏' -> 'b' */
	set( 0x1d450, "c"); /* '𝑐' -> 'c' */
	set( 0x1d451, "d"); /* '𝑑' -> 'd' */
	set( 0x1d452, "e"); /* '𝑒' -> 'e' */
	set( 0x1d453, "f"); /* '𝑓' -> 'f' */
	set( 0x1d454, "g"); /* '𝑔' -> 'g' */
	set( 0x1d456, "i"); /* '𝑖' -> 'i' */
	set( 0x1d457, "j"); /* '𝑗' -> 'j' */
	set( 0x1d458, "k"); /* '𝑘' -> 'k' */
	set( 0x1d459, "l"); /* '𝑙' -> 'l' */
	set( 0x1d45a, "m"); /* '𝑚' -> 'm' */
	set( 0x1d45b, "n"); /* '𝑛' -> 'n' */
	set( 0x1d45c, "o"); /* '𝑜' -> 'o' */
	set( 0x1d45d, "p"); /* '𝑝' -> 'p' */
	set( 0x1d45e, "q"); /* '𝑞' -> 'q' */
	set( 0x1d45f, "r"); /* '𝑟' -> 'r' */
	set( 0x1d460, "s"); /* '𝑠' -> 's' */
	set( 0x1d461, "t"); /* '𝑡' -> 't' */
	set( 0x1d462, "u"); /* '𝑢' -> 'u' */
	set( 0x1d463, "v"); /* '𝑣' -> 'v' */
	set( 0x1d464, "w"); /* '𝑤' -> 'w' */
	set( 0x1d465, "x"); /* '𝑥' -> 'x' */
	set( 0x1d466, "y"); /* '𝑦' -> 'y' */
	set( 0x1d467, "z"); /* '𝑧' -> 'z' */
	set( 0x1d468, "A"); /* '𝑨' -> 'A' */
	set( 0x1d469, "B"); /* '𝑩' -> 'B' */
	set( 0x1d46a, "C"); /* '𝑪' -> 'C' */
	set( 0x1d46b, "D"); /* '𝑫' -> 'D' */
	set( 0x1d46c, "E"); /* '𝑬' -> 'E' */
	set( 0x1d46d, "F"); /* '𝑭' -> 'F' */
	set( 0x1d46e, "G"); /* '𝑮' -> 'G' */
	set( 0x1d46f, "H"); /* '𝑯' -> 'H' */
	set( 0x1d470, "I"); /* '𝑰' -> 'I' */
	set( 0x1d471, "J"); /* '𝑱' -> 'J' */
	set( 0x1d472, "K"); /* '𝑲' -> 'K' */
	set( 0x1d473, "L"); /* '𝑳' -> 'L' */
	set( 0x1d474, "M"); /* '𝑴' -> 'M' */
	set( 0x1d475, "N"); /* '𝑵' -> 'N' */
	set( 0x1d476, "O"); /* '𝑶' -> 'O' */
	set( 0x1d477, "P"); /* '𝑷' -> 'P' */
	set( 0x1d478, "Q"); /* '𝑸' -> 'Q' */
	set( 0x1d479, "R"); /* '𝑹' -> 'R' */
	set( 0x1d47a, "S"); /* '𝑺' -> 'S' */
	set( 0x1d47b, "T"); /* '𝑻' -> 'T' */
	set( 0x1d47c, "U"); /* '𝑼' -> 'U' */
	set( 0x1d47d, "V"); /* '𝑽' -> 'V' */
	set( 0x1d47e, "W"); /* '𝑾' -> 'W' */
	set( 0x1d47f, "X"); /* '𝑿' -> 'X' */
	set( 0x1d480, "Y"); /* '𝒀' -> 'Y' */
	set( 0x1d481, "Z"); /* '𝒁' -> 'Z' */
	set( 0x1d482, "a"); /* '𝒂' -> 'a' */
	set( 0x1d483, "b"); /* '𝒃' -> 'b' */
	set( 0x1d484, "c"); /* '𝒄' -> 'c' */
	set( 0x1d485, "d"); /* '𝒅' -> 'd' */
	set( 0x1d486, "e"); /* '𝒆' -> 'e' */
	set( 0x1d487, "f"); /* '𝒇' -> 'f' */
	set( 0x1d488, "g"); /* '𝒈' -> 'g' */
	set( 0x1d489, "h"); /* '𝒉' -> 'h' */
	set( 0x1d48a, "i"); /* '𝒊' -> 'i' */
	set( 0x1d48b, "j"); /* '𝒋' -> 'j' */
	set( 0x1d48c, "k"); /* '𝒌' -> 'k' */
	set( 0x1d48d, "l"); /* '𝒍' -> 'l' */
	set( 0x1d48e, "m"); /* '𝒎' -> 'm' */
	set( 0x1d48f, "n"); /* '𝒏' -> 'n' */
	set( 0x1d490, "o"); /* '𝒐' -> 'o' */
	set( 0x1d491, "p"); /* '𝒑' -> 'p' */
	set( 0x1d492, "q"); /* '𝒒' -> 'q' */
	set( 0x1d493, "r"); /* '𝒓' -> 'r' */
	set( 0x1d494, "s"); /* '𝒔' -> 's' */
	set( 0x1d495, "t"); /* '𝒕' -> 't' */
	set( 0x1d496, "u"); /* '𝒖' -> 'u' */
	set( 0x1d497, "v"); /* '𝒗' -> 'v' */
	set( 0x1d498, "w"); /* '𝒘' -> 'w' */
	set( 0x1d499, "x"); /* '𝒙' -> 'x' */
	set( 0x1d49a, "y"); /* '𝒚' -> 'y' */
	set( 0x1d49b, "z"); /* '𝒛' -> 'z' */
	set( 0x1d49c, "A"); /* '𝒜' -> 'A' */
	set( 0x1d49e, "C"); /* '𝒞' -> 'C' */
	set( 0x1d49f, "D"); /* '𝒟' -> 'D' */
	set( 0x1d4a2, "G"); /* '𝒢' -> 'G' */
	set( 0x1d4a5, "J"); /* '𝒥' -> 'J' */
	set( 0x1d4a6, "K"); /* '𝒦' -> 'K' */
	set( 0x1d4a9, "N"); /* '𝒩' -> 'N' */
	set( 0x1d4aa, "O"); /* '𝒪' -> 'O' */
	set( 0x1d4ab, "P"); /* '𝒫' -> 'P' */
	set( 0x1d4ac, "Q"); /* '𝒬' -> 'Q' */
	set( 0x1d4ae, "S"); /* '𝒮' -> 'S' */
	set( 0x1d4af, "T"); /* '𝒯' -> 'T' */
	set( 0x1d4b0, "U"); /* '𝒰' -> 'U' */
	set( 0x1d4b1, "V"); /* '𝒱' -> 'V' */
	set( 0x1d4b2, "W"); /* '𝒲' -> 'W' */
	set( 0x1d4b3, "X"); /* '𝒳' -> 'X' */
	set( 0x1d4b4, "Y"); /* '𝒴' -> 'Y' */
	set( 0x1d4b5, "Z"); /* '𝒵' -> 'Z' */
	set( 0x1d4b6, "a"); /* '𝒶' -> 'a' */
	set( 0x1d4b7, "b"); /* '𝒷' -> 'b' */
	set( 0x1d4b8, "c"); /* '𝒸' -> 'c' */
	set( 0x1d4b9, "d"); /* '𝒹' -> 'd' */
	set( 0x1d4bb, "f"); /* '𝒻' -> 'f' */
	set( 0x1d4bd, "h"); /* '𝒽' -> 'h' */
	set( 0x1d4be, "i"); /* '𝒾' -> 'i' */
	set( 0x1d4bf, "j"); /* '𝒿' -> 'j' */
	set( 0x1d4c0, "k"); /* '𝓀' -> 'k' */
	set( 0x1d4c1, "l"); /* '𝓁' -> 'l' */
	set( 0x1d4c2, "m"); /* '𝓂' -> 'm' */
	set( 0x1d4c3, "n"); /* '𝓃' -> 'n' */
	set( 0x1d4c5, "p"); /* '𝓅' -> 'p' */
	set( 0x1d4c6, "q"); /* '𝓆' -> 'q' */
	set( 0x1d4c7, "r"); /* '𝓇' -> 'r' */
	set( 0x1d4c8, "s"); /* '𝓈' -> 's' */
	set( 0x1d4c9, "t"); /* '𝓉' -> 't' */
	set( 0x1d4ca, "u"); /* '𝓊' -> 'u' */
	set( 0x1d4cb, "v"); /* '𝓋' -> 'v' */
	set( 0x1d4cc, "w"); /* '𝓌' -> 'w' */
	set( 0x1d4cd, "x"); /* '𝓍' -> 'x' */
	set( 0x1d4ce, "y"); /* '𝓎' -> 'y' */
	set( 0x1d4cf, "z"); /* '𝓏' -> 'z' */
	set( 0x1d4d0, "A"); /* '𝓐' -> 'A' */
	set( 0x1d4d1, "B"); /* '𝓑' -> 'B' */
	set( 0x1d4d2, "C"); /* '𝓒' -> 'C' */
	set( 0x1d4d3, "D"); /* '𝓓' -> 'D' */
	set( 0x1d4d4, "E"); /* '𝓔' -> 'E' */
	set( 0x1d4d5, "F"); /* '𝓕' -> 'F' */
	set( 0x1d4d6, "G"); /* '𝓖' -> 'G' */
	set( 0x1d4d7, "H"); /* '𝓗' -> 'H' */
	set( 0x1d4d8, "I"); /* '𝓘' -> 'I' */
	set( 0x1d4d9, "J"); /* '𝓙' -> 'J' */
	set( 0x1d4da, "K"); /* '𝓚' -> 'K' */
	set( 0x1d4db, "L"); /* '𝓛' -> 'L' */
	set( 0x1d4dc, "M"); /* '𝓜' -> 'M' */
	set( 0x1d4dd, "N"); /* '𝓝' -> 'N' */
	set( 0x1d4de, "O"); /* '𝓞' -> 'O' */
	set( 0x1d4df, "P"); /* '𝓟' -> 'P' */
	set( 0x1d4e0, "Q"); /* '𝓠' -> 'Q' */
	set( 0x1d4e1, "R"); /* '𝓡' -> 'R' */
	set( 0x1d4e2, "S"); /* '𝓢' -> 'S' */
	set( 0x1d4e3, "T"); /* '𝓣' -> 'T' */
	set( 0x1d4e4, "U"); /* '𝓤' -> 'U' */
	set( 0x1d4e5, "V"); /* '𝓥' -> 'V' */
	set( 0x1d4e6, "W"); /* '𝓦' -> 'W' */
	set( 0x1d4e7, "X"); /* '𝓧' -> 'X' */
	set( 0x1d4e8, "Y"); /* '𝓨' -> 'Y' */
	set( 0x1d4e9, "Z"); /* '𝓩' -> 'Z' */
	set( 0x1d4ea, "a"); /* '𝓪' -> 'a' */
	set( 0x1d4eb, "b"); /* '𝓫' -> 'b' */
	set( 0x1d4ec, "c"); /* '𝓬' -> 'c' */
	set( 0x1d4ed, "d"); /* '𝓭' -> 'd' */
	set( 0x1d4ee, "e"); /* '𝓮' -> 'e' */
	set( 0x1d4ef, "f"); /* '𝓯' -> 'f' */
	set( 0x1d4f0, "g"); /* '𝓰' -> 'g' */
	set( 0x1d4f1, "h"); /* '𝓱' -> 'h' */
	set( 0x1d4f2, "i"); /* '𝓲' -> 'i' */
	set( 0x1d4f3, "j"); /* '𝓳' -> 'j' */
	set( 0x1d4f4, "k"); /* '𝓴' -> 'k' */
	set( 0x1d4f5, "l"); /* '𝓵' -> 'l' */
	set( 0x1d4f6, "m"); /* '𝓶' -> 'm' */
	set( 0x1d4f7, "n"); /* '𝓷' -> 'n' */
	set( 0x1d4f8, "o"); /* '𝓸' -> 'o' */
	set( 0x1d4f9, "p"); /* '𝓹' -> 'p' */
	set( 0x1d4fa, "q"); /* '𝓺' -> 'q' */
	set( 0x1d4fb, "r"); /* '𝓻' -> 'r' */
	set( 0x1d4fc, "s"); /* '𝓼' -> 's' */
	set( 0x1d4fd, "t"); /* '𝓽' -> 't' */
	set( 0x1d4fe, "u"); /* '𝓾' -> 'u' */
	set( 0x1d4ff, "v"); /* '𝓿' -> 'v' */
	set( 0x1d500, "w"); /* '𝔀' -> 'w' */
	set( 0x1d501, "x"); /* '𝔁' -> 'x' */
	set( 0x1d502, "y"); /* '𝔂' -> 'y' */
	set( 0x1d503, "z"); /* '𝔃' -> 'z' */
	set( 0x1d504, "A"); /* '𝔄' -> 'A' */
	set( 0x1d505, "B"); /* '𝔅' -> 'B' */
	set( 0x1d507, "D"); /* '𝔇' -> 'D' */
	set( 0x1d508, "E"); /* '𝔈' -> 'E' */
	set( 0x1d509, "F"); /* '𝔉' -> 'F' */
	set( 0x1d50a, "G"); /* '𝔊' -> 'G' */
	set( 0x1d50d, "J"); /* '𝔍' -> 'J' */
	set( 0x1d50e, "K"); /* '𝔎' -> 'K' */
	set( 0x1d50f, "L"); /* '𝔏' -> 'L' */
	set( 0x1d510, "M"); /* '𝔐' -> 'M' */
	set( 0x1d511, "N"); /* '𝔑' -> 'N' */
	set( 0x1d512, "O"); /* '𝔒' -> 'O' */
	set( 0x1d513, "P"); /* '𝔓' -> 'P' */
	set( 0x1d514, "Q"); /* '𝔔' -> 'Q' */
	set( 0x1d516, "S"); /* '𝔖' -> 'S' */
	set( 0x1d517, "T"); /* '𝔗' -> 'T' */
	set( 0x1d518, "U"); /* '𝔘' -> 'U' */
	set( 0x1d519, "V"); /* '𝔙' -> 'V' */
	set( 0x1d51a, "W"); /* '𝔚' -> 'W' */
	set( 0x1d51b, "X"); /* '𝔛' -> 'X' */
	set( 0x1d51c, "Y"); /* '𝔜' -> 'Y' */
	set( 0x1d51e, "a"); /* '𝔞' -> 'a' */
	set( 0x1d51f, "b"); /* '𝔟' -> 'b' */
	set( 0x1d520, "c"); /* '𝔠' -> 'c' */
	set( 0x1d521, "d"); /* '𝔡' -> 'd' */
	set( 0x1d522, "e"); /* '𝔢' -> 'e' */
	set( 0x1d523, "f"); /* '𝔣' -> 'f' */
	set( 0x1d524, "g"); /* '𝔤' -> 'g' */
	set( 0x1d525, "h"); /* '𝔥' -> 'h' */
	set( 0x1d526, "i"); /* '𝔦' -> 'i' */
	set( 0x1d527, "j"); /* '𝔧' -> 'j' */
	set( 0x1d528, "k"); /* '𝔨' -> 'k' */
	set( 0x1d529, "l"); /* '𝔩' -> 'l' */
	set( 0x1d52a, "m"); /* '𝔪' -> 'm' */
	set( 0x1d52b, "n"); /* '𝔫' -> 'n' */
	set( 0x1d52c, "o"); /* '𝔬' -> 'o' */
	set( 0x1d52d, "p"); /* '𝔭' -> 'p' */
	set( 0x1d52e, "q"); /* '𝔮' -> 'q' */
	set( 0x1d52f, "r"); /* '𝔯' -> 'r' */
	set( 0x1d530, "s"); /* '𝔰' -> 's' */
	set( 0x1d531, "t"); /* '𝔱' -> 't' */
	set( 0x1d532, "u"); /* '𝔲' -> 'u' */
	set( 0x1d533, "v"); /* '𝔳' -> 'v' */
	set( 0x1d534, "w"); /* '𝔴' -> 'w' */
	set( 0x1d535, "x"); /* '𝔵' -> 'x' */
	set( 0x1d536, "y"); /* '𝔶' -> 'y' */
	set( 0x1d537, "z"); /* '𝔷' -> 'z' */
	set( 0x1d538, "A"); /* '𝔸' -> 'A' */
	set( 0x1d539, "B"); /* '𝔹' -> 'B' */
	set( 0x1d53b, "D"); /* '𝔻' -> 'D' */
	set( 0x1d53c, "E"); /* '𝔼' -> 'E' */
	set( 0x1d53d, "F"); /* '𝔽' -> 'F' */
	set( 0x1d53e, "G"); /* '𝔾' -> 'G' */
	set( 0x1d540, "I"); /* '𝕀' -> 'I' */
	set( 0x1d541, "J"); /* '𝕁' -> 'J' */
	set( 0x1d542, "K"); /* '𝕂' -> 'K' */
	set( 0x1d543, "L"); /* '𝕃' -> 'L' */
	set( 0x1d544, "M"); /* '𝕄' -> 'M' */
	set( 0x1d546, "O"); /* '𝕆' -> 'O' */
	set( 0x1d54a, "S"); /* '𝕊' -> 'S' */
	set( 0x1d54b, "T"); /* '𝕋' -> 'T' */
	set( 0x1d54c, "U"); /* '𝕌' -> 'U' */
	set( 0x1d54d, "V"); /* '𝕍' -> 'V' */
	set( 0x1d54e, "W"); /* '𝕎' -> 'W' */
	set( 0x1d54f, "X"); /* '𝕏' -> 'X' */
	set( 0x1d550, "Y"); /* '𝕐' -> 'Y' */
	set( 0x1d552, "a"); /* '𝕒' -> 'a' */
	set( 0x1d553, "b"); /* '𝕓' -> 'b' */
	set( 0x1d554, "c"); /* '𝕔' -> 'c' */
	set( 0x1d555, "d"); /* '𝕕' -> 'd' */
	set( 0x1d556, "e"); /* '𝕖' -> 'e' */
	set( 0x1d557, "f"); /* '𝕗' -> 'f' */
	set( 0x1d558, "g"); /* '𝕘' -> 'g' */
	set( 0x1d559, "h"); /* '𝕙' -> 'h' */
	set( 0x1d55a, "i"); /* '𝕚' -> 'i' */
	set( 0x1d55b, "j"); /* '𝕛' -> 'j' */
	set( 0x1d55c, "k"); /* '𝕜' -> 'k' */
	set( 0x1d55d, "l"); /* '𝕝' -> 'l' */
	set( 0x1d55e, "m"); /* '𝕞' -> 'm' */
	set( 0x1d55f, "n"); /* '𝕟' -> 'n' */
	set( 0x1d560, "o"); /* '𝕠' -> 'o' */
	set( 0x1d561, "p"); /* '𝕡' -> 'p' */
	set( 0x1d562, "q"); /* '𝕢' -> 'q' */
	set( 0x1d563, "r"); /* '𝕣' -> 'r' */
	set( 0x1d564, "s"); /* '𝕤' -> 's' */
	set( 0x1d565, "t"); /* '𝕥' -> 't' */
	set( 0x1d566, "u"); /* '𝕦' -> 'u' */
	set( 0x1d567, "v"); /* '𝕧' -> 'v' */
	set( 0x1d568, "w"); /* '𝕨' -> 'w' */
	set( 0x1d569, "x"); /* '𝕩' -> 'x' */
	set( 0x1d56a, "y"); /* '𝕪' -> 'y' */
	set( 0x1d56b, "z"); /* '𝕫' -> 'z' */
	set( 0x1d56c, "A"); /* '𝕬' -> 'A' */
	set( 0x1d56d, "B"); /* '𝕭' -> 'B' */
	set( 0x1d56e, "C"); /* '𝕮' -> 'C' */
	set( 0x1d56f, "D"); /* '𝕯' -> 'D' */
	set( 0x1d570, "E"); /* '𝕰' -> 'E' */
	set( 0x1d571, "F"); /* '𝕱' -> 'F' */
	set( 0x1d572, "G"); /* '𝕲' -> 'G' */
	set( 0x1d573, "H"); /* '𝕳' -> 'H' */
	set( 0x1d574, "I"); /* '𝕴' -> 'I' */
	set( 0x1d575, "J"); /* '𝕵' -> 'J' */
	set( 0x1d576, "K"); /* '𝕶' -> 'K' */
	set( 0x1d577, "L"); /* '𝕷' -> 'L' */
	set( 0x1d578, "M"); /* '𝕸' -> 'M' */
	set( 0x1d579, "N"); /* '𝕹' -> 'N' */
	set( 0x1d57a, "O"); /* '𝕺' -> 'O' */
	set( 0x1d57b, "P"); /* '𝕻' -> 'P' */
	set( 0x1d57c, "Q"); /* '𝕼' -> 'Q' */
	set( 0x1d57d, "R"); /* '𝕽' -> 'R' */
	set( 0x1d57e, "S"); /* '𝕾' -> 'S' */
	set( 0x1d57f, "T"); /* '𝕿' -> 'T' */
	set( 0x1d580, "U"); /* '𝖀' -> 'U' */
	set( 0x1d581, "V"); /* '𝖁' -> 'V' */
	set( 0x1d582, "W"); /* '𝖂' -> 'W' */
	set( 0x1d583, "X"); /* '𝖃' -> 'X' */
	set( 0x1d584, "Y"); /* '𝖄' -> 'Y' */
	set( 0x1d585, "Z"); /* '𝖅' -> 'Z' */
	set( 0x1d586, "a"); /* '𝖆' -> 'a' */
	set( 0x1d587, "b"); /* '𝖇' -> 'b' */
	set( 0x1d588, "c"); /* '𝖈' -> 'c' */
	set( 0x1d589, "d"); /* '𝖉' -> 'd' */
	set( 0x1d58a, "e"); /* '𝖊' -> 'e' */
	set( 0x1d58b, "f"); /* '𝖋' -> 'f' */
	set( 0x1d58c, "g"); /* '𝖌' -> 'g' */
	set( 0x1d58d, "h"); /* '𝖍' -> 'h' */
	set( 0x1d58e, "i"); /* '𝖎' -> 'i' */
	set( 0x1d58f, "j"); /* '𝖏' -> 'j' */
	set( 0x1d590, "k"); /* '𝖐' -> 'k' */
	set( 0x1d591, "l"); /* '𝖑' -> 'l' */
	set( 0x1d592, "m"); /* '𝖒' -> 'm' */
	set( 0x1d593, "n"); /* '𝖓' -> 'n' */
	set( 0x1d594, "o"); /* '𝖔' -> 'o' */
	set( 0x1d595, "p"); /* '𝖕' -> 'p' */
	set( 0x1d596, "q"); /* '𝖖' -> 'q' */
	set( 0x1d597, "r"); /* '𝖗' -> 'r' */
	set( 0x1d598, "s"); /* '𝖘' -> 's' */
	set( 0x1d599, "t"); /* '𝖙' -> 't' */
	set( 0x1d59a, "u"); /* '𝖚' -> 'u' */
	set( 0x1d59b, "v"); /* '𝖛' -> 'v' */
	set( 0x1d59c, "w"); /* '𝖜' -> 'w' */
	set( 0x1d59d, "x"); /* '𝖝' -> 'x' */
	set( 0x1d59e, "y"); /* '𝖞' -> 'y' */
	set( 0x1d59f, "z"); /* '𝖟' -> 'z' */
	set( 0x1d5a0, "A"); /* '𝖠' -> 'A' */
	set( 0x1d5a1, "B"); /* '𝖡' -> 'B' */
	set( 0x1d5a2, "C"); /* '𝖢' -> 'C' */
	set( 0x1d5a3, "D"); /* '𝖣' -> 'D' */
	set( 0x1d5a4, "E"); /* '𝖤' -> 'E' */
	set( 0x1d5a5, "F"); /* '𝖥' -> 'F' */
	set( 0x1d5a6, "G"); /* '𝖦' -> 'G' */
	set( 0x1d5a7, "H"); /* '𝖧' -> 'H' */
	set( 0x1d5a8, "I"); /* '𝖨' -> 'I' */
	set( 0x1d5a9, "J"); /* '𝖩' -> 'J' */
	set( 0x1d5aa, "K"); /* '𝖪' -> 'K' */
	set( 0x1d5ab, "L"); /* '𝖫' -> 'L' */
	set( 0x1d5ac, "M"); /* '𝖬' -> 'M' */
	set( 0x1d5ad, "N"); /* '𝖭' -> 'N' */
	set( 0x1d5ae, "O"); /* '𝖮' -> 'O' */
	set( 0x1d5af, "P"); /* '𝖯' -> 'P' */
	set( 0x1d5b0, "Q"); /* '𝖰' -> 'Q' */
	set( 0x1d5b1, "R"); /* '𝖱' -> 'R' */
	set( 0x1d5b2, "S"); /* '𝖲' -> 'S' */
	set( 0x1d5b3, "T"); /* '𝖳' -> 'T' */
	set( 0x1d5b4, "U"); /* '𝖴' -> 'U' */
	set( 0x1d5b5, "V"); /* '𝖵' -> 'V' */
	set( 0x1d5b6, "W"); /* '𝖶' -> 'W' */
	set( 0x1d5b7, "X"); /* '𝖷' -> 'X' */
	set( 0x1d5b8, "Y"); /* '𝖸' -> 'Y' */
	set( 0x1d5b9, "Z"); /* '𝖹' -> 'Z' */
	set( 0x1d5ba, "a"); /* '𝖺' -> 'a' */
	set( 0x1d5bb, "b"); /* '𝖻' -> 'b' */
	set( 0x1d5bc, "c"); /* '𝖼' -> 'c' */
	set( 0x1d5bd, "d"); /* '𝖽' -> 'd' */
	set( 0x1d5be, "e"); /* '𝖾' -> 'e' */
	set( 0x1d5bf, "f"); /* '𝖿' -> 'f' */
	set( 0x1d5c0, "g"); /* '𝗀' -> 'g' */
	set( 0x1d5c1, "h"); /* '𝗁' -> 'h' */
	set( 0x1d5c2, "i"); /* '𝗂' -> 'i' */
	set( 0x1d5c3, "j"); /* '𝗃' -> 'j' */
	set( 0x1d5c4, "k"); /* '𝗄' -> 'k' */
	set( 0x1d5c5, "l"); /* '𝗅' -> 'l' */
	set( 0x1d5c6, "m"); /* '𝗆' -> 'm' */
	set( 0x1d5c7, "n"); /* '𝗇' -> 'n' */
	set( 0x1d5c8, "o"); /* '𝗈' -> 'o' */
	set( 0x1d5c9, "p"); /* '𝗉' -> 'p' */
	set( 0x1d5ca, "q"); /* '𝗊' -> 'q' */
	set( 0x1d5cb, "r"); /* '𝗋' -> 'r' */
	set( 0x1d5cc, "s"); /* '𝗌' -> 's' */
	set( 0x1d5cd, "t"); /* '𝗍' -> 't' */
	set( 0x1d5ce, "u"); /* '𝗎' -> 'u' */
	set( 0x1d5cf, "v"); /* '𝗏' -> 'v' */
	set( 0x1d5d0, "w"); /* '𝗐' -> 'w' */
	set( 0x1d5d1, "x"); /* '𝗑' -> 'x' */
	set( 0x1d5d2, "y"); /* '𝗒' -> 'y' */
	set( 0x1d5d3, "z"); /* '𝗓' -> 'z' */
	set( 0x1d5d4, "A"); /* '𝗔' -> 'A' */
	set( 0x1d5d5, "B"); /* '𝗕' -> 'B' */
	set( 0x1d5d6, "C"); /* '𝗖' -> 'C' */
	set( 0x1d5d7, "D"); /* '𝗗' -> 'D' */
	set( 0x1d5d8, "E"); /* '𝗘' -> 'E' */
	set( 0x1d5d9, "F"); /* '𝗙' -> 'F' */
	set( 0x1d5da, "G"); /* '𝗚' -> 'G' */
	set( 0x1d5db, "H"); /* '𝗛' -> 'H' */
	set( 0x1d5dc, "I"); /* '𝗜' -> 'I' */
	set( 0x1d5dd, "J"); /* '𝗝' -> 'J' */
	set( 0x1d5de, "K"); /* '𝗞' -> 'K' */
	set( 0x1d5df, "L"); /* '𝗟' -> 'L' */
	set( 0x1d5e0, "M"); /* '𝗠' -> 'M' */
	set( 0x1d5e1, "N"); /* '𝗡' -> 'N' */
	set( 0x1d5e2, "O"); /* '𝗢' -> 'O' */
	set( 0x1d5e3, "P"); /* '𝗣' -> 'P' */
	set( 0x1d5e4, "Q"); /* '𝗤' -> 'Q' */
	set( 0x1d5e5, "R"); /* '𝗥' -> 'R' */
	set( 0x1d5e6, "S"); /* '𝗦' -> 'S' */
	set( 0x1d5e7, "T"); /* '𝗧' -> 'T' */
	set( 0x1d5e8, "U"); /* '𝗨' -> 'U' */
	set( 0x1d5e9, "V"); /* '𝗩' -> 'V' */
	set( 0x1d5ea, "W"); /* '𝗪' -> 'W' */
	set( 0x1d5eb, "X"); /* '𝗫' -> 'X' */
	set( 0x1d5ec, "Y"); /* '𝗬' -> 'Y' */
	set( 0x1d5ed, "Z"); /* '𝗭' -> 'Z' */
	set( 0x1d5ee, "a"); /* '𝗮' -> 'a' */
	set( 0x1d5ef, "b"); /* '𝗯' -> 'b' */
	set( 0x1d5f0, "c"); /* '𝗰' -> 'c' */
	set( 0x1d5f1, "d"); /* '𝗱' -> 'd' */
	set( 0x1d5f2, "e"); /* '𝗲' -> 'e' */
	set( 0x1d5f3, "f"); /* '𝗳' -> 'f' */
	set( 0x1d5f4, "g"); /* '𝗴' -> 'g' */
	set( 0x1d5f5, "h"); /* '𝗵' -> 'h' */
	set( 0x1d5f6, "i"); /* '𝗶' -> 'i' */
	set( 0x1d5f7, "j"); /* '𝗷' -> 'j' */
	set( 0x1d5f8, "k"); /* '𝗸' -> 'k' */
	set( 0x1d5f9, "l"); /* '𝗹' -> 'l' */
	set( 0x1d5fa, "m"); /* '𝗺' -> 'm' */
	set( 0x1d5fb, "n"); /* '𝗻' -> 'n' */
	set( 0x1d5fc, "o"); /* '𝗼' -> 'o' */
	set( 0x1d5fd, "p"); /* '𝗽' -> 'p' */
	set( 0x1d5fe, "q"); /* '𝗾' -> 'q' */
	set( 0x1d5ff, "r"); /* '𝗿' -> 'r' */
	set( 0x1d600, "s"); /* '𝘀' -> 's' */
	set( 0x1d601, "t"); /* '𝘁' -> 't' */
	set( 0x1d602, "u"); /* '𝘂' -> 'u' */
	set( 0x1d603, "v"); /* '𝘃' -> 'v' */
	set( 0x1d604, "w"); /* '𝘄' -> 'w' */
	set( 0x1d605, "x"); /* '𝘅' -> 'x' */
	set( 0x1d606, "y"); /* '𝘆' -> 'y' */
	set( 0x1d607, "z"); /* '𝘇' -> 'z' */
	set( 0x1d608, "A"); /* '𝘈' -> 'A' */
	set( 0x1d609, "B"); /* '𝘉' -> 'B' */
	set( 0x1d60a, "C"); /* '𝘊' -> 'C' */
	set( 0x1d60b, "D"); /* '𝘋' -> 'D' */
	set( 0x1d60c, "E"); /* '𝘌' -> 'E' */
	set( 0x1d60d, "F"); /* '𝘍' -> 'F' */
	set( 0x1d60e, "G"); /* '𝘎' -> 'G' */
	set( 0x1d60f, "H"); /* '𝘏' -> 'H' */
	set( 0x1d610, "I"); /* '𝘐' -> 'I' */
	set( 0x1d611, "J"); /* '𝘑' -> 'J' */
	set( 0x1d612, "K"); /* '𝘒' -> 'K' */
	set( 0x1d613, "L"); /* '𝘓' -> 'L' */
	set( 0x1d614, "M"); /* '𝘔' -> 'M' */
	set( 0x1d615, "N"); /* '𝘕' -> 'N' */
	set( 0x1d616, "O"); /* '𝘖' -> 'O' */
	set( 0x1d617, "P"); /* '𝘗' -> 'P' */
	set( 0x1d618, "Q"); /* '𝘘' -> 'Q' */
	set( 0x1d619, "R"); /* '𝘙' -> 'R' */
	set( 0x1d61a, "S"); /* '𝘚' -> 'S' */
	set( 0x1d61b, "T"); /* '𝘛' -> 'T' */
	set( 0x1d61c, "U"); /* '𝘜' -> 'U' */
	set( 0x1d61d, "V"); /* '𝘝' -> 'V' */
	set( 0x1d61e, "W"); /* '𝘞' -> 'W' */
	set( 0x1d61f, "X"); /* '𝘟' -> 'X' */
	set( 0x1d620, "Y"); /* '𝘠' -> 'Y' */
	set( 0x1d621, "Z"); /* '𝘡' -> 'Z' */
	set( 0x1d622, "a"); /* '𝘢' -> 'a' */
	set( 0x1d623, "b"); /* '𝘣' -> 'b' */
	set( 0x1d624, "c"); /* '𝘤' -> 'c' */
	set( 0x1d625, "d"); /* '𝘥' -> 'd' */
	set( 0x1d626, "e"); /* '𝘦' -> 'e' */
	set( 0x1d627, "f"); /* '𝘧' -> 'f' */
	set( 0x1d628, "g"); /* '𝘨' -> 'g' */
	set( 0x1d629, "h"); /* '𝘩' -> 'h' */
	set( 0x1d62a, "i"); /* '𝘪' -> 'i' */
	set( 0x1d62b, "j"); /* '𝘫' -> 'j' */
	set( 0x1d62c, "k"); /* '𝘬' -> 'k' */
	set( 0x1d62d, "l"); /* '𝘭' -> 'l' */
	set( 0x1d62e, "m"); /* '𝘮' -> 'm' */
	set( 0x1d62f, "n"); /* '𝘯' -> 'n' */
	set( 0x1d630, "o"); /* '𝘰' -> 'o' */
	set( 0x1d631, "p"); /* '𝘱' -> 'p' */
	set( 0x1d632, "q"); /* '𝘲' -> 'q' */
	set( 0x1d633, "r"); /* '𝘳' -> 'r' */
	set( 0x1d634, "s"); /* '𝘴' -> 's' */
	set( 0x1d635, "t"); /* '𝘵' -> 't' */
	set( 0x1d636, "u"); /* '𝘶' -> 'u' */
	set( 0x1d637, "v"); /* '𝘷' -> 'v' */
	set( 0x1d638, "w"); /* '𝘸' -> 'w' */
	set( 0x1d639, "x"); /* '𝘹' -> 'x' */
	set( 0x1d63a, "y"); /* '𝘺' -> 'y' */
	set( 0x1d63b, "z"); /* '𝘻' -> 'z' */
	set( 0x1d63c, "A"); /* '𝘼' -> 'A' */
	set( 0x1d63d, "B"); /* '𝘽' -> 'B' */
	set( 0x1d63e, "C"); /* '𝘾' -> 'C' */
	set( 0x1d63f, "D"); /* '𝘿' -> 'D' */
	set( 0x1d640, "E"); /* '𝙀' -> 'E' */
	set( 0x1d641, "F"); /* '𝙁' -> 'F' */
	set( 0x1d642, "G"); /* '𝙂' -> 'G' */
	set( 0x1d643, "H"); /* '𝙃' -> 'H' */
	set( 0x1d644, "I"); /* '𝙄' -> 'I' */
	set( 0x1d645, "J"); /* '𝙅' -> 'J' */
	set( 0x1d646, "K"); /* '𝙆' -> 'K' */
	set( 0x1d647, "L"); /* '𝙇' -> 'L' */
	set( 0x1d648, "M"); /* '𝙈' -> 'M' */
	set( 0x1d649, "N"); /* '𝙉' -> 'N' */
	set( 0x1d64a, "O"); /* '𝙊' -> 'O' */
	set( 0x1d64b, "P"); /* '𝙋' -> 'P' */
	set( 0x1d64c, "Q"); /* '𝙌' -> 'Q' */
	set( 0x1d64d, "R"); /* '𝙍' -> 'R' */
	set( 0x1d64e, "S"); /* '𝙎' -> 'S' */
	set( 0x1d64f, "T"); /* '𝙏' -> 'T' */
	set( 0x1d650, "U"); /* '𝙐' -> 'U' */
	set( 0x1d651, "V"); /* '𝙑' -> 'V' */
	set( 0x1d652, "W"); /* '𝙒' -> 'W' */
	set( 0x1d653, "X"); /* '𝙓' -> 'X' */
	set( 0x1d654, "Y"); /* '𝙔' -> 'Y' */
	set( 0x1d655, "Z"); /* '𝙕' -> 'Z' */
	set( 0x1d656, "a"); /* '𝙖' -> 'a' */
	set( 0x1d657, "b"); /* '𝙗' -> 'b' */
	set( 0x1d658, "c"); /* '𝙘' -> 'c' */
	set( 0x1d659, "d"); /* '𝙙' -> 'd' */
	set( 0x1d65a, "e"); /* '𝙚' -> 'e' */
	set( 0x1d65b, "f"); /* '𝙛' -> 'f' */
	set( 0x1d65c, "g"); /* '𝙜' -> 'g' */
	set( 0x1d65d, "h"); /* '𝙝' -> 'h' */
	set( 0x1d65e, "i"); /* '𝙞' -> 'i' */
	set( 0x1d65f, "j"); /* '𝙟' -> 'j' */
	set( 0x1d660, "k"); /* '𝙠' -> 'k' */
	set( 0x1d661, "l"); /* '𝙡' -> 'l' */
	set( 0x1d662, "m"); /* '𝙢' -> 'm' */
	set( 0x1d663, "n"); /* '𝙣' -> 'n' */
	set( 0x1d664, "o"); /* '𝙤' -> 'o' */
	set( 0x1d665, "p"); /* '𝙥' -> 'p' */
	set( 0x1d666, "q"); /* '𝙦' -> 'q' */
	set( 0x1d667, "r"); /* '𝙧' -> 'r' */
	set( 0x1d668, "s"); /* '𝙨' -> 's' */
	set( 0x1d669, "t"); /* '𝙩' -> 't' */
	set( 0x1d66a, "u"); /* '𝙪' -> 'u' */
	set( 0x1d66b, "v"); /* '𝙫' -> 'v' */
	set( 0x1d66c, "w"); /* '𝙬' -> 'w' */
	set( 0x1d66d, "x"); /* '𝙭' -> 'x' */
	set( 0x1d66e, "y"); /* '𝙮' -> 'y' */
	set( 0x1d66f, "z"); /* '𝙯' -> 'z' */
	set( 0x1d670, "A"); /* '𝙰' -> 'A' */
	set( 0x1d671, "B"); /* '𝙱' -> 'B' */
	set( 0x1d672, "C"); /* '𝙲' -> 'C' */
	set( 0x1d673, "D"); /* '𝙳' -> 'D' */
	set( 0x1d674, "E"); /* '𝙴' -> 'E' */
	set( 0x1d675, "F"); /* '𝙵' -> 'F' */
	set( 0x1d676, "G"); /* '𝙶' -> 'G' */
	set( 0x1d677, "H"); /* '𝙷' -> 'H' */
	set( 0x1d678, "I"); /* '𝙸' -> 'I' */
	set( 0x1d679, "J"); /* '𝙹' -> 'J' */
	set( 0x1d67a, "K"); /* '𝙺' -> 'K' */
	set( 0x1d67b, "L"); /* '𝙻' -> 'L' */
	set( 0x1d67c, "M"); /* '𝙼' -> 'M' */
	set( 0x1d67d, "N"); /* '𝙽' -> 'N' */
	set( 0x1d67e, "O"); /* '𝙾' -> 'O' */
	set( 0x1d67f, "P"); /* '𝙿' -> 'P' */
	set( 0x1d680, "Q"); /* '𝚀' -> 'Q' */
	set( 0x1d681, "R"); /* '𝚁' -> 'R' */
	set( 0x1d682, "S"); /* '𝚂' -> 'S' */
	set( 0x1d683, "T"); /* '𝚃' -> 'T' */
	set( 0x1d684, "U"); /* '𝚄' -> 'U' */
	set( 0x1d685, "V"); /* '𝚅' -> 'V' */
	set( 0x1d686, "W"); /* '𝚆' -> 'W' */
	set( 0x1d687, "X"); /* '𝚇' -> 'X' */
	set( 0x1d688, "Y"); /* '𝚈' -> 'Y' */
	set( 0x1d689, "Z"); /* '𝚉' -> 'Z' */
	set( 0x1d68a, "a"); /* '𝚊' -> 'a' */
	set( 0x1d68b, "b"); /* '𝚋' -> 'b' */
	set( 0x1d68c, "c"); /* '𝚌' -> 'c' */
	set( 0x1d68d, "d"); /* '𝚍' -> 'd' */
	set( 0x1d68e, "e"); /* '𝚎' -> 'e' */
	set( 0x1d68f, "f"); /* '𝚏' -> 'f' */
	set( 0x1d690, "g"); /* '𝚐' -> 'g' */
	set( 0x1d691, "h"); /* '𝚑' -> 'h' */
	set( 0x1d692, "i"); /* '𝚒' -> 'i' */
	set( 0x1d693, "j"); /* '𝚓' -> 'j' */
	set( 0x1d694, "k"); /* '𝚔' -> 'k' */
	set( 0x1d695, "l"); /* '𝚕' -> 'l' */
	set( 0x1d696, "m"); /* '𝚖' -> 'm' */
	set( 0x1d697, "n"); /* '𝚗' -> 'n' */
	set( 0x1d698, "o"); /* '𝚘' -> 'o' */
	set( 0x1d699, "p"); /* '𝚙' -> 'p' */
	set( 0x1d69a, "q"); /* '𝚚' -> 'q' */
	set( 0x1d69b, "r"); /* '𝚛' -> 'r' */
	set( 0x1d69c, "s"); /* '𝚜' -> 's' */
	set( 0x1d69d, "t"); /* '𝚝' -> 't' */
	set( 0x1d69e, "u"); /* '𝚞' -> 'u' */
	set( 0x1d69f, "v"); /* '𝚟' -> 'v' */
	set( 0x1d6a0, "w"); /* '𝚠' -> 'w' */
	set( 0x1d6a1, "x"); /* '𝚡' -> 'x' */
	set( 0x1d6a2, "y"); /* '𝚢' -> 'y' */
	set( 0x1d6a3, "z"); /* '𝚣' -> 'z' */
	set( 0x1d7ce, "0"); /* '𝟎' -> '0' */
	set( 0x1d7cf, "1"); /* '𝟏' -> '1' */
	set( 0x1d7d0, "2"); /* '𝟐' -> '2' */
	set( 0x1d7d1, "3"); /* '𝟑' -> '3' */
	set( 0x1d7d2, "4"); /* '𝟒' -> '4' */
	set( 0x1d7d3, "5"); /* '𝟓' -> '5' */
	set( 0x1d7d4, "6"); /* '𝟔' -> '6' */
	set( 0x1d7d5, "7"); /* '𝟕' -> '7' */
	set( 0x1d7d6, "8"); /* '𝟖' -> '8' */
	set( 0x1d7d7, "9"); /* '𝟗' -> '9' */
	set( 0x1d7d8, "0"); /* '𝟘' -> '0' */
	set( 0x1d7d9, "1"); /* '𝟙' -> '1' */
	set( 0x1d7da, "2"); /* '𝟚' -> '2' */
	set( 0x1d7db, "3"); /* '𝟛' -> '3' */
	set( 0x1d7dc, "4"); /* '𝟜' -> '4' */
	set( 0x1d7dd, "5"); /* '𝟝' -> '5' */
	set( 0x1d7de, "6"); /* '𝟞' -> '6' */
	set( 0x1d7df, "7"); /* '𝟟' -> '7' */
	set( 0x1d7e0, "8"); /* '𝟠' -> '8' */
	set( 0x1d7e1, "9"); /* '𝟡' -> '9' */
	set( 0x1d7e2, "0"); /* '𝟢' -> '0' */
	set( 0x1d7e3, "1"); /* '𝟣' -> '1' */
	set( 0x1d7e4, "2"); /* '𝟤' -> '2' */
	set( 0x1d7e5, "3"); /* '𝟥' -> '3' */
	set( 0x1d7e6, "4"); /* '𝟦' -> '4' */
	set( 0x1d7e7, "5"); /* '𝟧' -> '5' */
	set( 0x1d7e8, "6"); /* '𝟨' -> '6' */
	set( 0x1d7e9, "7"); /* '𝟩' -> '7' */
	set( 0x1d7ea, "8"); /* '𝟪' -> '8' */
	set( 0x1d7eb, "9"); /* '𝟫' -> '9' */
	set( 0x1d7ec, "0"); /* '𝟬' -> '0' */
	set( 0x1d7ed, "1"); /* '𝟭' -> '1' */
	set( 0x1d7ee, "2"); /* '𝟮' -> '2' */
	set( 0x1d7ef, "3"); /* '𝟯' -> '3' */
	set( 0x1d7f0, "4"); /* '𝟰' -> '4' */
	set( 0x1d7f1, "5"); /* '𝟱' -> '5' */
	set( 0x1d7f2, "6"); /* '𝟲' -> '6' */
	set( 0x1d7f3, "7"); /* '𝟳' -> '7' */
	set( 0x1d7f4, "8"); /* '𝟴' -> '8' */
	set( 0x1d7f5, "9"); /* '𝟵' -> '9' */
	set( 0x1d7f6, "0"); /* '𝟶' -> '0' */
	set( 0x1d7f7, "1"); /* '𝟷' -> '1' */
	set( 0x1d7f8, "2"); /* '𝟸' -> '2' */
	set( 0x1d7f9, "3"); /* '𝟹' -> '3' */
	set( 0x1d7fa, "4"); /* '𝟺' -> '4' */
	set( 0x1d7fb, "5"); /* '𝟻' -> '5' */
	set( 0x1d7fc, "6"); /* '𝟼' -> '6' */
	set( 0x1d7fd, "7"); /* '𝟽' -> '7' */
	set( 0x1d7fe, "8"); /* '𝟾' -> '8' */
	set( 0x1d7ff, "9"); /* '𝟿' -> '9' */
	set( 0x1f100, "0."); /* '🄀' -> '0.' */
	set( 0x1f101, "0,"); /* '🄁' -> '0,' */
	set( 0x1f102, "1,"); /* '🄂' -> '1,' */
	set( 0x1f103, "2,"); /* '🄃' -> '2,' */
	set( 0x1f104, "3,"); /* '🄄' -> '3,' */
	set( 0x1f105, "4,"); /* '🄅' -> '4,' */
	set( 0x1f106, "5,"); /* '🄆' -> '5,' */
	set( 0x1f107, "6,"); /* '🄇' -> '6,' */
	set( 0x1f108, "7,"); /* '🄈' -> '7,' */
	set( 0x1f109, "8,"); /* '🄉' -> '8,' */
	set( 0x1f10a, "9,"); /* '🄊' -> '9,' */
	set( 0x1f110, "(A)"); /* '🄐' -> '(A)' */
	set( 0x1f111, "(B)"); /* '🄑' -> '(B)' */
	set( 0x1f112, "(C)"); /* '🄒' -> '(C)' */
	set( 0x1f113, "(D)"); /* '🄓' -> '(D)' */
	set( 0x1f114, "(E)"); /* '🄔' -> '(E)' */
	set( 0x1f115, "(F)"); /* '🄕' -> '(F)' */
	set( 0x1f116, "(G)"); /* '🄖' -> '(G)' */
	set( 0x1f117, "(H)"); /* '🄗' -> '(H)' */
	set( 0x1f118, "(I)"); /* '🄘' -> '(I)' */
	set( 0x1f119, "(J)"); /* '🄙' -> '(J)' */
	set( 0x1f11a, "(K)"); /* '🄚' -> '(K)' */
	set( 0x1f11b, "(L)"); /* '🄛' -> '(L)' */
	set( 0x1f11c, "(M)"); /* '🄜' -> '(M)' */
	set( 0x1f11d, "(N)"); /* '🄝' -> '(N)' */
	set( 0x1f11e, "(O)"); /* '🄞' -> '(O)' */
	set( 0x1f11f, "(P)"); /* '🄟' -> '(P)' */
	set( 0x1f120, "(Q)"); /* '🄠' -> '(Q)' */
	set( 0x1f121, "(R)"); /* '🄡' -> '(R)' */
	set( 0x1f122, "(S)"); /* '🄢' -> '(S)' */
	set( 0x1f123, "(T)"); /* '🄣' -> '(T)' */
	set( 0x1f124, "(U)"); /* '🄤' -> '(U)' */
	set( 0x1f125, "(V)"); /* '🄥' -> '(V)' */
	set( 0x1f126, "(W)"); /* '🄦' -> '(W)' */
	set( 0x1f127, "(X)"); /* '🄧' -> '(X)' */
	set( 0x1f128, "(Y)"); /* '🄨' -> '(Y)' */
	set( 0x1f129, "(Z)"); /* '🄩' -> '(Z)' */
	set( 0x1f12a, "S"); /* '🄪' -> 'S' */
	set( 0x1f12b, "C"); /* '🄫' -> 'C' */
	set( 0x1f12c, "R"); /* '🄬' -> 'R' */
	set( 0x1f12d, "CD"); /* '🄭' -> 'CD' */
	set( 0x1f12e, "WZ"); /* '🄮' -> 'WZ' */
	set( 0x1f131, "B"); /* '🄱' -> 'B' */
	set( 0x1f13d, "N"); /* '🄽' -> 'N' */
	set( 0x1f13f, "P"); /* '🄿' -> 'P' */
	set( 0x1f142, "S"); /* '🅂' -> 'S' */
	set( 0x1f146, "W"); /* '🅆' -> 'W' */
	set( 0x1f14a, "HV"); /* '🅊' -> 'HV' */
	set( 0x1f14b, "MV"); /* '🅋' -> 'MV' */
	set( 0x1f14c, "SD"); /* '🅌' -> 'SD' */
	set( 0x1f14d, "SS"); /* '🅍' -> 'SS' */
	set( 0x1f14e, "PPV"); /* '🅎' -> 'PPV' */
	set( 0x1f190, "DJ"); /* '🆐' -> 'DJ' */
}

void CharMap::buildMapTolower()
{
	set( 0x41, 0x61); /* 'A' -> 'a' */
	set( 0x42, 0x62); /* 'B' -> 'b' */
	set( 0x43, 0x63); /* 'C' -> 'c' */
	set( 0x44, 0x64); /* 'D' -> 'd' */
	set( 0x45, 0x65); /* 'E' -> 'e' */
	set( 0x46, 0x66); /* 'F' -> 'f' */
	set( 0x47, 0x67); /* 'G' -> 'g' */
	set( 0x48, 0x68); /* 'H' -> 'h' */
	set( 0x49, 0x69); /* 'I' -> 'i' */
	set( 0x4a, 0x6a); /* 'J' -> 'j' */
	set( 0x4b, 0x6b); /* 'K' -> 'k' */
	set( 0x4c, 0x6c); /* 'L' -> 'l' */
	set( 0x4d, 0x6d); /* 'M' -> 'm' */
	set( 0x4e, 0x6e); /* 'N' -> 'n' */
	set( 0x4f, 0x6f); /* 'O' -> 'o' */
	set( 0x50, 0x70); /* 'P' -> 'p' */
	set( 0x51, 0x71); /* 'Q' -> 'q' */
	set( 0x52, 0x72); /* 'R' -> 'r' */
	set( 0x53, 0x73); /* 'S' -> 's' */
	set( 0x54, 0x74); /* 'T' -> 't' */
	set( 0x55, 0x75); /* 'U' -> 'u' */
	set( 0x56, 0x76); /* 'V' -> 'v' */
	set( 0x57, 0x77); /* 'W' -> 'w' */
	set( 0x58, 0x78); /* 'X' -> 'x' */
	set( 0x59, 0x79); /* 'Y' -> 'y' */
	set( 0x5a, 0x7a); /* 'Z' -> 'z' */
	set( 0xc0, 0xe0); /* 'À' -> 'à' */
	set( 0xc1, 0xe1); /* 'Á' -> 'á' */
	set( 0xc2, 0xe2); /* 'Â' -> 'â' */
	set( 0xc3, 0xe3); /* 'Ã' -> 'ã' */
	set( 0xc4, 0xe4); /* 'Ä' -> 'ä' */
	set( 0xc5, 0xe5); /* 'Å' -> 'å' */
	set( 0xc6, 0xe6); /* 'Æ' -> 'æ' */
	set( 0xc7, 0xe7); /* 'Ç' -> 'ç' */
	set( 0xc8, 0xe8); /* 'È' -> 'è' */
	set( 0xc9, 0xe9); /* 'É' -> 'é' */
	set( 0xca, 0xea); /* 'Ê' -> 'ê' */
	set( 0xcb, 0xeb); /* 'Ë' -> 'ë' */
	set( 0xcc, 0xec); /* 'Ì' -> 'ì' */
	set( 0xcd, 0xed); /* 'Í' -> 'í' */
	set( 0xce, 0xee); /* 'Î' -> 'î' */
	set( 0xcf, 0xef); /* 'Ï' -> 'ï' */
	set( 0xd0, 0xf0); /* 'Ð' -> 'ð' */
	set( 0xd1, 0xf1); /* 'Ñ' -> 'ñ' */
	set( 0xd2, 0xf2); /* 'Ò' -> 'ò' */
	set( 0xd3, 0xf3); /* 'Ó' -> 'ó' */
	set( 0xd4, 0xf4); /* 'Ô' -> 'ô' */
	set( 0xd5, 0xf5); /* 'Õ' -> 'õ' */
	set( 0xd6, 0xf6); /* 'Ö' -> 'ö' */
	set( 0xd8, 0xf8); /* 'Ø' -> 'ø' */
	set( 0xd9, 0xf9); /* 'Ù' -> 'ù' */
	set( 0xda, 0xfa); /* 'Ú' -> 'ú' */
	set( 0xdb, 0xfb); /* 'Û' -> 'û' */
	set( 0xdc, 0xfc); /* 'Ü' -> 'ü' */
	set( 0xdd, 0xfd); /* 'Ý' -> 'ý' */
	set( 0xde, 0xfe); /* 'Þ' -> 'þ' */
	set( 0x100, 0x101); /* 'Ā' -> 'ā' */
	set( 0x102, 0x103); /* 'Ă' -> 'ă' */
	set( 0x104, 0x105); /* 'Ą' -> 'ą' */
	set( 0x106, 0x107); /* 'Ć' -> 'ć' */
	set( 0x108, 0x109); /* 'Ĉ' -> 'ĉ' */
	set( 0x10a, 0x10b); /* 'Ċ' -> 'ċ' */
	set( 0x10c, 0x10d); /* 'Č' -> 'č' */
	set( 0x10e, 0x10f); /* 'Ď' -> 'ď' */
	set( 0x110, 0x111); /* 'Đ' -> 'đ' */
	set( 0x112, 0x113); /* 'Ē' -> 'ē' */
	set( 0x114, 0x115); /* 'Ĕ' -> 'ĕ' */
	set( 0x116, 0x117); /* 'Ė' -> 'ė' */
	set( 0x118, 0x119); /* 'Ę' -> 'ę' */
	set( 0x11a, 0x11b); /* 'Ě' -> 'ě' */
	set( 0x11c, 0x11d); /* 'Ĝ' -> 'ĝ' */
	set( 0x11e, 0x11f); /* 'Ğ' -> 'ğ' */
	set( 0x120, 0x121); /* 'Ġ' -> 'ġ' */
	set( 0x122, 0x123); /* 'Ģ' -> 'ģ' */
	set( 0x124, 0x125); /* 'Ĥ' -> 'ĥ' */
	set( 0x126, 0x127); /* 'Ħ' -> 'ħ' */
	set( 0x128, 0x129); /* 'Ĩ' -> 'ĩ' */
	set( 0x12a, 0x12b); /* 'Ī' -> 'ī' */
	set( 0x12c, 0x12d); /* 'Ĭ' -> 'ĭ' */
	set( 0x12e, 0x12f); /* 'Į' -> 'į' */
	set( 0x130, 0x69); /* 'İ' -> 'i' */
	set( 0x132, 0x133); /* 'Ĳ' -> 'ĳ' */
	set( 0x134, 0x135); /* 'Ĵ' -> 'ĵ' */
	set( 0x136, 0x137); /* 'Ķ' -> 'ķ' */
	set( 0x139, 0x13a); /* 'Ĺ' -> 'ĺ' */
	set( 0x13b, 0x13c); /* 'Ļ' -> 'ļ' */
	set( 0x13d, 0x13e); /* 'Ľ' -> 'ľ' */
	set( 0x13f, 0x140); /* 'Ŀ' -> 'ŀ' */
	set( 0x141, 0x142); /* 'Ł' -> 'ł' */
	set( 0x143, 0x144); /* 'Ń' -> 'ń' */
	set( 0x145, 0x146); /* 'Ņ' -> 'ņ' */
	set( 0x147, 0x148); /* 'Ň' -> 'ň' */
	set( 0x14a, 0x14b); /* 'Ŋ' -> 'ŋ' */
	set( 0x14c, 0x14d); /* 'Ō' -> 'ō' */
	set( 0x14e, 0x14f); /* 'Ŏ' -> 'ŏ' */
	set( 0x150, 0x151); /* 'Ő' -> 'ő' */
	set( 0x152, 0x153); /* 'Œ' -> 'œ' */
	set( 0x154, 0x155); /* 'Ŕ' -> 'ŕ' */
	set( 0x156, 0x157); /* 'Ŗ' -> 'ŗ' */
	set( 0x158, 0x159); /* 'Ř' -> 'ř' */
	set( 0x15a, 0x15b); /* 'Ś' -> 'ś' */
	set( 0x15c, 0x15d); /* 'Ŝ' -> 'ŝ' */
	set( 0x15e, 0x15f); /* 'Ş' -> 'ş' */
	set( 0x160, 0x161); /* 'Š' -> 'š' */
	set( 0x162, 0x163); /* 'Ţ' -> 'ţ' */
	set( 0x164, 0x165); /* 'Ť' -> 'ť' */
	set( 0x166, 0x167); /* 'Ŧ' -> 'ŧ' */
	set( 0x168, 0x169); /* 'Ũ' -> 'ũ' */
	set( 0x16a, 0x16b); /* 'Ū' -> 'ū' */
	set( 0x16c, 0x16d); /* 'Ŭ' -> 'ŭ' */
	set( 0x16e, 0x16f); /* 'Ů' -> 'ů' */
	set( 0x170, 0x171); /* 'Ű' -> 'ű' */
	set( 0x172, 0x173); /* 'Ų' -> 'ų' */
	set( 0x174, 0x175); /* 'Ŵ' -> 'ŵ' */
	set( 0x176, 0x177); /* 'Ŷ' -> 'ŷ' */
	set( 0x178, 0xff); /* 'Ÿ' -> 'ÿ' */
	set( 0x179, 0x17a); /* 'Ź' -> 'ź' */
	set( 0x17b, 0x17c); /* 'Ż' -> 'ż' */
	set( 0x17d, 0x17e); /* 'Ž' -> 'ž' */
	set( 0x181, 0x253); /* 'Ɓ' -> 'ɓ' */
	set( 0x182, 0x183); /* 'Ƃ' -> 'ƃ' */
	set( 0x184, 0x185); /* 'Ƅ' -> 'ƅ' */
	set( 0x186, 0x254); /* 'Ɔ' -> 'ɔ' */
	set( 0x187, 0x188); /* 'Ƈ' -> 'ƈ' */
	set( 0x189, 0x256); /* 'Ɖ' -> 'ɖ' */
	set( 0x18a, 0x257); /* 'Ɗ' -> 'ɗ' */
	set( 0x18b, 0x18c); /* 'Ƌ' -> 'ƌ' */
	set( 0x18e, 0x1dd); /* 'Ǝ' -> 'ǝ' */
	set( 0x18f, 0x259); /* 'Ə' -> 'ə' */
	set( 0x190, 0x25b); /* 'Ɛ' -> 'ɛ' */
	set( 0x191, 0x192); /* 'Ƒ' -> 'ƒ' */
	set( 0x193, 0x260); /* 'Ɠ' -> 'ɠ' */
	set( 0x194, 0x263); /* 'Ɣ' -> 'ɣ' */
	set( 0x196, 0x269); /* 'Ɩ' -> 'ɩ' */
	set( 0x197, 0x268); /* 'Ɨ' -> 'ɨ' */
	set( 0x198, 0x199); /* 'Ƙ' -> 'ƙ' */
	set( 0x19c, 0x26f); /* 'Ɯ' -> 'ɯ' */
	set( 0x19d, 0x272); /* 'Ɲ' -> 'ɲ' */
	set( 0x19f, 0x275); /* 'Ɵ' -> 'ɵ' */
	set( 0x1a0, 0x1a1); /* 'Ơ' -> 'ơ' */
	set( 0x1a2, 0x1a3); /* 'Ƣ' -> 'ƣ' */
	set( 0x1a4, 0x1a5); /* 'Ƥ' -> 'ƥ' */
	set( 0x1a6, 0x280); /* 'Ʀ' -> 'ʀ' */
	set( 0x1a7, 0x1a8); /* 'Ƨ' -> 'ƨ' */
	set( 0x1a9, 0x283); /* 'Ʃ' -> 'ʃ' */
	set( 0x1ac, 0x1ad); /* 'Ƭ' -> 'ƭ' */
	set( 0x1ae, 0x288); /* 'Ʈ' -> 'ʈ' */
	set( 0x1af, 0x1b0); /* 'Ư' -> 'ư' */
	set( 0x1b1, 0x28a); /* 'Ʊ' -> 'ʊ' */
	set( 0x1b2, 0x28b); /* 'Ʋ' -> 'ʋ' */
	set( 0x1b3, 0x1b4); /* 'Ƴ' -> 'ƴ' */
	set( 0x1b5, 0x1b6); /* 'Ƶ' -> 'ƶ' */
	set( 0x1b7, 0x292); /* 'Ʒ' -> 'ʒ' */
	set( 0x1b8, 0x1b9); /* 'Ƹ' -> 'ƹ' */
	set( 0x1bc, 0x1bd); /* 'Ƽ' -> 'ƽ' */
	set( 0x1c4, 0x1c6); /* 'Ǆ' -> 'ǆ' */
	set( 0x1c5, 0x1c6); /* 'ǅ' -> 'ǆ' */
	set( 0x1c7, 0x1c9); /* 'Ǉ' -> 'ǉ' */
	set( 0x1c8, 0x1c9); /* 'ǈ' -> 'ǉ' */
	set( 0x1ca, 0x1cc); /* 'Ǌ' -> 'ǌ' */
	set( 0x1cb, 0x1cc); /* 'ǋ' -> 'ǌ' */
	set( 0x1cd, 0x1ce); /* 'Ǎ' -> 'ǎ' */
	set( 0x1cf, 0x1d0); /* 'Ǐ' -> 'ǐ' */
	set( 0x1d1, 0x1d2); /* 'Ǒ' -> 'ǒ' */
	set( 0x1d3, 0x1d4); /* 'Ǔ' -> 'ǔ' */
	set( 0x1d5, 0x1d6); /* 'Ǖ' -> 'ǖ' */
	set( 0x1d7, 0x1d8); /* 'Ǘ' -> 'ǘ' */
	set( 0x1d9, 0x1da); /* 'Ǚ' -> 'ǚ' */
	set( 0x1db, 0x1dc); /* 'Ǜ' -> 'ǜ' */
	set( 0x1de, 0x1df); /* 'Ǟ' -> 'ǟ' */
	set( 0x1e0, 0x1e1); /* 'Ǡ' -> 'ǡ' */
	set( 0x1e2, 0x1e3); /* 'Ǣ' -> 'ǣ' */
	set( 0x1e4, 0x1e5); /* 'Ǥ' -> 'ǥ' */
	set( 0x1e6, 0x1e7); /* 'Ǧ' -> 'ǧ' */
	set( 0x1e8, 0x1e9); /* 'Ǩ' -> 'ǩ' */
	set( 0x1ea, 0x1eb); /* 'Ǫ' -> 'ǫ' */
	set( 0x1ec, 0x1ed); /* 'Ǭ' -> 'ǭ' */
	set( 0x1ee, 0x1ef); /* 'Ǯ' -> 'ǯ' */
	set( 0x1f1, 0x1f3); /* 'Ǳ' -> 'ǳ' */
	set( 0x1f2, 0x1f3); /* 'ǲ' -> 'ǳ' */
	set( 0x1f4, 0x1f5); /* 'Ǵ' -> 'ǵ' */
	set( 0x1f6, 0x195); /* 'Ƕ' -> 'ƕ' */
	set( 0x1f7, 0x1bf); /* 'Ƿ' -> 'ƿ' */
	set( 0x1f8, 0x1f9); /* 'Ǹ' -> 'ǹ' */
	set( 0x1fa, 0x1fb); /* 'Ǻ' -> 'ǻ' */
	set( 0x1fc, 0x1fd); /* 'Ǽ' -> 'ǽ' */
	set( 0x1fe, 0x1ff); /* 'Ǿ' -> 'ǿ' */
	set( 0x200, 0x201); /* 'Ȁ' -> 'ȁ' */
	set( 0x202, 0x203); /* 'Ȃ' -> 'ȃ' */
	set( 0x204, 0x205); /* 'Ȅ' -> 'ȅ' */
	set( 0x206, 0x207); /* 'Ȇ' -> 'ȇ' */
	set( 0x208, 0x209); /* 'Ȉ' -> 'ȉ' */
	set( 0x20a, 0x20b); /* 'Ȋ' -> 'ȋ' */
	set( 0x20c, 0x20d); /* 'Ȍ' -> 'ȍ' */
	set( 0x20e, 0x20f); /* 'Ȏ' -> 'ȏ' */
	set( 0x210, 0x211); /* 'Ȑ' -> 'ȑ' */
	set( 0x212, 0x213); /* 'Ȓ' -> 'ȓ' */
	set( 0x214, 0x215); /* 'Ȕ' -> 'ȕ' */
	set( 0x216, 0x217); /* 'Ȗ' -> 'ȗ' */
	set( 0x218, 0x219); /* 'Ș' -> 'ș' */
	set( 0x21a, 0x21b); /* 'Ț' -> 'ț' */
	set( 0x21c, 0x21d); /* 'Ȝ' -> 'ȝ' */
	set( 0x21e, 0x21f); /* 'Ȟ' -> 'ȟ' */
	set( 0x220, 0x19e); /* 'Ƞ' -> 'ƞ' */
	set( 0x222, 0x223); /* 'Ȣ' -> 'ȣ' */
	set( 0x224, 0x225); /* 'Ȥ' -> 'ȥ' */
	set( 0x226, 0x227); /* 'Ȧ' -> 'ȧ' */
	set( 0x228, 0x229); /* 'Ȩ' -> 'ȩ' */
	set( 0x22a, 0x22b); /* 'Ȫ' -> 'ȫ' */
	set( 0x22c, 0x22d); /* 'Ȭ' -> 'ȭ' */
	set( 0x22e, 0x22f); /* 'Ȯ' -> 'ȯ' */
	set( 0x230, 0x231); /* 'Ȱ' -> 'ȱ' */
	set( 0x232, 0x233); /* 'Ȳ' -> 'ȳ' */
	set( 0x23a, 0x2c65); /* 'Ⱥ' -> 'ⱥ' */
	set( 0x23b, 0x23c); /* 'Ȼ' -> 'ȼ' */
	set( 0x23d, 0x19a); /* 'Ƚ' -> 'ƚ' */
	set( 0x23e, 0x2c66); /* 'Ⱦ' -> 'ⱦ' */
	set( 0x241, 0x242); /* 'Ɂ' -> 'ɂ' */
	set( 0x243, 0x180); /* 'Ƀ' -> 'ƀ' */
	set( 0x244, 0x289); /* 'Ʉ' -> 'ʉ' */
	set( 0x245, 0x28c); /* 'Ʌ' -> 'ʌ' */
	set( 0x246, 0x247); /* 'Ɇ' -> 'ɇ' */
	set( 0x248, 0x249); /* 'Ɉ' -> 'ɉ' */
	set( 0x24a, 0x24b); /* 'Ɋ' -> 'ɋ' */
	set( 0x24c, 0x24d); /* 'Ɍ' -> 'ɍ' */
	set( 0x24e, 0x24f); /* 'Ɏ' -> 'ɏ' */
	set( 0x370, 0x371); /* 'Ͱ' -> 'ͱ' */
	set( 0x372, 0x373); /* 'Ͳ' -> 'ͳ' */
	set( 0x376, 0x377); /* 'Ͷ' -> 'ͷ' */
	set( 0x386, 0x3ac); /* 'Ά' -> 'ά' */
	set( 0x388, 0x3ad); /* 'Έ' -> 'έ' */
	set( 0x389, 0x3ae); /* 'Ή' -> 'ή' */
	set( 0x38a, 0x3af); /* 'Ί' -> 'ί' */
	set( 0x38c, 0x3cc); /* 'Ό' -> 'ό' */
	set( 0x38e, 0x3cd); /* 'Ύ' -> 'ύ' */
	set( 0x38f, 0x3ce); /* 'Ώ' -> 'ώ' */
	set( 0x391, 0x3b1); /* 'Α' -> 'α' */
	set( 0x392, 0x3b2); /* 'Β' -> 'β' */
	set( 0x393, 0x3b3); /* 'Γ' -> 'γ' */
	set( 0x394, 0x3b4); /* 'Δ' -> 'δ' */
	set( 0x395, 0x3b5); /* 'Ε' -> 'ε' */
	set( 0x396, 0x3b6); /* 'Ζ' -> 'ζ' */
	set( 0x397, 0x3b7); /* 'Η' -> 'η' */
	set( 0x398, 0x3b8); /* 'Θ' -> 'θ' */
	set( 0x399, 0x3b9); /* 'Ι' -> 'ι' */
	set( 0x39a, 0x3ba); /* 'Κ' -> 'κ' */
	set( 0x39b, 0x3bb); /* 'Λ' -> 'λ' */
	set( 0x39c, 0x3bc); /* 'Μ' -> 'μ' */
	set( 0x39d, 0x3bd); /* 'Ν' -> 'ν' */
	set( 0x39e, 0x3be); /* 'Ξ' -> 'ξ' */
	set( 0x39f, 0x3bf); /* 'Ο' -> 'ο' */
	set( 0x3a0, 0x3c0); /* 'Π' -> 'π' */
	set( 0x3a1, 0x3c1); /* 'Ρ' -> 'ρ' */
	set( 0x3a3, 0x3c3); /* 'Σ' -> 'σ' */
	set( 0x3a4, 0x3c4); /* 'Τ' -> 'τ' */
	set( 0x3a5, 0x3c5); /* 'Υ' -> 'υ' */
	set( 0x3a6, 0x3c6); /* 'Φ' -> 'φ' */
	set( 0x3a7, 0x3c7); /* 'Χ' -> 'χ' */
	set( 0x3a8, 0x3c8); /* 'Ψ' -> 'ψ' */
	set( 0x3a9, 0x3c9); /* 'Ω' -> 'ω' */
	set( 0x3aa, 0x3ca); /* 'Ϊ' -> 'ϊ' */
	set( 0x3ab, 0x3cb); /* 'Ϋ' -> 'ϋ' */
	set( 0x3cf, 0x3d7); /* 'Ϗ' -> 'ϗ' */
	set( 0x3d8, 0x3d9); /* 'Ϙ' -> 'ϙ' */
	set( 0x3da, 0x3db); /* 'Ϛ' -> 'ϛ' */
	set( 0x3dc, 0x3dd); /* 'Ϝ' -> 'ϝ' */
	set( 0x3de, 0x3df); /* 'Ϟ' -> 'ϟ' */
	set( 0x3e0, 0x3e1); /* 'Ϡ' -> 'ϡ' */
	set( 0x3e2, 0x3e3); /* 'Ϣ' -> 'ϣ' */
	set( 0x3e4, 0x3e5); /* 'Ϥ' -> 'ϥ' */
	set( 0x3e6, 0x3e7); /* 'Ϧ' -> 'ϧ' */
	set( 0x3e8, 0x3e9); /* 'Ϩ' -> 'ϩ' */
	set( 0x3ea, 0x3eb); /* 'Ϫ' -> 'ϫ' */
	set( 0x3ec, 0x3ed); /* 'Ϭ' -> 'ϭ' */
	set( 0x3ee, 0x3ef); /* 'Ϯ' -> 'ϯ' */
	set( 0x3f4, 0x3b8); /* 'ϴ' -> 'θ' */
	set( 0x3f7, 0x3f8); /* 'Ϸ' -> 'ϸ' */
	set( 0x3f9, 0x3f2); /* 'Ϲ' -> 'ϲ' */
	set( 0x3fa, 0x3fb); /* 'Ϻ' -> 'ϻ' */
	set( 0x3fd, 0x37b); /* 'Ͻ' -> 'ͻ' */
	set( 0x3fe, 0x37c); /* 'Ͼ' -> 'ͼ' */
	set( 0x3ff, 0x37d); /* 'Ͽ' -> 'ͽ' */
	set( 0x400, 0x450); /* 'Ѐ' -> 'ѐ' */
	set( 0x401, 0x451); /* 'Ё' -> 'ё' */
	set( 0x402, 0x452); /* 'Ђ' -> 'ђ' */
	set( 0x403, 0x453); /* 'Ѓ' -> 'ѓ' */
	set( 0x404, 0x454); /* 'Є' -> 'є' */
	set( 0x405, 0x455); /* 'Ѕ' -> 'ѕ' */
	set( 0x406, 0x456); /* 'І' -> 'і' */
	set( 0x407, 0x457); /* 'Ї' -> 'ї' */
	set( 0x408, 0x458); /* 'Ј' -> 'ј' */
	set( 0x409, 0x459); /* 'Љ' -> 'љ' */
	set( 0x40a, 0x45a); /* 'Њ' -> 'њ' */
	set( 0x40b, 0x45b); /* 'Ћ' -> 'ћ' */
	set( 0x40c, 0x45c); /* 'Ќ' -> 'ќ' */
	set( 0x40d, 0x45d); /* 'Ѝ' -> 'ѝ' */
	set( 0x40e, 0x45e); /* 'Ў' -> 'ў' */
	set( 0x40f, 0x45f); /* 'Џ' -> 'џ' */
	set( 0x410, 0x430); /* 'А' -> 'а' */
	set( 0x411, 0x431); /* 'Б' -> 'б' */
	set( 0x412, 0x432); /* 'В' -> 'в' */
	set( 0x413, 0x433); /* 'Г' -> 'г' */
	set( 0x414, 0x434); /* 'Д' -> 'д' */
	set( 0x415, 0x435); /* 'Е' -> 'е' */
	set( 0x416, 0x436); /* 'Ж' -> 'ж' */
	set( 0x417, 0x437); /* 'З' -> 'з' */
	set( 0x418, 0x438); /* 'И' -> 'и' */
	set( 0x419, 0x439); /* 'Й' -> 'й' */
	set( 0x41a, 0x43a); /* 'К' -> 'к' */
	set( 0x41b, 0x43b); /* 'Л' -> 'л' */
	set( 0x41c, 0x43c); /* 'М' -> 'м' */
	set( 0x41d, 0x43d); /* 'Н' -> 'н' */
	set( 0x41e, 0x43e); /* 'О' -> 'о' */
	set( 0x41f, 0x43f); /* 'П' -> 'п' */
	set( 0x420, 0x440); /* 'Р' -> 'р' */
	set( 0x421, 0x441); /* 'С' -> 'с' */
	set( 0x422, 0x442); /* 'Т' -> 'т' */
	set( 0x423, 0x443); /* 'У' -> 'у' */
	set( 0x424, 0x444); /* 'Ф' -> 'ф' */
	set( 0x425, 0x445); /* 'Х' -> 'х' */
	set( 0x426, 0x446); /* 'Ц' -> 'ц' */
	set( 0x427, 0x447); /* 'Ч' -> 'ч' */
	set( 0x428, 0x448); /* 'Ш' -> 'ш' */
	set( 0x429, 0x449); /* 'Щ' -> 'щ' */
	set( 0x42a, 0x44a); /* 'Ъ' -> 'ъ' */
	set( 0x42b, 0x44b); /* 'Ы' -> 'ы' */
	set( 0x42c, 0x44c); /* 'Ь' -> 'ь' */
	set( 0x42d, 0x44d); /* 'Э' -> 'э' */
	set( 0x42e, 0x44e); /* 'Ю' -> 'ю' */
	set( 0x42f, 0x44f); /* 'Я' -> 'я' */
	set( 0x460, 0x461); /* 'Ѡ' -> 'ѡ' */
	set( 0x462, 0x463); /* 'Ѣ' -> 'ѣ' */
	set( 0x464, 0x465); /* 'Ѥ' -> 'ѥ' */
	set( 0x466, 0x467); /* 'Ѧ' -> 'ѧ' */
	set( 0x468, 0x469); /* 'Ѩ' -> 'ѩ' */
	set( 0x46a, 0x46b); /* 'Ѫ' -> 'ѫ' */
	set( 0x46c, 0x46d); /* 'Ѭ' -> 'ѭ' */
	set( 0x46e, 0x46f); /* 'Ѯ' -> 'ѯ' */
	set( 0x470, 0x471); /* 'Ѱ' -> 'ѱ' */
	set( 0x472, 0x473); /* 'Ѳ' -> 'ѳ' */
	set( 0x474, 0x475); /* 'Ѵ' -> 'ѵ' */
	set( 0x476, 0x477); /* 'Ѷ' -> 'ѷ' */
	set( 0x478, 0x479); /* 'Ѹ' -> 'ѹ' */
	set( 0x47a, 0x47b); /* 'Ѻ' -> 'ѻ' */
	set( 0x47c, 0x47d); /* 'Ѽ' -> 'ѽ' */
	set( 0x47e, 0x47f); /* 'Ѿ' -> 'ѿ' */
	set( 0x480, 0x481); /* 'Ҁ' -> 'ҁ' */
	set( 0x48a, 0x48b); /* 'Ҋ' -> 'ҋ' */
	set( 0x48c, 0x48d); /* 'Ҍ' -> 'ҍ' */
	set( 0x48e, 0x48f); /* 'Ҏ' -> 'ҏ' */
	set( 0x490, 0x491); /* 'Ґ' -> 'ґ' */
	set( 0x492, 0x493); /* 'Ғ' -> 'ғ' */
	set( 0x494, 0x495); /* 'Ҕ' -> 'ҕ' */
	set( 0x496, 0x497); /* 'Җ' -> 'җ' */
	set( 0x498, 0x499); /* 'Ҙ' -> 'ҙ' */
	set( 0x49a, 0x49b); /* 'Қ' -> 'қ' */
	set( 0x49c, 0x49d); /* 'Ҝ' -> 'ҝ' */
	set( 0x49e, 0x49f); /* 'Ҟ' -> 'ҟ' */
	set( 0x4a0, 0x4a1); /* 'Ҡ' -> 'ҡ' */
	set( 0x4a2, 0x4a3); /* 'Ң' -> 'ң' */
	set( 0x4a4, 0x4a5); /* 'Ҥ' -> 'ҥ' */
	set( 0x4a6, 0x4a7); /* 'Ҧ' -> 'ҧ' */
	set( 0x4a8, 0x4a9); /* 'Ҩ' -> 'ҩ' */
	set( 0x4aa, 0x4ab); /* 'Ҫ' -> 'ҫ' */
	set( 0x4ac, 0x4ad); /* 'Ҭ' -> 'ҭ' */
	set( 0x4ae, 0x4af); /* 'Ү' -> 'ү' */
	set( 0x4b0, 0x4b1); /* 'Ұ' -> 'ұ' */
	set( 0x4b2, 0x4b3); /* 'Ҳ' -> 'ҳ' */
	set( 0x4b4, 0x4b5); /* 'Ҵ' -> 'ҵ' */
	set( 0x4b6, 0x4b7); /* 'Ҷ' -> 'ҷ' */
	set( 0x4b8, 0x4b9); /* 'Ҹ' -> 'ҹ' */
	set( 0x4ba, 0x4bb); /* 'Һ' -> 'һ' */
	set( 0x4bc, 0x4bd); /* 'Ҽ' -> 'ҽ' */
	set( 0x4be, 0x4bf); /* 'Ҿ' -> 'ҿ' */
	set( 0x4c0, 0x4cf); /* 'Ӏ' -> 'ӏ' */
	set( 0x4c1, 0x4c2); /* 'Ӂ' -> 'ӂ' */
	set( 0x4c3, 0x4c4); /* 'Ӄ' -> 'ӄ' */
	set( 0x4c5, 0x4c6); /* 'Ӆ' -> 'ӆ' */
	set( 0x4c7, 0x4c8); /* 'Ӈ' -> 'ӈ' */
	set( 0x4c9, 0x4ca); /* 'Ӊ' -> 'ӊ' */
	set( 0x4cb, 0x4cc); /* 'Ӌ' -> 'ӌ' */
	set( 0x4cd, 0x4ce); /* 'Ӎ' -> 'ӎ' */
	set( 0x4d0, 0x4d1); /* 'Ӑ' -> 'ӑ' */
	set( 0x4d2, 0x4d3); /* 'Ӓ' -> 'ӓ' */
	set( 0x4d4, 0x4d5); /* 'Ӕ' -> 'ӕ' */
	set( 0x4d6, 0x4d7); /* 'Ӗ' -> 'ӗ' */
	set( 0x4d8, 0x4d9); /* 'Ә' -> 'ә' */
	set( 0x4da, 0x4db); /* 'Ӛ' -> 'ӛ' */
	set( 0x4dc, 0x4dd); /* 'Ӝ' -> 'ӝ' */
	set( 0x4de, 0x4df); /* 'Ӟ' -> 'ӟ' */
	set( 0x4e0, 0x4e1); /* 'Ӡ' -> 'ӡ' */
	set( 0x4e2, 0x4e3); /* 'Ӣ' -> 'ӣ' */
	set( 0x4e4, 0x4e5); /* 'Ӥ' -> 'ӥ' */
	set( 0x4e6, 0x4e7); /* 'Ӧ' -> 'ӧ' */
	set( 0x4e8, 0x4e9); /* 'Ө' -> 'ө' */
	set( 0x4ea, 0x4eb); /* 'Ӫ' -> 'ӫ' */
	set( 0x4ec, 0x4ed); /* 'Ӭ' -> 'ӭ' */
	set( 0x4ee, 0x4ef); /* 'Ӯ' -> 'ӯ' */
	set( 0x4f0, 0x4f1); /* 'Ӱ' -> 'ӱ' */
	set( 0x4f2, 0x4f3); /* 'Ӳ' -> 'ӳ' */
	set( 0x4f4, 0x4f5); /* 'Ӵ' -> 'ӵ' */
	set( 0x4f6, 0x4f7); /* 'Ӷ' -> 'ӷ' */
	set( 0x4f8, 0x4f9); /* 'Ӹ' -> 'ӹ' */
	set( 0x4fa, 0x4fb); /* 'Ӻ' -> 'ӻ' */
	set( 0x4fc, 0x4fd); /* 'Ӽ' -> 'ӽ' */
	set( 0x4fe, 0x4ff); /* 'Ӿ' -> 'ӿ' */
	set( 0x500, 0x501); /* 'Ԁ' -> 'ԁ' */
	set( 0x502, 0x503); /* 'Ԃ' -> 'ԃ' */
	set( 0x504, 0x505); /* 'Ԅ' -> 'ԅ' */
	set( 0x506, 0x507); /* 'Ԇ' -> 'ԇ' */
	set( 0x508, 0x509); /* 'Ԉ' -> 'ԉ' */
	set( 0x50a, 0x50b); /* 'Ԋ' -> 'ԋ' */
	set( 0x50c, 0x50d); /* 'Ԍ' -> 'ԍ' */
	set( 0x50e, 0x50f); /* 'Ԏ' -> 'ԏ' */
	set( 0x510, 0x511); /* 'Ԑ' -> 'ԑ' */
	set( 0x512, 0x513); /* 'Ԓ' -> 'ԓ' */
	set( 0x514, 0x515); /* 'Ԕ' -> 'ԕ' */
	set( 0x516, 0x517); /* 'Ԗ' -> 'ԗ' */
	set( 0x518, 0x519); /* 'Ԙ' -> 'ԙ' */
	set( 0x51a, 0x51b); /* 'Ԛ' -> 'ԛ' */
	set( 0x51c, 0x51d); /* 'Ԝ' -> 'ԝ' */
	set( 0x51e, 0x51f); /* 'Ԟ' -> 'ԟ' */
	set( 0x520, 0x521); /* 'Ԡ' -> 'ԡ' */
	set( 0x522, 0x523); /* 'Ԣ' -> 'ԣ' */
	set( 0x524, 0x525); /* 'Ԥ' -> 'ԥ' */
	set( 0x531, 0x561); /* 'Ա' -> 'ա' */
	set( 0x532, 0x562); /* 'Բ' -> 'բ' */
	set( 0x533, 0x563); /* 'Գ' -> 'գ' */
	set( 0x534, 0x564); /* 'Դ' -> 'դ' */
	set( 0x535, 0x565); /* 'Ե' -> 'ե' */
	set( 0x536, 0x566); /* 'Զ' -> 'զ' */
	set( 0x537, 0x567); /* 'Է' -> 'է' */
	set( 0x538, 0x568); /* 'Ը' -> 'ը' */
	set( 0x539, 0x569); /* 'Թ' -> 'թ' */
	set( 0x53a, 0x56a); /* 'Ժ' -> 'ժ' */
	set( 0x53b, 0x56b); /* 'Ի' -> 'ի' */
	set( 0x53c, 0x56c); /* 'Լ' -> 'լ' */
	set( 0x53d, 0x56d); /* 'Խ' -> 'խ' */
	set( 0x53e, 0x56e); /* 'Ծ' -> 'ծ' */
	set( 0x53f, 0x56f); /* 'Կ' -> 'կ' */
	set( 0x540, 0x570); /* 'Հ' -> 'հ' */
	set( 0x541, 0x571); /* 'Ձ' -> 'ձ' */
	set( 0x542, 0x572); /* 'Ղ' -> 'ղ' */
	set( 0x543, 0x573); /* 'Ճ' -> 'ճ' */
	set( 0x544, 0x574); /* 'Մ' -> 'մ' */
	set( 0x545, 0x575); /* 'Յ' -> 'յ' */
	set( 0x546, 0x576); /* 'Ն' -> 'ն' */
	set( 0x547, 0x577); /* 'Շ' -> 'շ' */
	set( 0x548, 0x578); /* 'Ո' -> 'ո' */
	set( 0x549, 0x579); /* 'Չ' -> 'չ' */
	set( 0x54a, 0x57a); /* 'Պ' -> 'պ' */
	set( 0x54b, 0x57b); /* 'Ջ' -> 'ջ' */
	set( 0x54c, 0x57c); /* 'Ռ' -> 'ռ' */
	set( 0x54d, 0x57d); /* 'Ս' -> 'ս' */
	set( 0x54e, 0x57e); /* 'Վ' -> 'վ' */
	set( 0x54f, 0x57f); /* 'Տ' -> 'տ' */
	set( 0x550, 0x580); /* 'Ր' -> 'ր' */
	set( 0x551, 0x581); /* 'Ց' -> 'ց' */
	set( 0x552, 0x582); /* 'Ւ' -> 'ւ' */
	set( 0x553, 0x583); /* 'Փ' -> 'փ' */
	set( 0x554, 0x584); /* 'Ք' -> 'ք' */
	set( 0x555, 0x585); /* 'Օ' -> 'օ' */
	set( 0x556, 0x586); /* 'Ֆ' -> 'ֆ' */
	set( 0x10a0, 0x2d00); /* 'Ⴀ' -> 'ⴀ' */
	set( 0x10a1, 0x2d01); /* 'Ⴁ' -> 'ⴁ' */
	set( 0x10a2, 0x2d02); /* 'Ⴂ' -> 'ⴂ' */
	set( 0x10a3, 0x2d03); /* 'Ⴃ' -> 'ⴃ' */
	set( 0x10a4, 0x2d04); /* 'Ⴄ' -> 'ⴄ' */
	set( 0x10a5, 0x2d05); /* 'Ⴅ' -> 'ⴅ' */
	set( 0x10a6, 0x2d06); /* 'Ⴆ' -> 'ⴆ' */
	set( 0x10a7, 0x2d07); /* 'Ⴇ' -> 'ⴇ' */
	set( 0x10a8, 0x2d08); /* 'Ⴈ' -> 'ⴈ' */
	set( 0x10a9, 0x2d09); /* 'Ⴉ' -> 'ⴉ' */
	set( 0x10aa, 0x2d0a); /* 'Ⴊ' -> 'ⴊ' */
	set( 0x10ab, 0x2d0b); /* 'Ⴋ' -> 'ⴋ' */
	set( 0x10ac, 0x2d0c); /* 'Ⴌ' -> 'ⴌ' */
	set( 0x10ad, 0x2d0d); /* 'Ⴍ' -> 'ⴍ' */
	set( 0x10ae, 0x2d0e); /* 'Ⴎ' -> 'ⴎ' */
	set( 0x10af, 0x2d0f); /* 'Ⴏ' -> 'ⴏ' */
	set( 0x10b0, 0x2d10); /* 'Ⴐ' -> 'ⴐ' */
	set( 0x10b1, 0x2d11); /* 'Ⴑ' -> 'ⴑ' */
	set( 0x10b2, 0x2d12); /* 'Ⴒ' -> 'ⴒ' */
	set( 0x10b3, 0x2d13); /* 'Ⴓ' -> 'ⴓ' */
	set( 0x10b4, 0x2d14); /* 'Ⴔ' -> 'ⴔ' */
	set( 0x10b5, 0x2d15); /* 'Ⴕ' -> 'ⴕ' */
	set( 0x10b6, 0x2d16); /* 'Ⴖ' -> 'ⴖ' */
	set( 0x10b7, 0x2d17); /* 'Ⴗ' -> 'ⴗ' */
	set( 0x10b8, 0x2d18); /* 'Ⴘ' -> 'ⴘ' */
	set( 0x10b9, 0x2d19); /* 'Ⴙ' -> 'ⴙ' */
	set( 0x10ba, 0x2d1a); /* 'Ⴚ' -> 'ⴚ' */
	set( 0x10bb, 0x2d1b); /* 'Ⴛ' -> 'ⴛ' */
	set( 0x10bc, 0x2d1c); /* 'Ⴜ' -> 'ⴜ' */
	set( 0x10bd, 0x2d1d); /* 'Ⴝ' -> 'ⴝ' */
	set( 0x10be, 0x2d1e); /* 'Ⴞ' -> 'ⴞ' */
	set( 0x10bf, 0x2d1f); /* 'Ⴟ' -> 'ⴟ' */
	set( 0x10c0, 0x2d20); /* 'Ⴠ' -> 'ⴠ' */
	set( 0x10c1, 0x2d21); /* 'Ⴡ' -> 'ⴡ' */
	set( 0x10c2, 0x2d22); /* 'Ⴢ' -> 'ⴢ' */
	set( 0x10c3, 0x2d23); /* 'Ⴣ' -> 'ⴣ' */
	set( 0x10c4, 0x2d24); /* 'Ⴤ' -> 'ⴤ' */
	set( 0x10c5, 0x2d25); /* 'Ⴥ' -> 'ⴥ' */
	set( 0x1e00, 0x1e01); /* 'Ḁ' -> 'ḁ' */
	set( 0x1e02, 0x1e03); /* 'Ḃ' -> 'ḃ' */
	set( 0x1e04, 0x1e05); /* 'Ḅ' -> 'ḅ' */
	set( 0x1e06, 0x1e07); /* 'Ḇ' -> 'ḇ' */
	set( 0x1e08, 0x1e09); /* 'Ḉ' -> 'ḉ' */
	set( 0x1e0a, 0x1e0b); /* 'Ḋ' -> 'ḋ' */
	set( 0x1e0c, 0x1e0d); /* 'Ḍ' -> 'ḍ' */
	set( 0x1e0e, 0x1e0f); /* 'Ḏ' -> 'ḏ' */
	set( 0x1e10, 0x1e11); /* 'Ḑ' -> 'ḑ' */
	set( 0x1e12, 0x1e13); /* 'Ḓ' -> 'ḓ' */
	set( 0x1e14, 0x1e15); /* 'Ḕ' -> 'ḕ' */
	set( 0x1e16, 0x1e17); /* 'Ḗ' -> 'ḗ' */
	set( 0x1e18, 0x1e19); /* 'Ḙ' -> 'ḙ' */
	set( 0x1e1a, 0x1e1b); /* 'Ḛ' -> 'ḛ' */
	set( 0x1e1c, 0x1e1d); /* 'Ḝ' -> 'ḝ' */
	set( 0x1e1e, 0x1e1f); /* 'Ḟ' -> 'ḟ' */
	set( 0x1e20, 0x1e21); /* 'Ḡ' -> 'ḡ' */
	set( 0x1e22, 0x1e23); /* 'Ḣ' -> 'ḣ' */
	set( 0x1e24, 0x1e25); /* 'Ḥ' -> 'ḥ' */
	set( 0x1e26, 0x1e27); /* 'Ḧ' -> 'ḧ' */
	set( 0x1e28, 0x1e29); /* 'Ḩ' -> 'ḩ' */
	set( 0x1e2a, 0x1e2b); /* 'Ḫ' -> 'ḫ' */
	set( 0x1e2c, 0x1e2d); /* 'Ḭ' -> 'ḭ' */
	set( 0x1e2e, 0x1e2f); /* 'Ḯ' -> 'ḯ' */
	set( 0x1e30, 0x1e31); /* 'Ḱ' -> 'ḱ' */
	set( 0x1e32, 0x1e33); /* 'Ḳ' -> 'ḳ' */
	set( 0x1e34, 0x1e35); /* 'Ḵ' -> 'ḵ' */
	set( 0x1e36, 0x1e37); /* 'Ḷ' -> 'ḷ' */
	set( 0x1e38, 0x1e39); /* 'Ḹ' -> 'ḹ' */
	set( 0x1e3a, 0x1e3b); /* 'Ḻ' -> 'ḻ' */
	set( 0x1e3c, 0x1e3d); /* 'Ḽ' -> 'ḽ' */
	set( 0x1e3e, 0x1e3f); /* 'Ḿ' -> 'ḿ' */
	set( 0x1e40, 0x1e41); /* 'Ṁ' -> 'ṁ' */
	set( 0x1e42, 0x1e43); /* 'Ṃ' -> 'ṃ' */
	set( 0x1e44, 0x1e45); /* 'Ṅ' -> 'ṅ' */
	set( 0x1e46, 0x1e47); /* 'Ṇ' -> 'ṇ' */
	set( 0x1e48, 0x1e49); /* 'Ṉ' -> 'ṉ' */
	set( 0x1e4a, 0x1e4b); /* 'Ṋ' -> 'ṋ' */
	set( 0x1e4c, 0x1e4d); /* 'Ṍ' -> 'ṍ' */
	set( 0x1e4e, 0x1e4f); /* 'Ṏ' -> 'ṏ' */
	set( 0x1e50, 0x1e51); /* 'Ṑ' -> 'ṑ' */
	set( 0x1e52, 0x1e53); /* 'Ṓ' -> 'ṓ' */
	set( 0x1e54, 0x1e55); /* 'Ṕ' -> 'ṕ' */
	set( 0x1e56, 0x1e57); /* 'Ṗ' -> 'ṗ' */
	set( 0x1e58, 0x1e59); /* 'Ṙ' -> 'ṙ' */
	set( 0x1e5a, 0x1e5b); /* 'Ṛ' -> 'ṛ' */
	set( 0x1e5c, 0x1e5d); /* 'Ṝ' -> 'ṝ' */
	set( 0x1e5e, 0x1e5f); /* 'Ṟ' -> 'ṟ' */
	set( 0x1e60, 0x1e61); /* 'Ṡ' -> 'ṡ' */
	set( 0x1e62, 0x1e63); /* 'Ṣ' -> 'ṣ' */
	set( 0x1e64, 0x1e65); /* 'Ṥ' -> 'ṥ' */
	set( 0x1e66, 0x1e67); /* 'Ṧ' -> 'ṧ' */
	set( 0x1e68, 0x1e69); /* 'Ṩ' -> 'ṩ' */
	set( 0x1e6a, 0x1e6b); /* 'Ṫ' -> 'ṫ' */
	set( 0x1e6c, 0x1e6d); /* 'Ṭ' -> 'ṭ' */
	set( 0x1e6e, 0x1e6f); /* 'Ṯ' -> 'ṯ' */
	set( 0x1e70, 0x1e71); /* 'Ṱ' -> 'ṱ' */
	set( 0x1e72, 0x1e73); /* 'Ṳ' -> 'ṳ' */
	set( 0x1e74, 0x1e75); /* 'Ṵ' -> 'ṵ' */
	set( 0x1e76, 0x1e77); /* 'Ṷ' -> 'ṷ' */
	set( 0x1e78, 0x1e79); /* 'Ṹ' -> 'ṹ' */
	set( 0x1e7a, 0x1e7b); /* 'Ṻ' -> 'ṻ' */
	set( 0x1e7c, 0x1e7d); /* 'Ṽ' -> 'ṽ' */
	set( 0x1e7e, 0x1e7f); /* 'Ṿ' -> 'ṿ' */
	set( 0x1e80, 0x1e81); /* 'Ẁ' -> 'ẁ' */
	set( 0x1e82, 0x1e83); /* 'Ẃ' -> 'ẃ' */
	set( 0x1e84, 0x1e85); /* 'Ẅ' -> 'ẅ' */
	set( 0x1e86, 0x1e87); /* 'Ẇ' -> 'ẇ' */
	set( 0x1e88, 0x1e89); /* 'Ẉ' -> 'ẉ' */
	set( 0x1e8a, 0x1e8b); /* 'Ẋ' -> 'ẋ' */
	set( 0x1e8c, 0x1e8d); /* 'Ẍ' -> 'ẍ' */
	set( 0x1e8e, 0x1e8f); /* 'Ẏ' -> 'ẏ' */
	set( 0x1e90, 0x1e91); /* 'Ẑ' -> 'ẑ' */
	set( 0x1e92, 0x1e93); /* 'Ẓ' -> 'ẓ' */
	set( 0x1e94, 0x1e95); /* 'Ẕ' -> 'ẕ' */
	set( 0x1e9e, 0xdf); /* 'ẞ' -> 'ß' */
	set( 0x1ea0, 0x1ea1); /* 'Ạ' -> 'ạ' */
	set( 0x1ea2, 0x1ea3); /* 'Ả' -> 'ả' */
	set( 0x1ea4, 0x1ea5); /* 'Ấ' -> 'ấ' */
	set( 0x1ea6, 0x1ea7); /* 'Ầ' -> 'ầ' */
	set( 0x1ea8, 0x1ea9); /* 'Ẩ' -> 'ẩ' */
	set( 0x1eaa, 0x1eab); /* 'Ẫ' -> 'ẫ' */
	set( 0x1eac, 0x1ead); /* 'Ậ' -> 'ậ' */
	set( 0x1eae, 0x1eaf); /* 'Ắ' -> 'ắ' */
	set( 0x1eb0, 0x1eb1); /* 'Ằ' -> 'ằ' */
	set( 0x1eb2, 0x1eb3); /* 'Ẳ' -> 'ẳ' */
	set( 0x1eb4, 0x1eb5); /* 'Ẵ' -> 'ẵ' */
	set( 0x1eb6, 0x1eb7); /* 'Ặ' -> 'ặ' */
	set( 0x1eb8, 0x1eb9); /* 'Ẹ' -> 'ẹ' */
	set( 0x1eba, 0x1ebb); /* 'Ẻ' -> 'ẻ' */
	set( 0x1ebc, 0x1ebd); /* 'Ẽ' -> 'ẽ' */
	set( 0x1ebe, 0x1ebf); /* 'Ế' -> 'ế' */
	set( 0x1ec0, 0x1ec1); /* 'Ề' -> 'ề' */
	set( 0x1ec2, 0x1ec3); /* 'Ể' -> 'ể' */
	set( 0x1ec4, 0x1ec5); /* 'Ễ' -> 'ễ' */
	set( 0x1ec6, 0x1ec7); /* 'Ệ' -> 'ệ' */
	set( 0x1ec8, 0x1ec9); /* 'Ỉ' -> 'ỉ' */
	set( 0x1eca, 0x1ecb); /* 'Ị' -> 'ị' */
	set( 0x1ecc, 0x1ecd); /* 'Ọ' -> 'ọ' */
	set( 0x1ece, 0x1ecf); /* 'Ỏ' -> 'ỏ' */
	set( 0x1ed0, 0x1ed1); /* 'Ố' -> 'ố' */
	set( 0x1ed2, 0x1ed3); /* 'Ồ' -> 'ồ' */
	set( 0x1ed4, 0x1ed5); /* 'Ổ' -> 'ổ' */
	set( 0x1ed6, 0x1ed7); /* 'Ỗ' -> 'ỗ' */
	set( 0x1ed8, 0x1ed9); /* 'Ộ' -> 'ộ' */
	set( 0x1eda, 0x1edb); /* 'Ớ' -> 'ớ' */
	set( 0x1edc, 0x1edd); /* 'Ờ' -> 'ờ' */
	set( 0x1ede, 0x1edf); /* 'Ở' -> 'ở' */
	set( 0x1ee0, 0x1ee1); /* 'Ỡ' -> 'ỡ' */
	set( 0x1ee2, 0x1ee3); /* 'Ợ' -> 'ợ' */
	set( 0x1ee4, 0x1ee5); /* 'Ụ' -> 'ụ' */
	set( 0x1ee6, 0x1ee7); /* 'Ủ' -> 'ủ' */
	set( 0x1ee8, 0x1ee9); /* 'Ứ' -> 'ứ' */
	set( 0x1eea, 0x1eeb); /* 'Ừ' -> 'ừ' */
	set( 0x1eec, 0x1eed); /* 'Ử' -> 'ử' */
	set( 0x1eee, 0x1eef); /* 'Ữ' -> 'ữ' */
	set( 0x1ef0, 0x1ef1); /* 'Ự' -> 'ự' */
	set( 0x1ef2, 0x1ef3); /* 'Ỳ' -> 'ỳ' */
	set( 0x1ef4, 0x1ef5); /* 'Ỵ' -> 'ỵ' */
	set( 0x1ef6, 0x1ef7); /* 'Ỷ' -> 'ỷ' */
	set( 0x1ef8, 0x1ef9); /* 'Ỹ' -> 'ỹ' */
	set( 0x1efa, 0x1efb); /* 'Ỻ' -> 'ỻ' */
	set( 0x1efc, 0x1efd); /* 'Ỽ' -> 'ỽ' */
	set( 0x1efe, 0x1eff); /* 'Ỿ' -> 'ỿ' */
	set( 0x1f08, 0x1f00); /* 'Ἀ' -> 'ἀ' */
	set( 0x1f09, 0x1f01); /* 'Ἁ' -> 'ἁ' */
	set( 0x1f0a, 0x1f02); /* 'Ἂ' -> 'ἂ' */
	set( 0x1f0b, 0x1f03); /* 'Ἃ' -> 'ἃ' */
	set( 0x1f0c, 0x1f04); /* 'Ἄ' -> 'ἄ' */
	set( 0x1f0d, 0x1f05); /* 'Ἅ' -> 'ἅ' */
	set( 0x1f0e, 0x1f06); /* 'Ἆ' -> 'ἆ' */
	set( 0x1f0f, 0x1f07); /* 'Ἇ' -> 'ἇ' */
	set( 0x1f18, 0x1f10); /* 'Ἐ' -> 'ἐ' */
	set( 0x1f19, 0x1f11); /* 'Ἑ' -> 'ἑ' */
	set( 0x1f1a, 0x1f12); /* 'Ἒ' -> 'ἒ' */
	set( 0x1f1b, 0x1f13); /* 'Ἓ' -> 'ἓ' */
	set( 0x1f1c, 0x1f14); /* 'Ἔ' -> 'ἔ' */
	set( 0x1f1d, 0x1f15); /* 'Ἕ' -> 'ἕ' */
	set( 0x1f28, 0x1f20); /* 'Ἠ' -> 'ἠ' */
	set( 0x1f29, 0x1f21); /* 'Ἡ' -> 'ἡ' */
	set( 0x1f2a, 0x1f22); /* 'Ἢ' -> 'ἢ' */
	set( 0x1f2b, 0x1f23); /* 'Ἣ' -> 'ἣ' */
	set( 0x1f2c, 0x1f24); /* 'Ἤ' -> 'ἤ' */
	set( 0x1f2d, 0x1f25); /* 'Ἥ' -> 'ἥ' */
	set( 0x1f2e, 0x1f26); /* 'Ἦ' -> 'ἦ' */
	set( 0x1f2f, 0x1f27); /* 'Ἧ' -> 'ἧ' */
	set( 0x1f38, 0x1f30); /* 'Ἰ' -> 'ἰ' */
	set( 0x1f39, 0x1f31); /* 'Ἱ' -> 'ἱ' */
	set( 0x1f3a, 0x1f32); /* 'Ἲ' -> 'ἲ' */
	set( 0x1f3b, 0x1f33); /* 'Ἳ' -> 'ἳ' */
	set( 0x1f3c, 0x1f34); /* 'Ἴ' -> 'ἴ' */
	set( 0x1f3d, 0x1f35); /* 'Ἵ' -> 'ἵ' */
	set( 0x1f3e, 0x1f36); /* 'Ἶ' -> 'ἶ' */
	set( 0x1f3f, 0x1f37); /* 'Ἷ' -> 'ἷ' */
	set( 0x1f48, 0x1f40); /* 'Ὀ' -> 'ὀ' */
	set( 0x1f49, 0x1f41); /* 'Ὁ' -> 'ὁ' */
	set( 0x1f4a, 0x1f42); /* 'Ὂ' -> 'ὂ' */
	set( 0x1f4b, 0x1f43); /* 'Ὃ' -> 'ὃ' */
	set( 0x1f4c, 0x1f44); /* 'Ὄ' -> 'ὄ' */
	set( 0x1f4d, 0x1f45); /* 'Ὅ' -> 'ὅ' */
	set( 0x1f59, 0x1f51); /* 'Ὑ' -> 'ὑ' */
	set( 0x1f5b, 0x1f53); /* 'Ὓ' -> 'ὓ' */
	set( 0x1f5d, 0x1f55); /* 'Ὕ' -> 'ὕ' */
	set( 0x1f5f, 0x1f57); /* 'Ὗ' -> 'ὗ' */
	set( 0x1f68, 0x1f60); /* 'Ὠ' -> 'ὠ' */
	set( 0x1f69, 0x1f61); /* 'Ὡ' -> 'ὡ' */
	set( 0x1f6a, 0x1f62); /* 'Ὢ' -> 'ὢ' */
	set( 0x1f6b, 0x1f63); /* 'Ὣ' -> 'ὣ' */
	set( 0x1f6c, 0x1f64); /* 'Ὤ' -> 'ὤ' */
	set( 0x1f6d, 0x1f65); /* 'Ὥ' -> 'ὥ' */
	set( 0x1f6e, 0x1f66); /* 'Ὦ' -> 'ὦ' */
	set( 0x1f6f, 0x1f67); /* 'Ὧ' -> 'ὧ' */
	set( 0x1f88, 0x1f80); /* 'ᾈ' -> 'ᾀ' */
	set( 0x1f89, 0x1f81); /* 'ᾉ' -> 'ᾁ' */
	set( 0x1f8a, 0x1f82); /* 'ᾊ' -> 'ᾂ' */
	set( 0x1f8b, 0x1f83); /* 'ᾋ' -> 'ᾃ' */
	set( 0x1f8c, 0x1f84); /* 'ᾌ' -> 'ᾄ' */
	set( 0x1f8d, 0x1f85); /* 'ᾍ' -> 'ᾅ' */
	set( 0x1f8e, 0x1f86); /* 'ᾎ' -> 'ᾆ' */
	set( 0x1f8f, 0x1f87); /* 'ᾏ' -> 'ᾇ' */
	set( 0x1f98, 0x1f90); /* 'ᾘ' -> 'ᾐ' */
	set( 0x1f99, 0x1f91); /* 'ᾙ' -> 'ᾑ' */
	set( 0x1f9a, 0x1f92); /* 'ᾚ' -> 'ᾒ' */
	set( 0x1f9b, 0x1f93); /* 'ᾛ' -> 'ᾓ' */
	set( 0x1f9c, 0x1f94); /* 'ᾜ' -> 'ᾔ' */
	set( 0x1f9d, 0x1f95); /* 'ᾝ' -> 'ᾕ' */
	set( 0x1f9e, 0x1f96); /* 'ᾞ' -> 'ᾖ' */
	set( 0x1f9f, 0x1f97); /* 'ᾟ' -> 'ᾗ' */
	set( 0x1fa8, 0x1fa0); /* 'ᾨ' -> 'ᾠ' */
	set( 0x1fa9, 0x1fa1); /* 'ᾩ' -> 'ᾡ' */
	set( 0x1faa, 0x1fa2); /* 'ᾪ' -> 'ᾢ' */
	set( 0x1fab, 0x1fa3); /* 'ᾫ' -> 'ᾣ' */
	set( 0x1fac, 0x1fa4); /* 'ᾬ' -> 'ᾤ' */
	set( 0x1fad, 0x1fa5); /* 'ᾭ' -> 'ᾥ' */
	set( 0x1fae, 0x1fa6); /* 'ᾮ' -> 'ᾦ' */
	set( 0x1faf, 0x1fa7); /* 'ᾯ' -> 'ᾧ' */
	set( 0x1fb8, 0x1fb0); /* 'Ᾰ' -> 'ᾰ' */
	set( 0x1fb9, 0x1fb1); /* 'Ᾱ' -> 'ᾱ' */
	set( 0x1fba, 0x1f70); /* 'Ὰ' -> 'ὰ' */
	set( 0x1fbb, 0x1f71); /* 'Ά' -> 'ά' */
	set( 0x1fbc, 0x1fb3); /* 'ᾼ' -> 'ᾳ' */
	set( 0x1fc8, 0x1f72); /* 'Ὲ' -> 'ὲ' */
	set( 0x1fc9, 0x1f73); /* 'Έ' -> 'έ' */
	set( 0x1fca, 0x1f74); /* 'Ὴ' -> 'ὴ' */
	set( 0x1fcb, 0x1f75); /* 'Ή' -> 'ή' */
	set( 0x1fcc, 0x1fc3); /* 'ῌ' -> 'ῃ' */
	set( 0x1fd8, 0x1fd0); /* 'Ῐ' -> 'ῐ' */
	set( 0x1fd9, 0x1fd1); /* 'Ῑ' -> 'ῑ' */
	set( 0x1fda, 0x1f76); /* 'Ὶ' -> 'ὶ' */
	set( 0x1fdb, 0x1f77); /* 'Ί' -> 'ί' */
	set( 0x1fe8, 0x1fe0); /* 'Ῠ' -> 'ῠ' */
	set( 0x1fe9, 0x1fe1); /* 'Ῡ' -> 'ῡ' */
	set( 0x1fea, 0x1f7a); /* 'Ὺ' -> 'ὺ' */
	set( 0x1feb, 0x1f7b); /* 'Ύ' -> 'ύ' */
	set( 0x1fec, 0x1fe5); /* 'Ῥ' -> 'ῥ' */
	set( 0x1ff8, 0x1f78); /* 'Ὸ' -> 'ὸ' */
	set( 0x1ff9, 0x1f79); /* 'Ό' -> 'ό' */
	set( 0x1ffa, 0x1f7c); /* 'Ὼ' -> 'ὼ' */
	set( 0x1ffb, 0x1f7d); /* 'Ώ' -> 'ώ' */
	set( 0x1ffc, 0x1ff3); /* 'ῼ' -> 'ῳ' */
	set( 0x2126, 0x3c9); /* 'Ω' -> 'ω' */
	set( 0x212a, 0x6b); /* 'K' -> 'k' */
	set( 0x212b, 0xe5); /* 'Å' -> 'å' */
	set( 0x2132, 0x214e); /* 'Ⅎ' -> 'ⅎ' */
	set( 0x2160, 0x2170); /* 'Ⅰ' -> 'ⅰ' */
	set( 0x2161, 0x2171); /* 'Ⅱ' -> 'ⅱ' */
	set( 0x2162, 0x2172); /* 'Ⅲ' -> 'ⅲ' */
	set( 0x2163, 0x2173); /* 'Ⅳ' -> 'ⅳ' */
	set( 0x2164, 0x2174); /* 'Ⅴ' -> 'ⅴ' */
	set( 0x2165, 0x2175); /* 'Ⅵ' -> 'ⅵ' */
	set( 0x2166, 0x2176); /* 'Ⅶ' -> 'ⅶ' */
	set( 0x2167, 0x2177); /* 'Ⅷ' -> 'ⅷ' */
	set( 0x2168, 0x2178); /* 'Ⅸ' -> 'ⅸ' */
	set( 0x2169, 0x2179); /* 'Ⅹ' -> 'ⅹ' */
	set( 0x216a, 0x217a); /* 'Ⅺ' -> 'ⅺ' */
	set( 0x216b, 0x217b); /* 'Ⅻ' -> 'ⅻ' */
	set( 0x216c, 0x217c); /* 'Ⅼ' -> 'ⅼ' */
	set( 0x216d, 0x217d); /* 'Ⅽ' -> 'ⅽ' */
	set( 0x216e, 0x217e); /* 'Ⅾ' -> 'ⅾ' */
	set( 0x216f, 0x217f); /* 'Ⅿ' -> 'ⅿ' */
	set( 0x2183, 0x2184); /* 'Ↄ' -> 'ↄ' */
	set( 0x24b6, 0x24d0); /* 'Ⓐ' -> 'ⓐ' */
	set( 0x24b7, 0x24d1); /* 'Ⓑ' -> 'ⓑ' */
	set( 0x24b8, 0x24d2); /* 'Ⓒ' -> 'ⓒ' */
	set( 0x24b9, 0x24d3); /* 'Ⓓ' -> 'ⓓ' */
	set( 0x24ba, 0x24d4); /* 'Ⓔ' -> 'ⓔ' */
	set( 0x24bb, 0x24d5); /* 'Ⓕ' -> 'ⓕ' */
	set( 0x24bc, 0x24d6); /* 'Ⓖ' -> 'ⓖ' */
	set( 0x24bd, 0x24d7); /* 'Ⓗ' -> 'ⓗ' */
	set( 0x24be, 0x24d8); /* 'Ⓘ' -> 'ⓘ' */
	set( 0x24bf, 0x24d9); /* 'Ⓙ' -> 'ⓙ' */
	set( 0x24c0, 0x24da); /* 'Ⓚ' -> 'ⓚ' */
	set( 0x24c1, 0x24db); /* 'Ⓛ' -> 'ⓛ' */
	set( 0x24c2, 0x24dc); /* 'Ⓜ' -> 'ⓜ' */
	set( 0x24c3, 0x24dd); /* 'Ⓝ' -> 'ⓝ' */
	set( 0x24c4, 0x24de); /* 'Ⓞ' -> 'ⓞ' */
	set( 0x24c5, 0x24df); /* 'Ⓟ' -> 'ⓟ' */
	set( 0x24c6, 0x24e0); /* 'Ⓠ' -> 'ⓠ' */
	set( 0x24c7, 0x24e1); /* 'Ⓡ' -> 'ⓡ' */
	set( 0x24c8, 0x24e2); /* 'Ⓢ' -> 'ⓢ' */
	set( 0x24c9, 0x24e3); /* 'Ⓣ' -> 'ⓣ' */
	set( 0x24ca, 0x24e4); /* 'Ⓤ' -> 'ⓤ' */
	set( 0x24cb, 0x24e5); /* 'Ⓥ' -> 'ⓥ' */
	set( 0x24cc, 0x24e6); /* 'Ⓦ' -> 'ⓦ' */
	set( 0x24cd, 0x24e7); /* 'Ⓧ' -> 'ⓧ' */
	set( 0x24ce, 0x24e8); /* 'Ⓨ' -> 'ⓨ' */
	set( 0x24cf, 0x24e9); /* 'Ⓩ' -> 'ⓩ' */
	set( 0x2c00, 0x2c30); /* 'Ⰰ' -> 'ⰰ' */
	set( 0x2c01, 0x2c31); /* 'Ⰱ' -> 'ⰱ' */
	set( 0x2c02, 0x2c32); /* 'Ⰲ' -> 'ⰲ' */
	set( 0x2c03, 0x2c33); /* 'Ⰳ' -> 'ⰳ' */
	set( 0x2c04, 0x2c34); /* 'Ⰴ' -> 'ⰴ' */
	set( 0x2c05, 0x2c35); /* 'Ⰵ' -> 'ⰵ' */
	set( 0x2c06, 0x2c36); /* 'Ⰶ' -> 'ⰶ' */
	set( 0x2c07, 0x2c37); /* 'Ⰷ' -> 'ⰷ' */
	set( 0x2c08, 0x2c38); /* 'Ⰸ' -> 'ⰸ' */
	set( 0x2c09, 0x2c39); /* 'Ⰹ' -> 'ⰹ' */
	set( 0x2c0a, 0x2c3a); /* 'Ⰺ' -> 'ⰺ' */
	set( 0x2c0b, 0x2c3b); /* 'Ⰻ' -> 'ⰻ' */
	set( 0x2c0c, 0x2c3c); /* 'Ⰼ' -> 'ⰼ' */
	set( 0x2c0d, 0x2c3d); /* 'Ⰽ' -> 'ⰽ' */
	set( 0x2c0e, 0x2c3e); /* 'Ⰾ' -> 'ⰾ' */
	set( 0x2c0f, 0x2c3f); /* 'Ⰿ' -> 'ⰿ' */
	set( 0x2c10, 0x2c40); /* 'Ⱀ' -> 'ⱀ' */
	set( 0x2c11, 0x2c41); /* 'Ⱁ' -> 'ⱁ' */
	set( 0x2c12, 0x2c42); /* 'Ⱂ' -> 'ⱂ' */
	set( 0x2c13, 0x2c43); /* 'Ⱃ' -> 'ⱃ' */
	set( 0x2c14, 0x2c44); /* 'Ⱄ' -> 'ⱄ' */
	set( 0x2c15, 0x2c45); /* 'Ⱅ' -> 'ⱅ' */
	set( 0x2c16, 0x2c46); /* 'Ⱆ' -> 'ⱆ' */
	set( 0x2c17, 0x2c47); /* 'Ⱇ' -> 'ⱇ' */
	set( 0x2c18, 0x2c48); /* 'Ⱈ' -> 'ⱈ' */
	set( 0x2c19, 0x2c49); /* 'Ⱉ' -> 'ⱉ' */
	set( 0x2c1a, 0x2c4a); /* 'Ⱊ' -> 'ⱊ' */
	set( 0x2c1b, 0x2c4b); /* 'Ⱋ' -> 'ⱋ' */
	set( 0x2c1c, 0x2c4c); /* 'Ⱌ' -> 'ⱌ' */
	set( 0x2c1d, 0x2c4d); /* 'Ⱍ' -> 'ⱍ' */
	set( 0x2c1e, 0x2c4e); /* 'Ⱎ' -> 'ⱎ' */
	set( 0x2c1f, 0x2c4f); /* 'Ⱏ' -> 'ⱏ' */
	set( 0x2c20, 0x2c50); /* 'Ⱐ' -> 'ⱐ' */
	set( 0x2c21, 0x2c51); /* 'Ⱑ' -> 'ⱑ' */
	set( 0x2c22, 0x2c52); /* 'Ⱒ' -> 'ⱒ' */
	set( 0x2c23, 0x2c53); /* 'Ⱓ' -> 'ⱓ' */
	set( 0x2c24, 0x2c54); /* 'Ⱔ' -> 'ⱔ' */
	set( 0x2c25, 0x2c55); /* 'Ⱕ' -> 'ⱕ' */
	set( 0x2c26, 0x2c56); /* 'Ⱖ' -> 'ⱖ' */
	set( 0x2c27, 0x2c57); /* 'Ⱗ' -> 'ⱗ' */
	set( 0x2c28, 0x2c58); /* 'Ⱘ' -> 'ⱘ' */
	set( 0x2c29, 0x2c59); /* 'Ⱙ' -> 'ⱙ' */
	set( 0x2c2a, 0x2c5a); /* 'Ⱚ' -> 'ⱚ' */
	set( 0x2c2b, 0x2c5b); /* 'Ⱛ' -> 'ⱛ' */
	set( 0x2c2c, 0x2c5c); /* 'Ⱜ' -> 'ⱜ' */
	set( 0x2c2d, 0x2c5d); /* 'Ⱝ' -> 'ⱝ' */
	set( 0x2c2e, 0x2c5e); /* 'Ⱞ' -> 'ⱞ' */
	set( 0x2c60, 0x2c61); /* 'Ⱡ' -> 'ⱡ' */
	set( 0x2c62, 0x26b); /* 'Ɫ' -> 'ɫ' */
	set( 0x2c63, 0x1d7d); /* 'Ᵽ' -> 'ᵽ' */
	set( 0x2c64, 0x27d); /* 'Ɽ' -> 'ɽ' */
	set( 0x2c67, 0x2c68); /* 'Ⱨ' -> 'ⱨ' */
	set( 0x2c69, 0x2c6a); /* 'Ⱪ' -> 'ⱪ' */
	set( 0x2c6b, 0x2c6c); /* 'Ⱬ' -> 'ⱬ' */
	set( 0x2c6d, 0x251); /* 'Ɑ' -> 'ɑ' */
	set( 0x2c6e, 0x271); /* 'Ɱ' -> 'ɱ' */
	set( 0x2c6f, 0x250); /* 'Ɐ' -> 'ɐ' */
	set( 0x2c70, 0x252); /* 'Ɒ' -> 'ɒ' */
	set( 0x2c72, 0x2c73); /* 'Ⱳ' -> 'ⱳ' */
	set( 0x2c75, 0x2c76); /* 'Ⱶ' -> 'ⱶ' */
	set( 0x2c7e, 0x23f); /* 'Ȿ' -> 'ȿ' */
	set( 0x2c7f, 0x240); /* 'Ɀ' -> 'ɀ' */
	set( 0x2c80, 0x2c81); /* 'Ⲁ' -> 'ⲁ' */
	set( 0x2c82, 0x2c83); /* 'Ⲃ' -> 'ⲃ' */
	set( 0x2c84, 0x2c85); /* 'Ⲅ' -> 'ⲅ' */
	set( 0x2c86, 0x2c87); /* 'Ⲇ' -> 'ⲇ' */
	set( 0x2c88, 0x2c89); /* 'Ⲉ' -> 'ⲉ' */
	set( 0x2c8a, 0x2c8b); /* 'Ⲋ' -> 'ⲋ' */
	set( 0x2c8c, 0x2c8d); /* 'Ⲍ' -> 'ⲍ' */
	set( 0x2c8e, 0x2c8f); /* 'Ⲏ' -> 'ⲏ' */
	set( 0x2c90, 0x2c91); /* 'Ⲑ' -> 'ⲑ' */
	set( 0x2c92, 0x2c93); /* 'Ⲓ' -> 'ⲓ' */
	set( 0x2c94, 0x2c95); /* 'Ⲕ' -> 'ⲕ' */
	set( 0x2c96, 0x2c97); /* 'Ⲗ' -> 'ⲗ' */
	set( 0x2c98, 0x2c99); /* 'Ⲙ' -> 'ⲙ' */
	set( 0x2c9a, 0x2c9b); /* 'Ⲛ' -> 'ⲛ' */
	set( 0x2c9c, 0x2c9d); /* 'Ⲝ' -> 'ⲝ' */
	set( 0x2c9e, 0x2c9f); /* 'Ⲟ' -> 'ⲟ' */
	set( 0x2ca0, 0x2ca1); /* 'Ⲡ' -> 'ⲡ' */
	set( 0x2ca2, 0x2ca3); /* 'Ⲣ' -> 'ⲣ' */
	set( 0x2ca4, 0x2ca5); /* 'Ⲥ' -> 'ⲥ' */
	set( 0x2ca6, 0x2ca7); /* 'Ⲧ' -> 'ⲧ' */
	set( 0x2ca8, 0x2ca9); /* 'Ⲩ' -> 'ⲩ' */
	set( 0x2caa, 0x2cab); /* 'Ⲫ' -> 'ⲫ' */
	set( 0x2cac, 0x2cad); /* 'Ⲭ' -> 'ⲭ' */
	set( 0x2cae, 0x2caf); /* 'Ⲯ' -> 'ⲯ' */
	set( 0x2cb0, 0x2cb1); /* 'Ⲱ' -> 'ⲱ' */
	set( 0x2cb2, 0x2cb3); /* 'Ⲳ' -> 'ⲳ' */
	set( 0x2cb4, 0x2cb5); /* 'Ⲵ' -> 'ⲵ' */
	set( 0x2cb6, 0x2cb7); /* 'Ⲷ' -> 'ⲷ' */
	set( 0x2cb8, 0x2cb9); /* 'Ⲹ' -> 'ⲹ' */
	set( 0x2cba, 0x2cbb); /* 'Ⲻ' -> 'ⲻ' */
	set( 0x2cbc, 0x2cbd); /* 'Ⲽ' -> 'ⲽ' */
	set( 0x2cbe, 0x2cbf); /* 'Ⲿ' -> 'ⲿ' */
	set( 0x2cc0, 0x2cc1); /* 'Ⳁ' -> 'ⳁ' */
	set( 0x2cc2, 0x2cc3); /* 'Ⳃ' -> 'ⳃ' */
	set( 0x2cc4, 0x2cc5); /* 'Ⳅ' -> 'ⳅ' */
	set( 0x2cc6, 0x2cc7); /* 'Ⳇ' -> 'ⳇ' */
	set( 0x2cc8, 0x2cc9); /* 'Ⳉ' -> 'ⳉ' */
	set( 0x2cca, 0x2ccb); /* 'Ⳋ' -> 'ⳋ' */
	set( 0x2ccc, 0x2ccd); /* 'Ⳍ' -> 'ⳍ' */
	set( 0x2cce, 0x2ccf); /* 'Ⳏ' -> 'ⳏ' */
	set( 0x2cd0, 0x2cd1); /* 'Ⳑ' -> 'ⳑ' */
	set( 0x2cd2, 0x2cd3); /* 'Ⳓ' -> 'ⳓ' */
	set( 0x2cd4, 0x2cd5); /* 'Ⳕ' -> 'ⳕ' */
	set( 0x2cd6, 0x2cd7); /* 'Ⳗ' -> 'ⳗ' */
	set( 0x2cd8, 0x2cd9); /* 'Ⳙ' -> 'ⳙ' */
	set( 0x2cda, 0x2cdb); /* 'Ⳛ' -> 'ⳛ' */
	set( 0x2cdc, 0x2cdd); /* 'Ⳝ' -> 'ⳝ' */
	set( 0x2cde, 0x2cdf); /* 'Ⳟ' -> 'ⳟ' */
	set( 0x2ce0, 0x2ce1); /* 'Ⳡ' -> 'ⳡ' */
	set( 0x2ce2, 0x2ce3); /* 'Ⳣ' -> 'ⳣ' */
	set( 0x2ceb, 0x2cec); /* 'Ⳬ' -> 'ⳬ' */
	set( 0x2ced, 0x2cee); /* 'Ⳮ' -> 'ⳮ' */
	set( 0xa640, 0xa641); /* 'Ꙁ' -> 'ꙁ' */
	set( 0xa642, 0xa643); /* 'Ꙃ' -> 'ꙃ' */
	set( 0xa644, 0xa645); /* 'Ꙅ' -> 'ꙅ' */
	set( 0xa646, 0xa647); /* 'Ꙇ' -> 'ꙇ' */
	set( 0xa648, 0xa649); /* 'Ꙉ' -> 'ꙉ' */
	set( 0xa64a, 0xa64b); /* 'Ꙋ' -> 'ꙋ' */
	set( 0xa64c, 0xa64d); /* 'Ꙍ' -> 'ꙍ' */
	set( 0xa64e, 0xa64f); /* 'Ꙏ' -> 'ꙏ' */
	set( 0xa650, 0xa651); /* 'Ꙑ' -> 'ꙑ' */
	set( 0xa652, 0xa653); /* 'Ꙓ' -> 'ꙓ' */
	set( 0xa654, 0xa655); /* 'Ꙕ' -> 'ꙕ' */
	set( 0xa656, 0xa657); /* 'Ꙗ' -> 'ꙗ' */
	set( 0xa658, 0xa659); /* 'Ꙙ' -> 'ꙙ' */
	set( 0xa65a, 0xa65b); /* 'Ꙛ' -> 'ꙛ' */
	set( 0xa65c, 0xa65d); /* 'Ꙝ' -> 'ꙝ' */
	set( 0xa65e, 0xa65f); /* 'Ꙟ' -> 'ꙟ' */
	set( 0xa662, 0xa663); /* 'Ꙣ' -> 'ꙣ' */
	set( 0xa664, 0xa665); /* 'Ꙥ' -> 'ꙥ' */
	set( 0xa666, 0xa667); /* 'Ꙧ' -> 'ꙧ' */
	set( 0xa668, 0xa669); /* 'Ꙩ' -> 'ꙩ' */
	set( 0xa66a, 0xa66b); /* 'Ꙫ' -> 'ꙫ' */
	set( 0xa66c, 0xa66d); /* 'Ꙭ' -> 'ꙭ' */
	set( 0xa680, 0xa681); /* 'Ꚁ' -> 'ꚁ' */
	set( 0xa682, 0xa683); /* 'Ꚃ' -> 'ꚃ' */
	set( 0xa684, 0xa685); /* 'Ꚅ' -> 'ꚅ' */
	set( 0xa686, 0xa687); /* 'Ꚇ' -> 'ꚇ' */
	set( 0xa688, 0xa689); /* 'Ꚉ' -> 'ꚉ' */
	set( 0xa68a, 0xa68b); /* 'Ꚋ' -> 'ꚋ' */
	set( 0xa68c, 0xa68d); /* 'Ꚍ' -> 'ꚍ' */
	set( 0xa68e, 0xa68f); /* 'Ꚏ' -> 'ꚏ' */
	set( 0xa690, 0xa691); /* 'Ꚑ' -> 'ꚑ' */
	set( 0xa692, 0xa693); /* 'Ꚓ' -> 'ꚓ' */
	set( 0xa694, 0xa695); /* 'Ꚕ' -> 'ꚕ' */
	set( 0xa696, 0xa697); /* 'Ꚗ' -> 'ꚗ' */
	set( 0xa722, 0xa723); /* 'Ꜣ' -> 'ꜣ' */
	set( 0xa724, 0xa725); /* 'Ꜥ' -> 'ꜥ' */
	set( 0xa726, 0xa727); /* 'Ꜧ' -> 'ꜧ' */
	set( 0xa728, 0xa729); /* 'Ꜩ' -> 'ꜩ' */
	set( 0xa72a, 0xa72b); /* 'Ꜫ' -> 'ꜫ' */
	set( 0xa72c, 0xa72d); /* 'Ꜭ' -> 'ꜭ' */
	set( 0xa72e, 0xa72f); /* 'Ꜯ' -> 'ꜯ' */
	set( 0xa732, 0xa733); /* 'Ꜳ' -> 'ꜳ' */
	set( 0xa734, 0xa735); /* 'Ꜵ' -> 'ꜵ' */
	set( 0xa736, 0xa737); /* 'Ꜷ' -> 'ꜷ' */
	set( 0xa738, 0xa739); /* 'Ꜹ' -> 'ꜹ' */
	set( 0xa73a, 0xa73b); /* 'Ꜻ' -> 'ꜻ' */
	set( 0xa73c, 0xa73d); /* 'Ꜽ' -> 'ꜽ' */
	set( 0xa73e, 0xa73f); /* 'Ꜿ' -> 'ꜿ' */
	set( 0xa740, 0xa741); /* 'Ꝁ' -> 'ꝁ' */
	set( 0xa742, 0xa743); /* 'Ꝃ' -> 'ꝃ' */
	set( 0xa744, 0xa745); /* 'Ꝅ' -> 'ꝅ' */
	set( 0xa746, 0xa747); /* 'Ꝇ' -> 'ꝇ' */
	set( 0xa748, 0xa749); /* 'Ꝉ' -> 'ꝉ' */
	set( 0xa74a, 0xa74b); /* 'Ꝋ' -> 'ꝋ' */
	set( 0xa74c, 0xa74d); /* 'Ꝍ' -> 'ꝍ' */
	set( 0xa74e, 0xa74f); /* 'Ꝏ' -> 'ꝏ' */
	set( 0xa750, 0xa751); /* 'Ꝑ' -> 'ꝑ' */
	set( 0xa752, 0xa753); /* 'Ꝓ' -> 'ꝓ' */
	set( 0xa754, 0xa755); /* 'Ꝕ' -> 'ꝕ' */
	set( 0xa756, 0xa757); /* 'Ꝗ' -> 'ꝗ' */
	set( 0xa758, 0xa759); /* 'Ꝙ' -> 'ꝙ' */
	set( 0xa75a, 0xa75b); /* 'Ꝛ' -> 'ꝛ' */
	set( 0xa75c, 0xa75d); /* 'Ꝝ' -> 'ꝝ' */
	set( 0xa75e, 0xa75f); /* 'Ꝟ' -> 'ꝟ' */
	set( 0xa760, 0xa761); /* 'Ꝡ' -> 'ꝡ' */
	set( 0xa762, 0xa763); /* 'Ꝣ' -> 'ꝣ' */
	set( 0xa764, 0xa765); /* 'Ꝥ' -> 'ꝥ' */
	set( 0xa766, 0xa767); /* 'Ꝧ' -> 'ꝧ' */
	set( 0xa768, 0xa769); /* 'Ꝩ' -> 'ꝩ' */
	set( 0xa76a, 0xa76b); /* 'Ꝫ' -> 'ꝫ' */
	set( 0xa76c, 0xa76d); /* 'Ꝭ' -> 'ꝭ' */
	set( 0xa76e, 0xa76f); /* 'Ꝯ' -> 'ꝯ' */
	set( 0xa779, 0xa77a); /* 'Ꝺ' -> 'ꝺ' */
	set( 0xa77b, 0xa77c); /* 'Ꝼ' -> 'ꝼ' */
	set( 0xa77d, 0x1d79); /* 'Ᵹ' -> 'ᵹ' */
	set( 0xa77e, 0xa77f); /* 'Ꝿ' -> 'ꝿ' */
	set( 0xa780, 0xa781); /* 'Ꞁ' -> 'ꞁ' */
	set( 0xa782, 0xa783); /* 'Ꞃ' -> 'ꞃ' */
	set( 0xa784, 0xa785); /* 'Ꞅ' -> 'ꞅ' */
	set( 0xa786, 0xa787); /* 'Ꞇ' -> 'ꞇ' */
	set( 0xa78b, 0xa78c); /* 'Ꞌ' -> 'ꞌ' */
	set( 0xff21, 0xff41); /* 'Ａ' -> 'ａ' */
	set( 0xff22, 0xff42); /* 'Ｂ' -> 'ｂ' */
	set( 0xff23, 0xff43); /* 'Ｃ' -> 'ｃ' */
	set( 0xff24, 0xff44); /* 'Ｄ' -> 'ｄ' */
	set( 0xff25, 0xff45); /* 'Ｅ' -> 'ｅ' */
	set( 0xff26, 0xff46); /* 'Ｆ' -> 'ｆ' */
	set( 0xff27, 0xff47); /* 'Ｇ' -> 'ｇ' */
	set( 0xff28, 0xff48); /* 'Ｈ' -> 'ｈ' */
	set( 0xff29, 0xff49); /* 'Ｉ' -> 'ｉ' */
	set( 0xff2a, 0xff4a); /* 'Ｊ' -> 'ｊ' */
	set( 0xff2b, 0xff4b); /* 'Ｋ' -> 'ｋ' */
	set( 0xff2c, 0xff4c); /* 'Ｌ' -> 'ｌ' */
	set( 0xff2d, 0xff4d); /* 'Ｍ' -> 'ｍ' */
	set( 0xff2e, 0xff4e); /* 'Ｎ' -> 'ｎ' */
	set( 0xff2f, 0xff4f); /* 'Ｏ' -> 'ｏ' */
	set( 0xff30, 0xff50); /* 'Ｐ' -> 'ｐ' */
	set( 0xff31, 0xff51); /* 'Ｑ' -> 'ｑ' */
	set( 0xff32, 0xff52); /* 'Ｒ' -> 'ｒ' */
	set( 0xff33, 0xff53); /* 'Ｓ' -> 'ｓ' */
	set( 0xff34, 0xff54); /* 'Ｔ' -> 'ｔ' */
	set( 0xff35, 0xff55); /* 'Ｕ' -> 'ｕ' */
	set( 0xff36, 0xff56); /* 'Ｖ' -> 'ｖ' */
	set( 0xff37, 0xff57); /* 'Ｗ' -> 'ｗ' */
	set( 0xff38, 0xff58); /* 'Ｘ' -> 'ｘ' */
	set( 0xff39, 0xff59); /* 'Ｙ' -> 'ｙ' */
	set( 0xff3a, 0xff5a); /* 'Ｚ' -> 'ｚ' */
	set( 0x10400, 0x10428); /* '𐐀' -> '𐐨' */
	set( 0x10401, 0x10429); /* '𐐁' -> '𐐩' */
	set( 0x10402, 0x1042a); /* '𐐂' -> '𐐪' */
	set( 0x10403, 0x1042b); /* '𐐃' -> '𐐫' */
	set( 0x10404, 0x1042c); /* '𐐄' -> '𐐬' */
	set( 0x10405, 0x1042d); /* '𐐅' -> '𐐭' */
	set( 0x10406, 0x1042e); /* '𐐆' -> '𐐮' */
	set( 0x10407, 0x1042f); /* '𐐇' -> '𐐯' */
	set( 0x10408, 0x10430); /* '𐐈' -> '𐐰' */
	set( 0x10409, 0x10431); /* '𐐉' -> '𐐱' */
	set( 0x1040a, 0x10432); /* '𐐊' -> '𐐲' */
	set( 0x1040b, 0x10433); /* '𐐋' -> '𐐳' */
	set( 0x1040c, 0x10434); /* '𐐌' -> '𐐴' */
	set( 0x1040d, 0x10435); /* '𐐍' -> '𐐵' */
	set( 0x1040e, 0x10436); /* '𐐎' -> '𐐶' */
	set( 0x1040f, 0x10437); /* '𐐏' -> '𐐷' */
	set( 0x10410, 0x10438); /* '𐐐' -> '𐐸' */
	set( 0x10411, 0x10439); /* '𐐑' -> '𐐹' */
	set( 0x10412, 0x1043a); /* '𐐒' -> '𐐺' */
	set( 0x10413, 0x1043b); /* '𐐓' -> '𐐻' */
	set( 0x10414, 0x1043c); /* '𐐔' -> '𐐼' */
	set( 0x10415, 0x1043d); /* '𐐕' -> '𐐽' */
	set( 0x10416, 0x1043e); /* '𐐖' -> '𐐾' */
	set( 0x10417, 0x1043f); /* '𐐗' -> '𐐿' */
	set( 0x10418, 0x10440); /* '𐐘' -> '𐑀' */
	set( 0x10419, 0x10441); /* '𐐙' -> '𐑁' */
	set( 0x1041a, 0x10442); /* '𐐚' -> '𐑂' */
	set( 0x1041b, 0x10443); /* '𐐛' -> '𐑃' */
	set( 0x1041c, 0x10444); /* '𐐜' -> '𐑄' */
	set( 0x1041d, 0x10445); /* '𐐝' -> '𐑅' */
	set( 0x1041e, 0x10446); /* '𐐞' -> '𐑆' */
	set( 0x1041f, 0x10447); /* '𐐟' -> '𐑇' */
	set( 0x10420, 0x10448); /* '𐐠' -> '𐑈' */
	set( 0x10421, 0x10449); /* '𐐡' -> '𐑉' */
	set( 0x10422, 0x1044a); /* '𐐢' -> '𐑊' */
	set( 0x10423, 0x1044b); /* '𐐣' -> '𐑋' */
	set( 0x10424, 0x1044c); /* '𐐤' -> '𐑌' */
	set( 0x10425, 0x1044d); /* '𐐥' -> '𐑍' */
	set( 0x10426, 0x1044e); /* '𐐦' -> '𐑎' */
	set( 0x10427, 0x1044f); /* '𐐧' -> '𐑏' */
}

void CharMap::buildMapToupper()
{
	set( 0x61, 0x41); /* 'a' -> 'A' */
	set( 0x62, 0x42); /* 'b' -> 'B' */
	set( 0x63, 0x43); /* 'c' -> 'C' */
	set( 0x64, 0x44); /* 'd' -> 'D' */
	set( 0x65, 0x45); /* 'e' -> 'E' */
	set( 0x66, 0x46); /* 'f' -> 'F' */
	set( 0x67, 0x47); /* 'g' -> 'G' */
	set( 0x68, 0x48); /* 'h' -> 'H' */
	set( 0x69, 0x49); /* 'i' -> 'I' */
	set( 0x6a, 0x4a); /* 'j' -> 'J' */
	set( 0x6b, 0x4b); /* 'k' -> 'K' */
	set( 0x6c, 0x4c); /* 'l' -> 'L' */
	set( 0x6d, 0x4d); /* 'm' -> 'M' */
	set( 0x6e, 0x4e); /* 'n' -> 'N' */
	set( 0x6f, 0x4f); /* 'o' -> 'O' */
	set( 0x70, 0x50); /* 'p' -> 'P' */
	set( 0x71, 0x51); /* 'q' -> 'Q' */
	set( 0x72, 0x52); /* 'r' -> 'R' */
	set( 0x73, 0x53); /* 's' -> 'S' */
	set( 0x74, 0x54); /* 't' -> 'T' */
	set( 0x75, 0x55); /* 'u' -> 'U' */
	set( 0x76, 0x56); /* 'v' -> 'V' */
	set( 0x77, 0x57); /* 'w' -> 'W' */
	set( 0x78, 0x58); /* 'x' -> 'X' */
	set( 0x79, 0x59); /* 'y' -> 'Y' */
	set( 0x7a, 0x5a); /* 'z' -> 'Z' */
	set( 0xb5, 0x39c); /* 'µ' -> 'Μ' */
	set( 0xe0, 0xc0); /* 'à' -> 'À' */
	set( 0xe1, 0xc1); /* 'á' -> 'Á' */
	set( 0xe2, 0xc2); /* 'â' -> 'Â' */
	set( 0xe3, 0xc3); /* 'ã' -> 'Ã' */
	set( 0xe4, 0xc4); /* 'ä' -> 'Ä' */
	set( 0xe5, 0xc5); /* 'å' -> 'Å' */
	set( 0xe6, 0xc6); /* 'æ' -> 'Æ' */
	set( 0xe7, 0xc7); /* 'ç' -> 'Ç' */
	set( 0xe8, 0xc8); /* 'è' -> 'È' */
	set( 0xe9, 0xc9); /* 'é' -> 'É' */
	set( 0xea, 0xca); /* 'ê' -> 'Ê' */
	set( 0xeb, 0xcb); /* 'ë' -> 'Ë' */
	set( 0xec, 0xcc); /* 'ì' -> 'Ì' */
	set( 0xed, 0xcd); /* 'í' -> 'Í' */
	set( 0xee, 0xce); /* 'î' -> 'Î' */
	set( 0xef, 0xcf); /* 'ï' -> 'Ï' */
	set( 0xf0, 0xd0); /* 'ð' -> 'Ð' */
	set( 0xf1, 0xd1); /* 'ñ' -> 'Ñ' */
	set( 0xf2, 0xd2); /* 'ò' -> 'Ò' */
	set( 0xf3, 0xd3); /* 'ó' -> 'Ó' */
	set( 0xf4, 0xd4); /* 'ô' -> 'Ô' */
	set( 0xf5, 0xd5); /* 'õ' -> 'Õ' */
	set( 0xf6, 0xd6); /* 'ö' -> 'Ö' */
	set( 0xf8, 0xd8); /* 'ø' -> 'Ø' */
	set( 0xf9, 0xd9); /* 'ù' -> 'Ù' */
	set( 0xfa, 0xda); /* 'ú' -> 'Ú' */
	set( 0xfb, 0xdb); /* 'û' -> 'Û' */
	set( 0xfc, 0xdc); /* 'ü' -> 'Ü' */
	set( 0xfd, 0xdd); /* 'ý' -> 'Ý' */
	set( 0xfe, 0xde); /* 'þ' -> 'Þ' */
	set( 0xff, 0x178); /* 'ÿ' -> 'Ÿ' */
	set( 0x101, 0x100); /* 'ā' -> 'Ā' */
	set( 0x103, 0x102); /* 'ă' -> 'Ă' */
	set( 0x105, 0x104); /* 'ą' -> 'Ą' */
	set( 0x107, 0x106); /* 'ć' -> 'Ć' */
	set( 0x109, 0x108); /* 'ĉ' -> 'Ĉ' */
	set( 0x10b, 0x10a); /* 'ċ' -> 'Ċ' */
	set( 0x10d, 0x10c); /* 'č' -> 'Č' */
	set( 0x10f, 0x10e); /* 'ď' -> 'Ď' */
	set( 0x111, 0x110); /* 'đ' -> 'Đ' */
	set( 0x113, 0x112); /* 'ē' -> 'Ē' */
	set( 0x115, 0x114); /* 'ĕ' -> 'Ĕ' */
	set( 0x117, 0x116); /* 'ė' -> 'Ė' */
	set( 0x119, 0x118); /* 'ę' -> 'Ę' */
	set( 0x11b, 0x11a); /* 'ě' -> 'Ě' */
	set( 0x11d, 0x11c); /* 'ĝ' -> 'Ĝ' */
	set( 0x11f, 0x11e); /* 'ğ' -> 'Ğ' */
	set( 0x121, 0x120); /* 'ġ' -> 'Ġ' */
	set( 0x123, 0x122); /* 'ģ' -> 'Ģ' */
	set( 0x125, 0x124); /* 'ĥ' -> 'Ĥ' */
	set( 0x127, 0x126); /* 'ħ' -> 'Ħ' */
	set( 0x129, 0x128); /* 'ĩ' -> 'Ĩ' */
	set( 0x12b, 0x12a); /* 'ī' -> 'Ī' */
	set( 0x12d, 0x12c); /* 'ĭ' -> 'Ĭ' */
	set( 0x12f, 0x12e); /* 'į' -> 'Į' */
	set( 0x131, 0x49); /* 'ı' -> 'I' */
	set( 0x133, 0x132); /* 'ĳ' -> 'Ĳ' */
	set( 0x135, 0x134); /* 'ĵ' -> 'Ĵ' */
	set( 0x137, 0x136); /* 'ķ' -> 'Ķ' */
	set( 0x13a, 0x139); /* 'ĺ' -> 'Ĺ' */
	set( 0x13c, 0x13b); /* 'ļ' -> 'Ļ' */
	set( 0x13e, 0x13d); /* 'ľ' -> 'Ľ' */
	set( 0x140, 0x13f); /* 'ŀ' -> 'Ŀ' */
	set( 0x142, 0x141); /* 'ł' -> 'Ł' */
	set( 0x144, 0x143); /* 'ń' -> 'Ń' */
	set( 0x146, 0x145); /* 'ņ' -> 'Ņ' */
	set( 0x148, 0x147); /* 'ň' -> 'Ň' */
	set( 0x14b, 0x14a); /* 'ŋ' -> 'Ŋ' */
	set( 0x14d, 0x14c); /* 'ō' -> 'Ō' */
	set( 0x14f, 0x14e); /* 'ŏ' -> 'Ŏ' */
	set( 0x151, 0x150); /* 'ő' -> 'Ő' */
	set( 0x153, 0x152); /* 'œ' -> 'Œ' */
	set( 0x155, 0x154); /* 'ŕ' -> 'Ŕ' */
	set( 0x157, 0x156); /* 'ŗ' -> 'Ŗ' */
	set( 0x159, 0x158); /* 'ř' -> 'Ř' */
	set( 0x15b, 0x15a); /* 'ś' -> 'Ś' */
	set( 0x15d, 0x15c); /* 'ŝ' -> 'Ŝ' */
	set( 0x15f, 0x15e); /* 'ş' -> 'Ş' */
	set( 0x161, 0x160); /* 'š' -> 'Š' */
	set( 0x163, 0x162); /* 'ţ' -> 'Ţ' */
	set( 0x165, 0x164); /* 'ť' -> 'Ť' */
	set( 0x167, 0x166); /* 'ŧ' -> 'Ŧ' */
	set( 0x169, 0x168); /* 'ũ' -> 'Ũ' */
	set( 0x16b, 0x16a); /* 'ū' -> 'Ū' */
	set( 0x16d, 0x16c); /* 'ŭ' -> 'Ŭ' */
	set( 0x16f, 0x16e); /* 'ů' -> 'Ů' */
	set( 0x171, 0x170); /* 'ű' -> 'Ű' */
	set( 0x173, 0x172); /* 'ų' -> 'Ų' */
	set( 0x175, 0x174); /* 'ŵ' -> 'Ŵ' */
	set( 0x177, 0x176); /* 'ŷ' -> 'Ŷ' */
	set( 0x17a, 0x179); /* 'ź' -> 'Ź' */
	set( 0x17c, 0x17b); /* 'ż' -> 'Ż' */
	set( 0x17e, 0x17d); /* 'ž' -> 'Ž' */
	set( 0x17f, 0x53); /* 'ſ' -> 'S' */
	set( 0x180, 0x243); /* 'ƀ' -> 'Ƀ' */
	set( 0x183, 0x182); /* 'ƃ' -> 'Ƃ' */
	set( 0x185, 0x184); /* 'ƅ' -> 'Ƅ' */
	set( 0x188, 0x187); /* 'ƈ' -> 'Ƈ' */
	set( 0x18c, 0x18b); /* 'ƌ' -> 'Ƌ' */
	set( 0x192, 0x191); /* 'ƒ' -> 'Ƒ' */
	set( 0x195, 0x1f6); /* 'ƕ' -> 'Ƕ' */
	set( 0x199, 0x198); /* 'ƙ' -> 'Ƙ' */
	set( 0x19a, 0x23d); /* 'ƚ' -> 'Ƚ' */
	set( 0x19e, 0x220); /* 'ƞ' -> 'Ƞ' */
	set( 0x1a1, 0x1a0); /* 'ơ' -> 'Ơ' */
	set( 0x1a3, 0x1a2); /* 'ƣ' -> 'Ƣ' */
	set( 0x1a5, 0x1a4); /* 'ƥ' -> 'Ƥ' */
	set( 0x1a8, 0x1a7); /* 'ƨ' -> 'Ƨ' */
	set( 0x1ad, 0x1ac); /* 'ƭ' -> 'Ƭ' */
	set( 0x1b0, 0x1af); /* 'ư' -> 'Ư' */
	set( 0x1b4, 0x1b3); /* 'ƴ' -> 'Ƴ' */
	set( 0x1b6, 0x1b5); /* 'ƶ' -> 'Ƶ' */
	set( 0x1b9, 0x1b8); /* 'ƹ' -> 'Ƹ' */
	set( 0x1bd, 0x1bc); /* 'ƽ' -> 'Ƽ' */
	set( 0x1bf, 0x1f7); /* 'ƿ' -> 'Ƿ' */
	set( 0x1c5, 0x1c4); /* 'ǅ' -> 'Ǆ' */
	set( 0x1c6, 0x1c4); /* 'ǆ' -> 'Ǆ' */
	set( 0x1c8, 0x1c7); /* 'ǈ' -> 'Ǉ' */
	set( 0x1c9, 0x1c7); /* 'ǉ' -> 'Ǉ' */
	set( 0x1cb, 0x1ca); /* 'ǋ' -> 'Ǌ' */
	set( 0x1cc, 0x1ca); /* 'ǌ' -> 'Ǌ' */
	set( 0x1ce, 0x1cd); /* 'ǎ' -> 'Ǎ' */
	set( 0x1d0, 0x1cf); /* 'ǐ' -> 'Ǐ' */
	set( 0x1d2, 0x1d1); /* 'ǒ' -> 'Ǒ' */
	set( 0x1d4, 0x1d3); /* 'ǔ' -> 'Ǔ' */
	set( 0x1d6, 0x1d5); /* 'ǖ' -> 'Ǖ' */
	set( 0x1d8, 0x1d7); /* 'ǘ' -> 'Ǘ' */
	set( 0x1da, 0x1d9); /* 'ǚ' -> 'Ǚ' */
	set( 0x1dc, 0x1db); /* 'ǜ' -> 'Ǜ' */
	set( 0x1dd, 0x18e); /* 'ǝ' -> 'Ǝ' */
	set( 0x1df, 0x1de); /* 'ǟ' -> 'Ǟ' */
	set( 0x1e1, 0x1e0); /* 'ǡ' -> 'Ǡ' */
	set( 0x1e3, 0x1e2); /* 'ǣ' -> 'Ǣ' */
	set( 0x1e5, 0x1e4); /* 'ǥ' -> 'Ǥ' */
	set( 0x1e7, 0x1e6); /* 'ǧ' -> 'Ǧ' */
	set( 0x1e9, 0x1e8); /* 'ǩ' -> 'Ǩ' */
	set( 0x1eb, 0x1ea); /* 'ǫ' -> 'Ǫ' */
	set( 0x1ed, 0x1ec); /* 'ǭ' -> 'Ǭ' */
	set( 0x1ef, 0x1ee); /* 'ǯ' -> 'Ǯ' */
	set( 0x1f2, 0x1f1); /* 'ǲ' -> 'Ǳ' */
	set( 0x1f3, 0x1f1); /* 'ǳ' -> 'Ǳ' */
	set( 0x1f5, 0x1f4); /* 'ǵ' -> 'Ǵ' */
	set( 0x1f9, 0x1f8); /* 'ǹ' -> 'Ǹ' */
	set( 0x1fb, 0x1fa); /* 'ǻ' -> 'Ǻ' */
	set( 0x1fd, 0x1fc); /* 'ǽ' -> 'Ǽ' */
	set( 0x1ff, 0x1fe); /* 'ǿ' -> 'Ǿ' */
	set( 0x201, 0x200); /* 'ȁ' -> 'Ȁ' */
	set( 0x203, 0x202); /* 'ȃ' -> 'Ȃ' */
	set( 0x205, 0x204); /* 'ȅ' -> 'Ȅ' */
	set( 0x207, 0x206); /* 'ȇ' -> 'Ȇ' */
	set( 0x209, 0x208); /* 'ȉ' -> 'Ȉ' */
	set( 0x20b, 0x20a); /* 'ȋ' -> 'Ȋ' */
	set( 0x20d, 0x20c); /* 'ȍ' -> 'Ȍ' */
	set( 0x20f, 0x20e); /* 'ȏ' -> 'Ȏ' */
	set( 0x211, 0x210); /* 'ȑ' -> 'Ȑ' */
	set( 0x213, 0x212); /* 'ȓ' -> 'Ȓ' */
	set( 0x215, 0x214); /* 'ȕ' -> 'Ȕ' */
	set( 0x217, 0x216); /* 'ȗ' -> 'Ȗ' */
	set( 0x219, 0x218); /* 'ș' -> 'Ș' */
	set( 0x21b, 0x21a); /* 'ț' -> 'Ț' */
	set( 0x21d, 0x21c); /* 'ȝ' -> 'Ȝ' */
	set( 0x21f, 0x21e); /* 'ȟ' -> 'Ȟ' */
	set( 0x223, 0x222); /* 'ȣ' -> 'Ȣ' */
	set( 0x225, 0x224); /* 'ȥ' -> 'Ȥ' */
	set( 0x227, 0x226); /* 'ȧ' -> 'Ȧ' */
	set( 0x229, 0x228); /* 'ȩ' -> 'Ȩ' */
	set( 0x22b, 0x22a); /* 'ȫ' -> 'Ȫ' */
	set( 0x22d, 0x22c); /* 'ȭ' -> 'Ȭ' */
	set( 0x22f, 0x22e); /* 'ȯ' -> 'Ȯ' */
	set( 0x231, 0x230); /* 'ȱ' -> 'Ȱ' */
	set( 0x233, 0x232); /* 'ȳ' -> 'Ȳ' */
	set( 0x23c, 0x23b); /* 'ȼ' -> 'Ȼ' */
	set( 0x23f, 0x2c7e); /* 'ȿ' -> 'Ȿ' */
	set( 0x240, 0x2c7f); /* 'ɀ' -> 'Ɀ' */
	set( 0x242, 0x241); /* 'ɂ' -> 'Ɂ' */
	set( 0x247, 0x246); /* 'ɇ' -> 'Ɇ' */
	set( 0x249, 0x248); /* 'ɉ' -> 'Ɉ' */
	set( 0x24b, 0x24a); /* 'ɋ' -> 'Ɋ' */
	set( 0x24d, 0x24c); /* 'ɍ' -> 'Ɍ' */
	set( 0x24f, 0x24e); /* 'ɏ' -> 'Ɏ' */
	set( 0x250, 0x2c6f); /* 'ɐ' -> 'Ɐ' */
	set( 0x251, 0x2c6d); /* 'ɑ' -> 'Ɑ' */
	set( 0x252, 0x2c70); /* 'ɒ' -> 'Ɒ' */
	set( 0x253, 0x181); /* 'ɓ' -> 'Ɓ' */
	set( 0x254, 0x186); /* 'ɔ' -> 'Ɔ' */
	set( 0x256, 0x189); /* 'ɖ' -> 'Ɖ' */
	set( 0x257, 0x18a); /* 'ɗ' -> 'Ɗ' */
	set( 0x259, 0x18f); /* 'ə' -> 'Ə' */
	set( 0x25b, 0x190); /* 'ɛ' -> 'Ɛ' */
	set( 0x260, 0x193); /* 'ɠ' -> 'Ɠ' */
	set( 0x263, 0x194); /* 'ɣ' -> 'Ɣ' */
	set( 0x268, 0x197); /* 'ɨ' -> 'Ɨ' */
	set( 0x269, 0x196); /* 'ɩ' -> 'Ɩ' */
	set( 0x26b, 0x2c62); /* 'ɫ' -> 'Ɫ' */
	set( 0x26f, 0x19c); /* 'ɯ' -> 'Ɯ' */
	set( 0x271, 0x2c6e); /* 'ɱ' -> 'Ɱ' */
	set( 0x272, 0x19d); /* 'ɲ' -> 'Ɲ' */
	set( 0x275, 0x19f); /* 'ɵ' -> 'Ɵ' */
	set( 0x27d, 0x2c64); /* 'ɽ' -> 'Ɽ' */
	set( 0x280, 0x1a6); /* 'ʀ' -> 'Ʀ' */
	set( 0x283, 0x1a9); /* 'ʃ' -> 'Ʃ' */
	set( 0x288, 0x1ae); /* 'ʈ' -> 'Ʈ' */
	set( 0x289, 0x244); /* 'ʉ' -> 'Ʉ' */
	set( 0x28a, 0x1b1); /* 'ʊ' -> 'Ʊ' */
	set( 0x28b, 0x1b2); /* 'ʋ' -> 'Ʋ' */
	set( 0x28c, 0x245); /* 'ʌ' -> 'Ʌ' */
	set( 0x292, 0x1b7); /* 'ʒ' -> 'Ʒ' */
	set( 0x345, 0x399); /* 'ͅ' -> 'Ι' */
	set( 0x371, 0x370); /* 'ͱ' -> 'Ͱ' */
	set( 0x373, 0x372); /* 'ͳ' -> 'Ͳ' */
	set( 0x377, 0x376); /* 'ͷ' -> 'Ͷ' */
	set( 0x37b, 0x3fd); /* 'ͻ' -> 'Ͻ' */
	set( 0x37c, 0x3fe); /* 'ͼ' -> 'Ͼ' */
	set( 0x37d, 0x3ff); /* 'ͽ' -> 'Ͽ' */
	set( 0x3ac, 0x386); /* 'ά' -> 'Ά' */
	set( 0x3ad, 0x388); /* 'έ' -> 'Έ' */
	set( 0x3ae, 0x389); /* 'ή' -> 'Ή' */
	set( 0x3af, 0x38a); /* 'ί' -> 'Ί' */
	set( 0x3b1, 0x391); /* 'α' -> 'Α' */
	set( 0x3b2, 0x392); /* 'β' -> 'Β' */
	set( 0x3b3, 0x393); /* 'γ' -> 'Γ' */
	set( 0x3b4, 0x394); /* 'δ' -> 'Δ' */
	set( 0x3b5, 0x395); /* 'ε' -> 'Ε' */
	set( 0x3b6, 0x396); /* 'ζ' -> 'Ζ' */
	set( 0x3b7, 0x397); /* 'η' -> 'Η' */
	set( 0x3b8, 0x398); /* 'θ' -> 'Θ' */
	set( 0x3b9, 0x399); /* 'ι' -> 'Ι' */
	set( 0x3ba, 0x39a); /* 'κ' -> 'Κ' */
	set( 0x3bb, 0x39b); /* 'λ' -> 'Λ' */
	set( 0x3bc, 0x39c); /* 'μ' -> 'Μ' */
	set( 0x3bd, 0x39d); /* 'ν' -> 'Ν' */
	set( 0x3be, 0x39e); /* 'ξ' -> 'Ξ' */
	set( 0x3bf, 0x39f); /* 'ο' -> 'Ο' */
	set( 0x3c0, 0x3a0); /* 'π' -> 'Π' */
	set( 0x3c1, 0x3a1); /* 'ρ' -> 'Ρ' */
	set( 0x3c2, 0x3a3); /* 'ς' -> 'Σ' */
	set( 0x3c3, 0x3a3); /* 'σ' -> 'Σ' */
	set( 0x3c4, 0x3a4); /* 'τ' -> 'Τ' */
	set( 0x3c5, 0x3a5); /* 'υ' -> 'Υ' */
	set( 0x3c6, 0x3a6); /* 'φ' -> 'Φ' */
	set( 0x3c7, 0x3a7); /* 'χ' -> 'Χ' */
	set( 0x3c8, 0x3a8); /* 'ψ' -> 'Ψ' */
	set( 0x3c9, 0x3a9); /* 'ω' -> 'Ω' */
	set( 0x3ca, 0x3aa); /* 'ϊ' -> 'Ϊ' */
	set( 0x3cb, 0x3ab); /* 'ϋ' -> 'Ϋ' */
	set( 0x3cc, 0x38c); /* 'ό' -> 'Ό' */
	set( 0x3cd, 0x38e); /* 'ύ' -> 'Ύ' */
	set( 0x3ce, 0x38f); /* 'ώ' -> 'Ώ' */
	set( 0x3d0, 0x392); /* 'ϐ' -> 'Β' */
	set( 0x3d1, 0x398); /* 'ϑ' -> 'Θ' */
	set( 0x3d5, 0x3a6); /* 'ϕ' -> 'Φ' */
	set( 0x3d6, 0x3a0); /* 'ϖ' -> 'Π' */
	set( 0x3d7, 0x3cf); /* 'ϗ' -> 'Ϗ' */
	set( 0x3d9, 0x3d8); /* 'ϙ' -> 'Ϙ' */
	set( 0x3db, 0x3da); /* 'ϛ' -> 'Ϛ' */
	set( 0x3dd, 0x3dc); /* 'ϝ' -> 'Ϝ' */
	set( 0x3df, 0x3de); /* 'ϟ' -> 'Ϟ' */
	set( 0x3e1, 0x3e0); /* 'ϡ' -> 'Ϡ' */
	set( 0x3e3, 0x3e2); /* 'ϣ' -> 'Ϣ' */
	set( 0x3e5, 0x3e4); /* 'ϥ' -> 'Ϥ' */
	set( 0x3e7, 0x3e6); /* 'ϧ' -> 'Ϧ' */
	set( 0x3e9, 0x3e8); /* 'ϩ' -> 'Ϩ' */
	set( 0x3eb, 0x3ea); /* 'ϫ' -> 'Ϫ' */
	set( 0x3ed, 0x3ec); /* 'ϭ' -> 'Ϭ' */
	set( 0x3ef, 0x3ee); /* 'ϯ' -> 'Ϯ' */
	set( 0x3f0, 0x39a); /* 'ϰ' -> 'Κ' */
	set( 0x3f1, 0x3a1); /* 'ϱ' -> 'Ρ' */
	set( 0x3f2, 0x3f9); /* 'ϲ' -> 'Ϲ' */
	set( 0x3f5, 0x395); /* 'ϵ' -> 'Ε' */
	set( 0x3f8, 0x3f7); /* 'ϸ' -> 'Ϸ' */
	set( 0x3fb, 0x3fa); /* 'ϻ' -> 'Ϻ' */
	set( 0x430, 0x410); /* 'а' -> 'А' */
	set( 0x431, 0x411); /* 'б' -> 'Б' */
	set( 0x432, 0x412); /* 'в' -> 'В' */
	set( 0x433, 0x413); /* 'г' -> 'Г' */
	set( 0x434, 0x414); /* 'д' -> 'Д' */
	set( 0x435, 0x415); /* 'е' -> 'Е' */
	set( 0x436, 0x416); /* 'ж' -> 'Ж' */
	set( 0x437, 0x417); /* 'з' -> 'З' */
	set( 0x438, 0x418); /* 'и' -> 'И' */
	set( 0x439, 0x419); /* 'й' -> 'Й' */
	set( 0x43a, 0x41a); /* 'к' -> 'К' */
	set( 0x43b, 0x41b); /* 'л' -> 'Л' */
	set( 0x43c, 0x41c); /* 'м' -> 'М' */
	set( 0x43d, 0x41d); /* 'н' -> 'Н' */
	set( 0x43e, 0x41e); /* 'о' -> 'О' */
	set( 0x43f, 0x41f); /* 'п' -> 'П' */
	set( 0x440, 0x420); /* 'р' -> 'Р' */
	set( 0x441, 0x421); /* 'с' -> 'С' */
	set( 0x442, 0x422); /* 'т' -> 'Т' */
	set( 0x443, 0x423); /* 'у' -> 'У' */
	set( 0x444, 0x424); /* 'ф' -> 'Ф' */
	set( 0x445, 0x425); /* 'х' -> 'Х' */
	set( 0x446, 0x426); /* 'ц' -> 'Ц' */
	set( 0x447, 0x427); /* 'ч' -> 'Ч' */
	set( 0x448, 0x428); /* 'ш' -> 'Ш' */
	set( 0x449, 0x429); /* 'щ' -> 'Щ' */
	set( 0x44a, 0x42a); /* 'ъ' -> 'Ъ' */
	set( 0x44b, 0x42b); /* 'ы' -> 'Ы' */
	set( 0x44c, 0x42c); /* 'ь' -> 'Ь' */
	set( 0x44d, 0x42d); /* 'э' -> 'Э' */
	set( 0x44e, 0x42e); /* 'ю' -> 'Ю' */
	set( 0x44f, 0x42f); /* 'я' -> 'Я' */
	set( 0x450, 0x400); /* 'ѐ' -> 'Ѐ' */
	set( 0x451, 0x401); /* 'ё' -> 'Ё' */
	set( 0x452, 0x402); /* 'ђ' -> 'Ђ' */
	set( 0x453, 0x403); /* 'ѓ' -> 'Ѓ' */
	set( 0x454, 0x404); /* 'є' -> 'Є' */
	set( 0x455, 0x405); /* 'ѕ' -> 'Ѕ' */
	set( 0x456, 0x406); /* 'і' -> 'І' */
	set( 0x457, 0x407); /* 'ї' -> 'Ї' */
	set( 0x458, 0x408); /* 'ј' -> 'Ј' */
	set( 0x459, 0x409); /* 'љ' -> 'Љ' */
	set( 0x45a, 0x40a); /* 'њ' -> 'Њ' */
	set( 0x45b, 0x40b); /* 'ћ' -> 'Ћ' */
	set( 0x45c, 0x40c); /* 'ќ' -> 'Ќ' */
	set( 0x45d, 0x40d); /* 'ѝ' -> 'Ѝ' */
	set( 0x45e, 0x40e); /* 'ў' -> 'Ў' */
	set( 0x45f, 0x40f); /* 'џ' -> 'Џ' */
	set( 0x461, 0x460); /* 'ѡ' -> 'Ѡ' */
	set( 0x463, 0x462); /* 'ѣ' -> 'Ѣ' */
	set( 0x465, 0x464); /* 'ѥ' -> 'Ѥ' */
	set( 0x467, 0x466); /* 'ѧ' -> 'Ѧ' */
	set( 0x469, 0x468); /* 'ѩ' -> 'Ѩ' */
	set( 0x46b, 0x46a); /* 'ѫ' -> 'Ѫ' */
	set( 0x46d, 0x46c); /* 'ѭ' -> 'Ѭ' */
	set( 0x46f, 0x46e); /* 'ѯ' -> 'Ѯ' */
	set( 0x471, 0x470); /* 'ѱ' -> 'Ѱ' */
	set( 0x473, 0x472); /* 'ѳ' -> 'Ѳ' */
	set( 0x475, 0x474); /* 'ѵ' -> 'Ѵ' */
	set( 0x477, 0x476); /* 'ѷ' -> 'Ѷ' */
	set( 0x479, 0x478); /* 'ѹ' -> 'Ѹ' */
	set( 0x47b, 0x47a); /* 'ѻ' -> 'Ѻ' */
	set( 0x47d, 0x47c); /* 'ѽ' -> 'Ѽ' */
	set( 0x47f, 0x47e); /* 'ѿ' -> 'Ѿ' */
	set( 0x481, 0x480); /* 'ҁ' -> 'Ҁ' */
	set( 0x48b, 0x48a); /* 'ҋ' -> 'Ҋ' */
	set( 0x48d, 0x48c); /* 'ҍ' -> 'Ҍ' */
	set( 0x48f, 0x48e); /* 'ҏ' -> 'Ҏ' */
	set( 0x491, 0x490); /* 'ґ' -> 'Ґ' */
	set( 0x493, 0x492); /* 'ғ' -> 'Ғ' */
	set( 0x495, 0x494); /* 'ҕ' -> 'Ҕ' */
	set( 0x497, 0x496); /* 'җ' -> 'Җ' */
	set( 0x499, 0x498); /* 'ҙ' -> 'Ҙ' */
	set( 0x49b, 0x49a); /* 'қ' -> 'Қ' */
	set( 0x49d, 0x49c); /* 'ҝ' -> 'Ҝ' */
	set( 0x49f, 0x49e); /* 'ҟ' -> 'Ҟ' */
	set( 0x4a1, 0x4a0); /* 'ҡ' -> 'Ҡ' */
	set( 0x4a3, 0x4a2); /* 'ң' -> 'Ң' */
	set( 0x4a5, 0x4a4); /* 'ҥ' -> 'Ҥ' */
	set( 0x4a7, 0x4a6); /* 'ҧ' -> 'Ҧ' */
	set( 0x4a9, 0x4a8); /* 'ҩ' -> 'Ҩ' */
	set( 0x4ab, 0x4aa); /* 'ҫ' -> 'Ҫ' */
	set( 0x4ad, 0x4ac); /* 'ҭ' -> 'Ҭ' */
	set( 0x4af, 0x4ae); /* 'ү' -> 'Ү' */
	set( 0x4b1, 0x4b0); /* 'ұ' -> 'Ұ' */
	set( 0x4b3, 0x4b2); /* 'ҳ' -> 'Ҳ' */
	set( 0x4b5, 0x4b4); /* 'ҵ' -> 'Ҵ' */
	set( 0x4b7, 0x4b6); /* 'ҷ' -> 'Ҷ' */
	set( 0x4b9, 0x4b8); /* 'ҹ' -> 'Ҹ' */
	set( 0x4bb, 0x4ba); /* 'һ' -> 'Һ' */
	set( 0x4bd, 0x4bc); /* 'ҽ' -> 'Ҽ' */
	set( 0x4bf, 0x4be); /* 'ҿ' -> 'Ҿ' */
	set( 0x4c2, 0x4c1); /* 'ӂ' -> 'Ӂ' */
	set( 0x4c4, 0x4c3); /* 'ӄ' -> 'Ӄ' */
	set( 0x4c6, 0x4c5); /* 'ӆ' -> 'Ӆ' */
	set( 0x4c8, 0x4c7); /* 'ӈ' -> 'Ӈ' */
	set( 0x4ca, 0x4c9); /* 'ӊ' -> 'Ӊ' */
	set( 0x4cc, 0x4cb); /* 'ӌ' -> 'Ӌ' */
	set( 0x4ce, 0x4cd); /* 'ӎ' -> 'Ӎ' */
	set( 0x4cf, 0x4c0); /* 'ӏ' -> 'Ӏ' */
	set( 0x4d1, 0x4d0); /* 'ӑ' -> 'Ӑ' */
	set( 0x4d3, 0x4d2); /* 'ӓ' -> 'Ӓ' */
	set( 0x4d5, 0x4d4); /* 'ӕ' -> 'Ӕ' */
	set( 0x4d7, 0x4d6); /* 'ӗ' -> 'Ӗ' */
	set( 0x4d9, 0x4d8); /* 'ә' -> 'Ә' */
	set( 0x4db, 0x4da); /* 'ӛ' -> 'Ӛ' */
	set( 0x4dd, 0x4dc); /* 'ӝ' -> 'Ӝ' */
	set( 0x4df, 0x4de); /* 'ӟ' -> 'Ӟ' */
	set( 0x4e1, 0x4e0); /* 'ӡ' -> 'Ӡ' */
	set( 0x4e3, 0x4e2); /* 'ӣ' -> 'Ӣ' */
	set( 0x4e5, 0x4e4); /* 'ӥ' -> 'Ӥ' */
	set( 0x4e7, 0x4e6); /* 'ӧ' -> 'Ӧ' */
	set( 0x4e9, 0x4e8); /* 'ө' -> 'Ө' */
	set( 0x4eb, 0x4ea); /* 'ӫ' -> 'Ӫ' */
	set( 0x4ed, 0x4ec); /* 'ӭ' -> 'Ӭ' */
	set( 0x4ef, 0x4ee); /* 'ӯ' -> 'Ӯ' */
	set( 0x4f1, 0x4f0); /* 'ӱ' -> 'Ӱ' */
	set( 0x4f3, 0x4f2); /* 'ӳ' -> 'Ӳ' */
	set( 0x4f5, 0x4f4); /* 'ӵ' -> 'Ӵ' */
	set( 0x4f7, 0x4f6); /* 'ӷ' -> 'Ӷ' */
	set( 0x4f9, 0x4f8); /* 'ӹ' -> 'Ӹ' */
	set( 0x4fb, 0x4fa); /* 'ӻ' -> 'Ӻ' */
	set( 0x4fd, 0x4fc); /* 'ӽ' -> 'Ӽ' */
	set( 0x4ff, 0x4fe); /* 'ӿ' -> 'Ӿ' */
	set( 0x501, 0x500); /* 'ԁ' -> 'Ԁ' */
	set( 0x503, 0x502); /* 'ԃ' -> 'Ԃ' */
	set( 0x505, 0x504); /* 'ԅ' -> 'Ԅ' */
	set( 0x507, 0x506); /* 'ԇ' -> 'Ԇ' */
	set( 0x509, 0x508); /* 'ԉ' -> 'Ԉ' */
	set( 0x50b, 0x50a); /* 'ԋ' -> 'Ԋ' */
	set( 0x50d, 0x50c); /* 'ԍ' -> 'Ԍ' */
	set( 0x50f, 0x50e); /* 'ԏ' -> 'Ԏ' */
	set( 0x511, 0x510); /* 'ԑ' -> 'Ԑ' */
	set( 0x513, 0x512); /* 'ԓ' -> 'Ԓ' */
	set( 0x515, 0x514); /* 'ԕ' -> 'Ԕ' */
	set( 0x517, 0x516); /* 'ԗ' -> 'Ԗ' */
	set( 0x519, 0x518); /* 'ԙ' -> 'Ԙ' */
	set( 0x51b, 0x51a); /* 'ԛ' -> 'Ԛ' */
	set( 0x51d, 0x51c); /* 'ԝ' -> 'Ԝ' */
	set( 0x51f, 0x51e); /* 'ԟ' -> 'Ԟ' */
	set( 0x521, 0x520); /* 'ԡ' -> 'Ԡ' */
	set( 0x523, 0x522); /* 'ԣ' -> 'Ԣ' */
	set( 0x525, 0x524); /* 'ԥ' -> 'Ԥ' */
	set( 0x561, 0x531); /* 'ա' -> 'Ա' */
	set( 0x562, 0x532); /* 'բ' -> 'Բ' */
	set( 0x563, 0x533); /* 'գ' -> 'Գ' */
	set( 0x564, 0x534); /* 'դ' -> 'Դ' */
	set( 0x565, 0x535); /* 'ե' -> 'Ե' */
	set( 0x566, 0x536); /* 'զ' -> 'Զ' */
	set( 0x567, 0x537); /* 'է' -> 'Է' */
	set( 0x568, 0x538); /* 'ը' -> 'Ը' */
	set( 0x569, 0x539); /* 'թ' -> 'Թ' */
	set( 0x56a, 0x53a); /* 'ժ' -> 'Ժ' */
	set( 0x56b, 0x53b); /* 'ի' -> 'Ի' */
	set( 0x56c, 0x53c); /* 'լ' -> 'Լ' */
	set( 0x56d, 0x53d); /* 'խ' -> 'Խ' */
	set( 0x56e, 0x53e); /* 'ծ' -> 'Ծ' */
	set( 0x56f, 0x53f); /* 'կ' -> 'Կ' */
	set( 0x570, 0x540); /* 'հ' -> 'Հ' */
	set( 0x571, 0x541); /* 'ձ' -> 'Ձ' */
	set( 0x572, 0x542); /* 'ղ' -> 'Ղ' */
	set( 0x573, 0x543); /* 'ճ' -> 'Ճ' */
	set( 0x574, 0x544); /* 'մ' -> 'Մ' */
	set( 0x575, 0x545); /* 'յ' -> 'Յ' */
	set( 0x576, 0x546); /* 'ն' -> 'Ն' */
	set( 0x577, 0x547); /* 'շ' -> 'Շ' */
	set( 0x578, 0x548); /* 'ո' -> 'Ո' */
	set( 0x579, 0x549); /* 'չ' -> 'Չ' */
	set( 0x57a, 0x54a); /* 'պ' -> 'Պ' */
	set( 0x57b, 0x54b); /* 'ջ' -> 'Ջ' */
	set( 0x57c, 0x54c); /* 'ռ' -> 'Ռ' */
	set( 0x57d, 0x54d); /* 'ս' -> 'Ս' */
	set( 0x57e, 0x54e); /* 'վ' -> 'Վ' */
	set( 0x57f, 0x54f); /* 'տ' -> 'Տ' */
	set( 0x580, 0x550); /* 'ր' -> 'Ր' */
	set( 0x581, 0x551); /* 'ց' -> 'Ց' */
	set( 0x582, 0x552); /* 'ւ' -> 'Ւ' */
	set( 0x583, 0x553); /* 'փ' -> 'Փ' */
	set( 0x584, 0x554); /* 'ք' -> 'Ք' */
	set( 0x585, 0x555); /* 'օ' -> 'Օ' */
	set( 0x586, 0x556); /* 'ֆ' -> 'Ֆ' */
	set( 0x1d79, 0xa77d); /* 'ᵹ' -> 'Ᵹ' */
	set( 0x1d7d, 0x2c63); /* 'ᵽ' -> 'Ᵽ' */
	set( 0x1e01, 0x1e00); /* 'ḁ' -> 'Ḁ' */
	set( 0x1e03, 0x1e02); /* 'ḃ' -> 'Ḃ' */
	set( 0x1e05, 0x1e04); /* 'ḅ' -> 'Ḅ' */
	set( 0x1e07, 0x1e06); /* 'ḇ' -> 'Ḇ' */
	set( 0x1e09, 0x1e08); /* 'ḉ' -> 'Ḉ' */
	set( 0x1e0b, 0x1e0a); /* 'ḋ' -> 'Ḋ' */
	set( 0x1e0d, 0x1e0c); /* 'ḍ' -> 'Ḍ' */
	set( 0x1e0f, 0x1e0e); /* 'ḏ' -> 'Ḏ' */
	set( 0x1e11, 0x1e10); /* 'ḑ' -> 'Ḑ' */
	set( 0x1e13, 0x1e12); /* 'ḓ' -> 'Ḓ' */
	set( 0x1e15, 0x1e14); /* 'ḕ' -> 'Ḕ' */
	set( 0x1e17, 0x1e16); /* 'ḗ' -> 'Ḗ' */
	set( 0x1e19, 0x1e18); /* 'ḙ' -> 'Ḙ' */
	set( 0x1e1b, 0x1e1a); /* 'ḛ' -> 'Ḛ' */
	set( 0x1e1d, 0x1e1c); /* 'ḝ' -> 'Ḝ' */
	set( 0x1e1f, 0x1e1e); /* 'ḟ' -> 'Ḟ' */
	set( 0x1e21, 0x1e20); /* 'ḡ' -> 'Ḡ' */
	set( 0x1e23, 0x1e22); /* 'ḣ' -> 'Ḣ' */
	set( 0x1e25, 0x1e24); /* 'ḥ' -> 'Ḥ' */
	set( 0x1e27, 0x1e26); /* 'ḧ' -> 'Ḧ' */
	set( 0x1e29, 0x1e28); /* 'ḩ' -> 'Ḩ' */
	set( 0x1e2b, 0x1e2a); /* 'ḫ' -> 'Ḫ' */
	set( 0x1e2d, 0x1e2c); /* 'ḭ' -> 'Ḭ' */
	set( 0x1e2f, 0x1e2e); /* 'ḯ' -> 'Ḯ' */
	set( 0x1e31, 0x1e30); /* 'ḱ' -> 'Ḱ' */
	set( 0x1e33, 0x1e32); /* 'ḳ' -> 'Ḳ' */
	set( 0x1e35, 0x1e34); /* 'ḵ' -> 'Ḵ' */
	set( 0x1e37, 0x1e36); /* 'ḷ' -> 'Ḷ' */
	set( 0x1e39, 0x1e38); /* 'ḹ' -> 'Ḹ' */
	set( 0x1e3b, 0x1e3a); /* 'ḻ' -> 'Ḻ' */
	set( 0x1e3d, 0x1e3c); /* 'ḽ' -> 'Ḽ' */
	set( 0x1e3f, 0x1e3e); /* 'ḿ' -> 'Ḿ' */
	set( 0x1e41, 0x1e40); /* 'ṁ' -> 'Ṁ' */
	set( 0x1e43, 0x1e42); /* 'ṃ' -> 'Ṃ' */
	set( 0x1e45, 0x1e44); /* 'ṅ' -> 'Ṅ' */
	set( 0x1e47, 0x1e46); /* 'ṇ' -> 'Ṇ' */
	set( 0x1e49, 0x1e48); /* 'ṉ' -> 'Ṉ' */
	set( 0x1e4b, 0x1e4a); /* 'ṋ' -> 'Ṋ' */
	set( 0x1e4d, 0x1e4c); /* 'ṍ' -> 'Ṍ' */
	set( 0x1e4f, 0x1e4e); /* 'ṏ' -> 'Ṏ' */
	set( 0x1e51, 0x1e50); /* 'ṑ' -> 'Ṑ' */
	set( 0x1e53, 0x1e52); /* 'ṓ' -> 'Ṓ' */
	set( 0x1e55, 0x1e54); /* 'ṕ' -> 'Ṕ' */
	set( 0x1e57, 0x1e56); /* 'ṗ' -> 'Ṗ' */
	set( 0x1e59, 0x1e58); /* 'ṙ' -> 'Ṙ' */
	set( 0x1e5b, 0x1e5a); /* 'ṛ' -> 'Ṛ' */
	set( 0x1e5d, 0x1e5c); /* 'ṝ' -> 'Ṝ' */
	set( 0x1e5f, 0x1e5e); /* 'ṟ' -> 'Ṟ' */
	set( 0x1e61, 0x1e60); /* 'ṡ' -> 'Ṡ' */
	set( 0x1e63, 0x1e62); /* 'ṣ' -> 'Ṣ' */
	set( 0x1e65, 0x1e64); /* 'ṥ' -> 'Ṥ' */
	set( 0x1e67, 0x1e66); /* 'ṧ' -> 'Ṧ' */
	set( 0x1e69, 0x1e68); /* 'ṩ' -> 'Ṩ' */
	set( 0x1e6b, 0x1e6a); /* 'ṫ' -> 'Ṫ' */
	set( 0x1e6d, 0x1e6c); /* 'ṭ' -> 'Ṭ' */
	set( 0x1e6f, 0x1e6e); /* 'ṯ' -> 'Ṯ' */
	set( 0x1e71, 0x1e70); /* 'ṱ' -> 'Ṱ' */
	set( 0x1e73, 0x1e72); /* 'ṳ' -> 'Ṳ' */
	set( 0x1e75, 0x1e74); /* 'ṵ' -> 'Ṵ' */
	set( 0x1e77, 0x1e76); /* 'ṷ' -> 'Ṷ' */
	set( 0x1e79, 0x1e78); /* 'ṹ' -> 'Ṹ' */
	set( 0x1e7b, 0x1e7a); /* 'ṻ' -> 'Ṻ' */
	set( 0x1e7d, 0x1e7c); /* 'ṽ' -> 'Ṽ' */
	set( 0x1e7f, 0x1e7e); /* 'ṿ' -> 'Ṿ' */
	set( 0x1e81, 0x1e80); /* 'ẁ' -> 'Ẁ' */
	set( 0x1e83, 0x1e82); /* 'ẃ' -> 'Ẃ' */
	set( 0x1e85, 0x1e84); /* 'ẅ' -> 'Ẅ' */
	set( 0x1e87, 0x1e86); /* 'ẇ' -> 'Ẇ' */
	set( 0x1e89, 0x1e88); /* 'ẉ' -> 'Ẉ' */
	set( 0x1e8b, 0x1e8a); /* 'ẋ' -> 'Ẋ' */
	set( 0x1e8d, 0x1e8c); /* 'ẍ' -> 'Ẍ' */
	set( 0x1e8f, 0x1e8e); /* 'ẏ' -> 'Ẏ' */
	set( 0x1e91, 0x1e90); /* 'ẑ' -> 'Ẑ' */
	set( 0x1e93, 0x1e92); /* 'ẓ' -> 'Ẓ' */
	set( 0x1e95, 0x1e94); /* 'ẕ' -> 'Ẕ' */
	set( 0x1e9b, 0x1e60); /* 'ẛ' -> 'Ṡ' */
	set( 0x1ea1, 0x1ea0); /* 'ạ' -> 'Ạ' */
	set( 0x1ea3, 0x1ea2); /* 'ả' -> 'Ả' */
	set( 0x1ea5, 0x1ea4); /* 'ấ' -> 'Ấ' */
	set( 0x1ea7, 0x1ea6); /* 'ầ' -> 'Ầ' */
	set( 0x1ea9, 0x1ea8); /* 'ẩ' -> 'Ẩ' */
	set( 0x1eab, 0x1eaa); /* 'ẫ' -> 'Ẫ' */
	set( 0x1ead, 0x1eac); /* 'ậ' -> 'Ậ' */
	set( 0x1eaf, 0x1eae); /* 'ắ' -> 'Ắ' */
	set( 0x1eb1, 0x1eb0); /* 'ằ' -> 'Ằ' */
	set( 0x1eb3, 0x1eb2); /* 'ẳ' -> 'Ẳ' */
	set( 0x1eb5, 0x1eb4); /* 'ẵ' -> 'Ẵ' */
	set( 0x1eb7, 0x1eb6); /* 'ặ' -> 'Ặ' */
	set( 0x1eb9, 0x1eb8); /* 'ẹ' -> 'Ẹ' */
	set( 0x1ebb, 0x1eba); /* 'ẻ' -> 'Ẻ' */
	set( 0x1ebd, 0x1ebc); /* 'ẽ' -> 'Ẽ' */
	set( 0x1ebf, 0x1ebe); /* 'ế' -> 'Ế' */
	set( 0x1ec1, 0x1ec0); /* 'ề' -> 'Ề' */
	set( 0x1ec3, 0x1ec2); /* 'ể' -> 'Ể' */
	set( 0x1ec5, 0x1ec4); /* 'ễ' -> 'Ễ' */
	set( 0x1ec7, 0x1ec6); /* 'ệ' -> 'Ệ' */
	set( 0x1ec9, 0x1ec8); /* 'ỉ' -> 'Ỉ' */
	set( 0x1ecb, 0x1eca); /* 'ị' -> 'Ị' */
	set( 0x1ecd, 0x1ecc); /* 'ọ' -> 'Ọ' */
	set( 0x1ecf, 0x1ece); /* 'ỏ' -> 'Ỏ' */
	set( 0x1ed1, 0x1ed0); /* 'ố' -> 'Ố' */
	set( 0x1ed3, 0x1ed2); /* 'ồ' -> 'Ồ' */
	set( 0x1ed5, 0x1ed4); /* 'ổ' -> 'Ổ' */
	set( 0x1ed7, 0x1ed6); /* 'ỗ' -> 'Ỗ' */
	set( 0x1ed9, 0x1ed8); /* 'ộ' -> 'Ộ' */
	set( 0x1edb, 0x1eda); /* 'ớ' -> 'Ớ' */
	set( 0x1edd, 0x1edc); /* 'ờ' -> 'Ờ' */
	set( 0x1edf, 0x1ede); /* 'ở' -> 'Ở' */
	set( 0x1ee1, 0x1ee0); /* 'ỡ' -> 'Ỡ' */
	set( 0x1ee3, 0x1ee2); /* 'ợ' -> 'Ợ' */
	set( 0x1ee5, 0x1ee4); /* 'ụ' -> 'Ụ' */
	set( 0x1ee7, 0x1ee6); /* 'ủ' -> 'Ủ' */
	set( 0x1ee9, 0x1ee8); /* 'ứ' -> 'Ứ' */
	set( 0x1eeb, 0x1eea); /* 'ừ' -> 'Ừ' */
	set( 0x1eed, 0x1eec); /* 'ử' -> 'Ử' */
	set( 0x1eef, 0x1eee); /* 'ữ' -> 'Ữ' */
	set( 0x1ef1, 0x1ef0); /* 'ự' -> 'Ự' */
	set( 0x1ef3, 0x1ef2); /* 'ỳ' -> 'Ỳ' */
	set( 0x1ef5, 0x1ef4); /* 'ỵ' -> 'Ỵ' */
	set( 0x1ef7, 0x1ef6); /* 'ỷ' -> 'Ỷ' */
	set( 0x1ef9, 0x1ef8); /* 'ỹ' -> 'Ỹ' */
	set( 0x1efb, 0x1efa); /* 'ỻ' -> 'Ỻ' */
	set( 0x1efd, 0x1efc); /* 'ỽ' -> 'Ỽ' */
	set( 0x1eff, 0x1efe); /* 'ỿ' -> 'Ỿ' */
	set( 0x1f00, 0x1f08); /* 'ἀ' -> 'Ἀ' */
	set( 0x1f01, 0x1f09); /* 'ἁ' -> 'Ἁ' */
	set( 0x1f02, 0x1f0a); /* 'ἂ' -> 'Ἂ' */
	set( 0x1f03, 0x1f0b); /* 'ἃ' -> 'Ἃ' */
	set( 0x1f04, 0x1f0c); /* 'ἄ' -> 'Ἄ' */
	set( 0x1f05, 0x1f0d); /* 'ἅ' -> 'Ἅ' */
	set( 0x1f06, 0x1f0e); /* 'ἆ' -> 'Ἆ' */
	set( 0x1f07, 0x1f0f); /* 'ἇ' -> 'Ἇ' */
	set( 0x1f10, 0x1f18); /* 'ἐ' -> 'Ἐ' */
	set( 0x1f11, 0x1f19); /* 'ἑ' -> 'Ἑ' */
	set( 0x1f12, 0x1f1a); /* 'ἒ' -> 'Ἒ' */
	set( 0x1f13, 0x1f1b); /* 'ἓ' -> 'Ἓ' */
	set( 0x1f14, 0x1f1c); /* 'ἔ' -> 'Ἔ' */
	set( 0x1f15, 0x1f1d); /* 'ἕ' -> 'Ἕ' */
	set( 0x1f20, 0x1f28); /* 'ἠ' -> 'Ἠ' */
	set( 0x1f21, 0x1f29); /* 'ἡ' -> 'Ἡ' */
	set( 0x1f22, 0x1f2a); /* 'ἢ' -> 'Ἢ' */
	set( 0x1f23, 0x1f2b); /* 'ἣ' -> 'Ἣ' */
	set( 0x1f24, 0x1f2c); /* 'ἤ' -> 'Ἤ' */
	set( 0x1f25, 0x1f2d); /* 'ἥ' -> 'Ἥ' */
	set( 0x1f26, 0x1f2e); /* 'ἦ' -> 'Ἦ' */
	set( 0x1f27, 0x1f2f); /* 'ἧ' -> 'Ἧ' */
	set( 0x1f30, 0x1f38); /* 'ἰ' -> 'Ἰ' */
	set( 0x1f31, 0x1f39); /* 'ἱ' -> 'Ἱ' */
	set( 0x1f32, 0x1f3a); /* 'ἲ' -> 'Ἲ' */
	set( 0x1f33, 0x1f3b); /* 'ἳ' -> 'Ἳ' */
	set( 0x1f34, 0x1f3c); /* 'ἴ' -> 'Ἴ' */
	set( 0x1f35, 0x1f3d); /* 'ἵ' -> 'Ἵ' */
	set( 0x1f36, 0x1f3e); /* 'ἶ' -> 'Ἶ' */
	set( 0x1f37, 0x1f3f); /* 'ἷ' -> 'Ἷ' */
	set( 0x1f40, 0x1f48); /* 'ὀ' -> 'Ὀ' */
	set( 0x1f41, 0x1f49); /* 'ὁ' -> 'Ὁ' */
	set( 0x1f42, 0x1f4a); /* 'ὂ' -> 'Ὂ' */
	set( 0x1f43, 0x1f4b); /* 'ὃ' -> 'Ὃ' */
	set( 0x1f44, 0x1f4c); /* 'ὄ' -> 'Ὄ' */
	set( 0x1f45, 0x1f4d); /* 'ὅ' -> 'Ὅ' */
	set( 0x1f51, 0x1f59); /* 'ὑ' -> 'Ὑ' */
	set( 0x1f53, 0x1f5b); /* 'ὓ' -> 'Ὓ' */
	set( 0x1f55, 0x1f5d); /* 'ὕ' -> 'Ὕ' */
	set( 0x1f57, 0x1f5f); /* 'ὗ' -> 'Ὗ' */
	set( 0x1f60, 0x1f68); /* 'ὠ' -> 'Ὠ' */
	set( 0x1f61, 0x1f69); /* 'ὡ' -> 'Ὡ' */
	set( 0x1f62, 0x1f6a); /* 'ὢ' -> 'Ὢ' */
	set( 0x1f63, 0x1f6b); /* 'ὣ' -> 'Ὣ' */
	set( 0x1f64, 0x1f6c); /* 'ὤ' -> 'Ὤ' */
	set( 0x1f65, 0x1f6d); /* 'ὥ' -> 'Ὥ' */
	set( 0x1f66, 0x1f6e); /* 'ὦ' -> 'Ὦ' */
	set( 0x1f67, 0x1f6f); /* 'ὧ' -> 'Ὧ' */
	set( 0x1f70, 0x1fba); /* 'ὰ' -> 'Ὰ' */
	set( 0x1f71, 0x1fbb); /* 'ά' -> 'Ά' */
	set( 0x1f72, 0x1fc8); /* 'ὲ' -> 'Ὲ' */
	set( 0x1f73, 0x1fc9); /* 'έ' -> 'Έ' */
	set( 0x1f74, 0x1fca); /* 'ὴ' -> 'Ὴ' */
	set( 0x1f75, 0x1fcb); /* 'ή' -> 'Ή' */
	set( 0x1f76, 0x1fda); /* 'ὶ' -> 'Ὶ' */
	set( 0x1f77, 0x1fdb); /* 'ί' -> 'Ί' */
	set( 0x1f78, 0x1ff8); /* 'ὸ' -> 'Ὸ' */
	set( 0x1f79, 0x1ff9); /* 'ό' -> 'Ό' */
	set( 0x1f7a, 0x1fea); /* 'ὺ' -> 'Ὺ' */
	set( 0x1f7b, 0x1feb); /* 'ύ' -> 'Ύ' */
	set( 0x1f7c, 0x1ffa); /* 'ὼ' -> 'Ὼ' */
	set( 0x1f7d, 0x1ffb); /* 'ώ' -> 'Ώ' */
	set( 0x1f80, 0x1f88); /* 'ᾀ' -> 'ᾈ' */
	set( 0x1f81, 0x1f89); /* 'ᾁ' -> 'ᾉ' */
	set( 0x1f82, 0x1f8a); /* 'ᾂ' -> 'ᾊ' */
	set( 0x1f83, 0x1f8b); /* 'ᾃ' -> 'ᾋ' */
	set( 0x1f84, 0x1f8c); /* 'ᾄ' -> 'ᾌ' */
	set( 0x1f85, 0x1f8d); /* 'ᾅ' -> 'ᾍ' */
	set( 0x1f86, 0x1f8e); /* 'ᾆ' -> 'ᾎ' */
	set( 0x1f87, 0x1f8f); /* 'ᾇ' -> 'ᾏ' */
	set( 0x1f90, 0x1f98); /* 'ᾐ' -> 'ᾘ' */
	set( 0x1f91, 0x1f99); /* 'ᾑ' -> 'ᾙ' */
	set( 0x1f92, 0x1f9a); /* 'ᾒ' -> 'ᾚ' */
	set( 0x1f93, 0x1f9b); /* 'ᾓ' -> 'ᾛ' */
	set( 0x1f94, 0x1f9c); /* 'ᾔ' -> 'ᾜ' */
	set( 0x1f95, 0x1f9d); /* 'ᾕ' -> 'ᾝ' */
	set( 0x1f96, 0x1f9e); /* 'ᾖ' -> 'ᾞ' */
	set( 0x1f97, 0x1f9f); /* 'ᾗ' -> 'ᾟ' */
	set( 0x1fa0, 0x1fa8); /* 'ᾠ' -> 'ᾨ' */
	set( 0x1fa1, 0x1fa9); /* 'ᾡ' -> 'ᾩ' */
	set( 0x1fa2, 0x1faa); /* 'ᾢ' -> 'ᾪ' */
	set( 0x1fa3, 0x1fab); /* 'ᾣ' -> 'ᾫ' */
	set( 0x1fa4, 0x1fac); /* 'ᾤ' -> 'ᾬ' */
	set( 0x1fa5, 0x1fad); /* 'ᾥ' -> 'ᾭ' */
	set( 0x1fa6, 0x1fae); /* 'ᾦ' -> 'ᾮ' */
	set( 0x1fa7, 0x1faf); /* 'ᾧ' -> 'ᾯ' */
	set( 0x1fb0, 0x1fb8); /* 'ᾰ' -> 'Ᾰ' */
	set( 0x1fb1, 0x1fb9); /* 'ᾱ' -> 'Ᾱ' */
	set( 0x1fb3, 0x1fbc); /* 'ᾳ' -> 'ᾼ' */
	set( 0x1fbe, 0x399); /* 'ι' -> 'Ι' */
	set( 0x1fc3, 0x1fcc); /* 'ῃ' -> 'ῌ' */
	set( 0x1fd0, 0x1fd8); /* 'ῐ' -> 'Ῐ' */
	set( 0x1fd1, 0x1fd9); /* 'ῑ' -> 'Ῑ' */
	set( 0x1fe0, 0x1fe8); /* 'ῠ' -> 'Ῠ' */
	set( 0x1fe1, 0x1fe9); /* 'ῡ' -> 'Ῡ' */
	set( 0x1fe5, 0x1fec); /* 'ῥ' -> 'Ῥ' */
	set( 0x1ff3, 0x1ffc); /* 'ῳ' -> 'ῼ' */
	set( 0x214e, 0x2132); /* 'ⅎ' -> 'Ⅎ' */
	set( 0x2170, 0x2160); /* 'ⅰ' -> 'Ⅰ' */
	set( 0x2171, 0x2161); /* 'ⅱ' -> 'Ⅱ' */
	set( 0x2172, 0x2162); /* 'ⅲ' -> 'Ⅲ' */
	set( 0x2173, 0x2163); /* 'ⅳ' -> 'Ⅳ' */
	set( 0x2174, 0x2164); /* 'ⅴ' -> 'Ⅴ' */
	set( 0x2175, 0x2165); /* 'ⅵ' -> 'Ⅵ' */
	set( 0x2176, 0x2166); /* 'ⅶ' -> 'Ⅶ' */
	set( 0x2177, 0x2167); /* 'ⅷ' -> 'Ⅷ' */
	set( 0x2178, 0x2168); /* 'ⅸ' -> 'Ⅸ' */
	set( 0x2179, 0x2169); /* 'ⅹ' -> 'Ⅹ' */
	set( 0x217a, 0x216a); /* 'ⅺ' -> 'Ⅺ' */
	set( 0x217b, 0x216b); /* 'ⅻ' -> 'Ⅻ' */
	set( 0x217c, 0x216c); /* 'ⅼ' -> 'Ⅼ' */
	set( 0x217d, 0x216d); /* 'ⅽ' -> 'Ⅽ' */
	set( 0x217e, 0x216e); /* 'ⅾ' -> 'Ⅾ' */
	set( 0x217f, 0x216f); /* 'ⅿ' -> 'Ⅿ' */
	set( 0x2184, 0x2183); /* 'ↄ' -> 'Ↄ' */
	set( 0x24d0, 0x24b6); /* 'ⓐ' -> 'Ⓐ' */
	set( 0x24d1, 0x24b7); /* 'ⓑ' -> 'Ⓑ' */
	set( 0x24d2, 0x24b8); /* 'ⓒ' -> 'Ⓒ' */
	set( 0x24d3, 0x24b9); /* 'ⓓ' -> 'Ⓓ' */
	set( 0x24d4, 0x24ba); /* 'ⓔ' -> 'Ⓔ' */
	set( 0x24d5, 0x24bb); /* 'ⓕ' -> 'Ⓕ' */
	set( 0x24d6, 0x24bc); /* 'ⓖ' -> 'Ⓖ' */
	set( 0x24d7, 0x24bd); /* 'ⓗ' -> 'Ⓗ' */
	set( 0x24d8, 0x24be); /* 'ⓘ' -> 'Ⓘ' */
	set( 0x24d9, 0x24bf); /* 'ⓙ' -> 'Ⓙ' */
	set( 0x24da, 0x24c0); /* 'ⓚ' -> 'Ⓚ' */
	set( 0x24db, 0x24c1); /* 'ⓛ' -> 'Ⓛ' */
	set( 0x24dc, 0x24c2); /* 'ⓜ' -> 'Ⓜ' */
	set( 0x24dd, 0x24c3); /* 'ⓝ' -> 'Ⓝ' */
	set( 0x24de, 0x24c4); /* 'ⓞ' -> 'Ⓞ' */
	set( 0x24df, 0x24c5); /* 'ⓟ' -> 'Ⓟ' */
	set( 0x24e0, 0x24c6); /* 'ⓠ' -> 'Ⓠ' */
	set( 0x24e1, 0x24c7); /* 'ⓡ' -> 'Ⓡ' */
	set( 0x24e2, 0x24c8); /* 'ⓢ' -> 'Ⓢ' */
	set( 0x24e3, 0x24c9); /* 'ⓣ' -> 'Ⓣ' */
	set( 0x24e4, 0x24ca); /* 'ⓤ' -> 'Ⓤ' */
	set( 0x24e5, 0x24cb); /* 'ⓥ' -> 'Ⓥ' */
	set( 0x24e6, 0x24cc); /* 'ⓦ' -> 'Ⓦ' */
	set( 0x24e7, 0x24cd); /* 'ⓧ' -> 'Ⓧ' */
	set( 0x24e8, 0x24ce); /* 'ⓨ' -> 'Ⓨ' */
	set( 0x24e9, 0x24cf); /* 'ⓩ' -> 'Ⓩ' */
	set( 0x2c30, 0x2c00); /* 'ⰰ' -> 'Ⰰ' */
	set( 0x2c31, 0x2c01); /* 'ⰱ' -> 'Ⰱ' */
	set( 0x2c32, 0x2c02); /* 'ⰲ' -> 'Ⰲ' */
	set( 0x2c33, 0x2c03); /* 'ⰳ' -> 'Ⰳ' */
	set( 0x2c34, 0x2c04); /* 'ⰴ' -> 'Ⰴ' */
	set( 0x2c35, 0x2c05); /* 'ⰵ' -> 'Ⰵ' */
	set( 0x2c36, 0x2c06); /* 'ⰶ' -> 'Ⰶ' */
	set( 0x2c37, 0x2c07); /* 'ⰷ' -> 'Ⰷ' */
	set( 0x2c38, 0x2c08); /* 'ⰸ' -> 'Ⰸ' */
	set( 0x2c39, 0x2c09); /* 'ⰹ' -> 'Ⰹ' */
	set( 0x2c3a, 0x2c0a); /* 'ⰺ' -> 'Ⰺ' */
	set( 0x2c3b, 0x2c0b); /* 'ⰻ' -> 'Ⰻ' */
	set( 0x2c3c, 0x2c0c); /* 'ⰼ' -> 'Ⰼ' */
	set( 0x2c3d, 0x2c0d); /* 'ⰽ' -> 'Ⰽ' */
	set( 0x2c3e, 0x2c0e); /* 'ⰾ' -> 'Ⰾ' */
	set( 0x2c3f, 0x2c0f); /* 'ⰿ' -> 'Ⰿ' */
	set( 0x2c40, 0x2c10); /* 'ⱀ' -> 'Ⱀ' */
	set( 0x2c41, 0x2c11); /* 'ⱁ' -> 'Ⱁ' */
	set( 0x2c42, 0x2c12); /* 'ⱂ' -> 'Ⱂ' */
	set( 0x2c43, 0x2c13); /* 'ⱃ' -> 'Ⱃ' */
	set( 0x2c44, 0x2c14); /* 'ⱄ' -> 'Ⱄ' */
	set( 0x2c45, 0x2c15); /* 'ⱅ' -> 'Ⱅ' */
	set( 0x2c46, 0x2c16); /* 'ⱆ' -> 'Ⱆ' */
	set( 0x2c47, 0x2c17); /* 'ⱇ' -> 'Ⱇ' */
	set( 0x2c48, 0x2c18); /* 'ⱈ' -> 'Ⱈ' */
	set( 0x2c49, 0x2c19); /* 'ⱉ' -> 'Ⱉ' */
	set( 0x2c4a, 0x2c1a); /* 'ⱊ' -> 'Ⱊ' */
	set( 0x2c4b, 0x2c1b); /* 'ⱋ' -> 'Ⱋ' */
	set( 0x2c4c, 0x2c1c); /* 'ⱌ' -> 'Ⱌ' */
	set( 0x2c4d, 0x2c1d); /* 'ⱍ' -> 'Ⱍ' */
	set( 0x2c4e, 0x2c1e); /* 'ⱎ' -> 'Ⱎ' */
	set( 0x2c4f, 0x2c1f); /* 'ⱏ' -> 'Ⱏ' */
	set( 0x2c50, 0x2c20); /* 'ⱐ' -> 'Ⱐ' */
	set( 0x2c51, 0x2c21); /* 'ⱑ' -> 'Ⱑ' */
	set( 0x2c52, 0x2c22); /* 'ⱒ' -> 'Ⱒ' */
	set( 0x2c53, 0x2c23); /* 'ⱓ' -> 'Ⱓ' */
	set( 0x2c54, 0x2c24); /* 'ⱔ' -> 'Ⱔ' */
	set( 0x2c55, 0x2c25); /* 'ⱕ' -> 'Ⱕ' */
	set( 0x2c56, 0x2c26); /* 'ⱖ' -> 'Ⱖ' */
	set( 0x2c57, 0x2c27); /* 'ⱗ' -> 'Ⱗ' */
	set( 0x2c58, 0x2c28); /* 'ⱘ' -> 'Ⱘ' */
	set( 0x2c59, 0x2c29); /* 'ⱙ' -> 'Ⱙ' */
	set( 0x2c5a, 0x2c2a); /* 'ⱚ' -> 'Ⱚ' */
	set( 0x2c5b, 0x2c2b); /* 'ⱛ' -> 'Ⱛ' */
	set( 0x2c5c, 0x2c2c); /* 'ⱜ' -> 'Ⱜ' */
	set( 0x2c5d, 0x2c2d); /* 'ⱝ' -> 'Ⱝ' */
	set( 0x2c5e, 0x2c2e); /* 'ⱞ' -> 'Ⱞ' */
	set( 0x2c61, 0x2c60); /* 'ⱡ' -> 'Ⱡ' */
	set( 0x2c65, 0x23a); /* 'ⱥ' -> 'Ⱥ' */
	set( 0x2c66, 0x23e); /* 'ⱦ' -> 'Ⱦ' */
	set( 0x2c68, 0x2c67); /* 'ⱨ' -> 'Ⱨ' */
	set( 0x2c6a, 0x2c69); /* 'ⱪ' -> 'Ⱪ' */
	set( 0x2c6c, 0x2c6b); /* 'ⱬ' -> 'Ⱬ' */
	set( 0x2c73, 0x2c72); /* 'ⱳ' -> 'Ⱳ' */
	set( 0x2c76, 0x2c75); /* 'ⱶ' -> 'Ⱶ' */
	set( 0x2c81, 0x2c80); /* 'ⲁ' -> 'Ⲁ' */
	set( 0x2c83, 0x2c82); /* 'ⲃ' -> 'Ⲃ' */
	set( 0x2c85, 0x2c84); /* 'ⲅ' -> 'Ⲅ' */
	set( 0x2c87, 0x2c86); /* 'ⲇ' -> 'Ⲇ' */
	set( 0x2c89, 0x2c88); /* 'ⲉ' -> 'Ⲉ' */
	set( 0x2c8b, 0x2c8a); /* 'ⲋ' -> 'Ⲋ' */
	set( 0x2c8d, 0x2c8c); /* 'ⲍ' -> 'Ⲍ' */
	set( 0x2c8f, 0x2c8e); /* 'ⲏ' -> 'Ⲏ' */
	set( 0x2c91, 0x2c90); /* 'ⲑ' -> 'Ⲑ' */
	set( 0x2c93, 0x2c92); /* 'ⲓ' -> 'Ⲓ' */
	set( 0x2c95, 0x2c94); /* 'ⲕ' -> 'Ⲕ' */
	set( 0x2c97, 0x2c96); /* 'ⲗ' -> 'Ⲗ' */
	set( 0x2c99, 0x2c98); /* 'ⲙ' -> 'Ⲙ' */
	set( 0x2c9b, 0x2c9a); /* 'ⲛ' -> 'Ⲛ' */
	set( 0x2c9d, 0x2c9c); /* 'ⲝ' -> 'Ⲝ' */
	set( 0x2c9f, 0x2c9e); /* 'ⲟ' -> 'Ⲟ' */
	set( 0x2ca1, 0x2ca0); /* 'ⲡ' -> 'Ⲡ' */
	set( 0x2ca3, 0x2ca2); /* 'ⲣ' -> 'Ⲣ' */
	set( 0x2ca5, 0x2ca4); /* 'ⲥ' -> 'Ⲥ' */
	set( 0x2ca7, 0x2ca6); /* 'ⲧ' -> 'Ⲧ' */
	set( 0x2ca9, 0x2ca8); /* 'ⲩ' -> 'Ⲩ' */
	set( 0x2cab, 0x2caa); /* 'ⲫ' -> 'Ⲫ' */
	set( 0x2cad, 0x2cac); /* 'ⲭ' -> 'Ⲭ' */
	set( 0x2caf, 0x2cae); /* 'ⲯ' -> 'Ⲯ' */
	set( 0x2cb1, 0x2cb0); /* 'ⲱ' -> 'Ⲱ' */
	set( 0x2cb3, 0x2cb2); /* 'ⲳ' -> 'Ⲳ' */
	set( 0x2cb5, 0x2cb4); /* 'ⲵ' -> 'Ⲵ' */
	set( 0x2cb7, 0x2cb6); /* 'ⲷ' -> 'Ⲷ' */
	set( 0x2cb9, 0x2cb8); /* 'ⲹ' -> 'Ⲹ' */
	set( 0x2cbb, 0x2cba); /* 'ⲻ' -> 'Ⲻ' */
	set( 0x2cbd, 0x2cbc); /* 'ⲽ' -> 'Ⲽ' */
	set( 0x2cbf, 0x2cbe); /* 'ⲿ' -> 'Ⲿ' */
	set( 0x2cc1, 0x2cc0); /* 'ⳁ' -> 'Ⳁ' */
	set( 0x2cc3, 0x2cc2); /* 'ⳃ' -> 'Ⳃ' */
	set( 0x2cc5, 0x2cc4); /* 'ⳅ' -> 'Ⳅ' */
	set( 0x2cc7, 0x2cc6); /* 'ⳇ' -> 'Ⳇ' */
	set( 0x2cc9, 0x2cc8); /* 'ⳉ' -> 'Ⳉ' */
	set( 0x2ccb, 0x2cca); /* 'ⳋ' -> 'Ⳋ' */
	set( 0x2ccd, 0x2ccc); /* 'ⳍ' -> 'Ⳍ' */
	set( 0x2ccf, 0x2cce); /* 'ⳏ' -> 'Ⳏ' */
	set( 0x2cd1, 0x2cd0); /* 'ⳑ' -> 'Ⳑ' */
	set( 0x2cd3, 0x2cd2); /* 'ⳓ' -> 'Ⳓ' */
	set( 0x2cd5, 0x2cd4); /* 'ⳕ' -> 'Ⳕ' */
	set( 0x2cd7, 0x2cd6); /* 'ⳗ' -> 'Ⳗ' */
	set( 0x2cd9, 0x2cd8); /* 'ⳙ' -> 'Ⳙ' */
	set( 0x2cdb, 0x2cda); /* 'ⳛ' -> 'Ⳛ' */
	set( 0x2cdd, 0x2cdc); /* 'ⳝ' -> 'Ⳝ' */
	set( 0x2cdf, 0x2cde); /* 'ⳟ' -> 'Ⳟ' */
	set( 0x2ce1, 0x2ce0); /* 'ⳡ' -> 'Ⳡ' */
	set( 0x2ce3, 0x2ce2); /* 'ⳣ' -> 'Ⳣ' */
	set( 0x2cec, 0x2ceb); /* 'ⳬ' -> 'Ⳬ' */
	set( 0x2cee, 0x2ced); /* 'ⳮ' -> 'Ⳮ' */
	set( 0x2d00, 0x10a0); /* 'ⴀ' -> 'Ⴀ' */
	set( 0x2d01, 0x10a1); /* 'ⴁ' -> 'Ⴁ' */
	set( 0x2d02, 0x10a2); /* 'ⴂ' -> 'Ⴂ' */
	set( 0x2d03, 0x10a3); /* 'ⴃ' -> 'Ⴃ' */
	set( 0x2d04, 0x10a4); /* 'ⴄ' -> 'Ⴄ' */
	set( 0x2d05, 0x10a5); /* 'ⴅ' -> 'Ⴅ' */
	set( 0x2d06, 0x10a6); /* 'ⴆ' -> 'Ⴆ' */
	set( 0x2d07, 0x10a7); /* 'ⴇ' -> 'Ⴇ' */
	set( 0x2d08, 0x10a8); /* 'ⴈ' -> 'Ⴈ' */
	set( 0x2d09, 0x10a9); /* 'ⴉ' -> 'Ⴉ' */
	set( 0x2d0a, 0x10aa); /* 'ⴊ' -> 'Ⴊ' */
	set( 0x2d0b, 0x10ab); /* 'ⴋ' -> 'Ⴋ' */
	set( 0x2d0c, 0x10ac); /* 'ⴌ' -> 'Ⴌ' */
	set( 0x2d0d, 0x10ad); /* 'ⴍ' -> 'Ⴍ' */
	set( 0x2d0e, 0x10ae); /* 'ⴎ' -> 'Ⴎ' */
	set( 0x2d0f, 0x10af); /* 'ⴏ' -> 'Ⴏ' */
	set( 0x2d10, 0x10b0); /* 'ⴐ' -> 'Ⴐ' */
	set( 0x2d11, 0x10b1); /* 'ⴑ' -> 'Ⴑ' */
	set( 0x2d12, 0x10b2); /* 'ⴒ' -> 'Ⴒ' */
	set( 0x2d13, 0x10b3); /* 'ⴓ' -> 'Ⴓ' */
	set( 0x2d14, 0x10b4); /* 'ⴔ' -> 'Ⴔ' */
	set( 0x2d15, 0x10b5); /* 'ⴕ' -> 'Ⴕ' */
	set( 0x2d16, 0x10b6); /* 'ⴖ' -> 'Ⴖ' */
	set( 0x2d17, 0x10b7); /* 'ⴗ' -> 'Ⴗ' */
	set( 0x2d18, 0x10b8); /* 'ⴘ' -> 'Ⴘ' */
	set( 0x2d19, 0x10b9); /* 'ⴙ' -> 'Ⴙ' */
	set( 0x2d1a, 0x10ba); /* 'ⴚ' -> 'Ⴚ' */
	set( 0x2d1b, 0x10bb); /* 'ⴛ' -> 'Ⴛ' */
	set( 0x2d1c, 0x10bc); /* 'ⴜ' -> 'Ⴜ' */
	set( 0x2d1d, 0x10bd); /* 'ⴝ' -> 'Ⴝ' */
	set( 0x2d1e, 0x10be); /* 'ⴞ' -> 'Ⴞ' */
	set( 0x2d1f, 0x10bf); /* 'ⴟ' -> 'Ⴟ' */
	set( 0x2d20, 0x10c0); /* 'ⴠ' -> 'Ⴠ' */
	set( 0x2d21, 0x10c1); /* 'ⴡ' -> 'Ⴡ' */
	set( 0x2d22, 0x10c2); /* 'ⴢ' -> 'Ⴢ' */
	set( 0x2d23, 0x10c3); /* 'ⴣ' -> 'Ⴣ' */
	set( 0x2d24, 0x10c4); /* 'ⴤ' -> 'Ⴤ' */
	set( 0x2d25, 0x10c5); /* 'ⴥ' -> 'Ⴥ' */
	set( 0xa641, 0xa640); /* 'ꙁ' -> 'Ꙁ' */
	set( 0xa643, 0xa642); /* 'ꙃ' -> 'Ꙃ' */
	set( 0xa645, 0xa644); /* 'ꙅ' -> 'Ꙅ' */
	set( 0xa647, 0xa646); /* 'ꙇ' -> 'Ꙇ' */
	set( 0xa649, 0xa648); /* 'ꙉ' -> 'Ꙉ' */
	set( 0xa64b, 0xa64a); /* 'ꙋ' -> 'Ꙋ' */
	set( 0xa64d, 0xa64c); /* 'ꙍ' -> 'Ꙍ' */
	set( 0xa64f, 0xa64e); /* 'ꙏ' -> 'Ꙏ' */
	set( 0xa651, 0xa650); /* 'ꙑ' -> 'Ꙑ' */
	set( 0xa653, 0xa652); /* 'ꙓ' -> 'Ꙓ' */
	set( 0xa655, 0xa654); /* 'ꙕ' -> 'Ꙕ' */
	set( 0xa657, 0xa656); /* 'ꙗ' -> 'Ꙗ' */
	set( 0xa659, 0xa658); /* 'ꙙ' -> 'Ꙙ' */
	set( 0xa65b, 0xa65a); /* 'ꙛ' -> 'Ꙛ' */
	set( 0xa65d, 0xa65c); /* 'ꙝ' -> 'Ꙝ' */
	set( 0xa65f, 0xa65e); /* 'ꙟ' -> 'Ꙟ' */
	set( 0xa663, 0xa662); /* 'ꙣ' -> 'Ꙣ' */
	set( 0xa665, 0xa664); /* 'ꙥ' -> 'Ꙥ' */
	set( 0xa667, 0xa666); /* 'ꙧ' -> 'Ꙧ' */
	set( 0xa669, 0xa668); /* 'ꙩ' -> 'Ꙩ' */
	set( 0xa66b, 0xa66a); /* 'ꙫ' -> 'Ꙫ' */
	set( 0xa66d, 0xa66c); /* 'ꙭ' -> 'Ꙭ' */
	set( 0xa681, 0xa680); /* 'ꚁ' -> 'Ꚁ' */
	set( 0xa683, 0xa682); /* 'ꚃ' -> 'Ꚃ' */
	set( 0xa685, 0xa684); /* 'ꚅ' -> 'Ꚅ' */
	set( 0xa687, 0xa686); /* 'ꚇ' -> 'Ꚇ' */
	set( 0xa689, 0xa688); /* 'ꚉ' -> 'Ꚉ' */
	set( 0xa68b, 0xa68a); /* 'ꚋ' -> 'Ꚋ' */
	set( 0xa68d, 0xa68c); /* 'ꚍ' -> 'Ꚍ' */
	set( 0xa68f, 0xa68e); /* 'ꚏ' -> 'Ꚏ' */
	set( 0xa691, 0xa690); /* 'ꚑ' -> 'Ꚑ' */
	set( 0xa693, 0xa692); /* 'ꚓ' -> 'Ꚓ' */
	set( 0xa695, 0xa694); /* 'ꚕ' -> 'Ꚕ' */
	set( 0xa697, 0xa696); /* 'ꚗ' -> 'Ꚗ' */
	set( 0xa723, 0xa722); /* 'ꜣ' -> 'Ꜣ' */
	set( 0xa725, 0xa724); /* 'ꜥ' -> 'Ꜥ' */
	set( 0xa727, 0xa726); /* 'ꜧ' -> 'Ꜧ' */
	set( 0xa729, 0xa728); /* 'ꜩ' -> 'Ꜩ' */
	set( 0xa72b, 0xa72a); /* 'ꜫ' -> 'Ꜫ' */
	set( 0xa72d, 0xa72c); /* 'ꜭ' -> 'Ꜭ' */
	set( 0xa72f, 0xa72e); /* 'ꜯ' -> 'Ꜯ' */
	set( 0xa733, 0xa732); /* 'ꜳ' -> 'Ꜳ' */
	set( 0xa735, 0xa734); /* 'ꜵ' -> 'Ꜵ' */
	set( 0xa737, 0xa736); /* 'ꜷ' -> 'Ꜷ' */
	set( 0xa739, 0xa738); /* 'ꜹ' -> 'Ꜹ' */
	set( 0xa73b, 0xa73a); /* 'ꜻ' -> 'Ꜻ' */
	set( 0xa73d, 0xa73c); /* 'ꜽ' -> 'Ꜽ' */
	set( 0xa73f, 0xa73e); /* 'ꜿ' -> 'Ꜿ' */
	set( 0xa741, 0xa740); /* 'ꝁ' -> 'Ꝁ' */
	set( 0xa743, 0xa742); /* 'ꝃ' -> 'Ꝃ' */
	set( 0xa745, 0xa744); /* 'ꝅ' -> 'Ꝅ' */
	set( 0xa747, 0xa746); /* 'ꝇ' -> 'Ꝇ' */
	set( 0xa749, 0xa748); /* 'ꝉ' -> 'Ꝉ' */
	set( 0xa74b, 0xa74a); /* 'ꝋ' -> 'Ꝋ' */
	set( 0xa74d, 0xa74c); /* 'ꝍ' -> 'Ꝍ' */
	set( 0xa74f, 0xa74e); /* 'ꝏ' -> 'Ꝏ' */
	set( 0xa751, 0xa750); /* 'ꝑ' -> 'Ꝑ' */
	set( 0xa753, 0xa752); /* 'ꝓ' -> 'Ꝓ' */
	set( 0xa755, 0xa754); /* 'ꝕ' -> 'Ꝕ' */
	set( 0xa757, 0xa756); /* 'ꝗ' -> 'Ꝗ' */
	set( 0xa759, 0xa758); /* 'ꝙ' -> 'Ꝙ' */
	set( 0xa75b, 0xa75a); /* 'ꝛ' -> 'Ꝛ' */
	set( 0xa75d, 0xa75c); /* 'ꝝ' -> 'Ꝝ' */
	set( 0xa75f, 0xa75e); /* 'ꝟ' -> 'Ꝟ' */
	set( 0xa761, 0xa760); /* 'ꝡ' -> 'Ꝡ' */
	set( 0xa763, 0xa762); /* 'ꝣ' -> 'Ꝣ' */
	set( 0xa765, 0xa764); /* 'ꝥ' -> 'Ꝥ' */
	set( 0xa767, 0xa766); /* 'ꝧ' -> 'Ꝧ' */
	set( 0xa769, 0xa768); /* 'ꝩ' -> 'Ꝩ' */
	set( 0xa76b, 0xa76a); /* 'ꝫ' -> 'Ꝫ' */
	set( 0xa76d, 0xa76c); /* 'ꝭ' -> 'Ꝭ' */
	set( 0xa76f, 0xa76e); /* 'ꝯ' -> 'Ꝯ' */
	set( 0xa77a, 0xa779); /* 'ꝺ' -> 'Ꝺ' */
	set( 0xa77c, 0xa77b); /* 'ꝼ' -> 'Ꝼ' */
	set( 0xa77f, 0xa77e); /* 'ꝿ' -> 'Ꝿ' */
	set( 0xa781, 0xa780); /* 'ꞁ' -> 'Ꞁ' */
	set( 0xa783, 0xa782); /* 'ꞃ' -> 'Ꞃ' */
	set( 0xa785, 0xa784); /* 'ꞅ' -> 'Ꞅ' */
	set( 0xa787, 0xa786); /* 'ꞇ' -> 'Ꞇ' */
	set( 0xa78c, 0xa78b); /* 'ꞌ' -> 'Ꞌ' */
	set( 0xff41, 0xff21); /* 'ａ' -> 'Ａ' */
	set( 0xff42, 0xff22); /* 'ｂ' -> 'Ｂ' */
	set( 0xff43, 0xff23); /* 'ｃ' -> 'Ｃ' */
	set( 0xff44, 0xff24); /* 'ｄ' -> 'Ｄ' */
	set( 0xff45, 0xff25); /* 'ｅ' -> 'Ｅ' */
	set( 0xff46, 0xff26); /* 'ｆ' -> 'Ｆ' */
	set( 0xff47, 0xff27); /* 'ｇ' -> 'Ｇ' */
	set( 0xff48, 0xff28); /* 'ｈ' -> 'Ｈ' */
	set( 0xff49, 0xff29); /* 'ｉ' -> 'Ｉ' */
	set( 0xff4a, 0xff2a); /* 'ｊ' -> 'Ｊ' */
	set( 0xff4b, 0xff2b); /* 'ｋ' -> 'Ｋ' */
	set( 0xff4c, 0xff2c); /* 'ｌ' -> 'Ｌ' */
	set( 0xff4d, 0xff2d); /* 'ｍ' -> 'Ｍ' */
	set( 0xff4e, 0xff2e); /* 'ｎ' -> 'Ｎ' */
	set( 0xff4f, 0xff2f); /* 'ｏ' -> 'Ｏ' */
	set( 0xff50, 0xff30); /* 'ｐ' -> 'Ｐ' */
	set( 0xff51, 0xff31); /* 'ｑ' -> 'Ｑ' */
	set( 0xff52, 0xff32); /* 'ｒ' -> 'Ｒ' */
	set( 0xff53, 0xff33); /* 'ｓ' -> 'Ｓ' */
	set( 0xff54, 0xff34); /* 'ｔ' -> 'Ｔ' */
	set( 0xff55, 0xff35); /* 'ｕ' -> 'Ｕ' */
	set( 0xff56, 0xff36); /* 'ｖ' -> 'Ｖ' */
	set( 0xff57, 0xff37); /* 'ｗ' -> 'Ｗ' */
	set( 0xff58, 0xff38); /* 'ｘ' -> 'Ｘ' */
	set( 0xff59, 0xff39); /* 'ｙ' -> 'Ｙ' */
	set( 0xff5a, 0xff3a); /* 'ｚ' -> 'Ｚ' */
	set( 0x10428, 0x10400); /* '𐐨' -> '𐐀' */
	set( 0x10429, 0x10401); /* '𐐩' -> '𐐁' */
	set( 0x1042a, 0x10402); /* '𐐪' -> '𐐂' */
	set( 0x1042b, 0x10403); /* '𐐫' -> '𐐃' */
	set( 0x1042c, 0x10404); /* '𐐬' -> '𐐄' */
	set( 0x1042d, 0x10405); /* '𐐭' -> '𐐅' */
	set( 0x1042e, 0x10406); /* '𐐮' -> '𐐆' */
	set( 0x1042f, 0x10407); /* '𐐯' -> '𐐇' */
	set( 0x10430, 0x10408); /* '𐐰' -> '𐐈' */
	set( 0x10431, 0x10409); /* '𐐱' -> '𐐉' */
	set( 0x10432, 0x1040a); /* '𐐲' -> '𐐊' */
	set( 0x10433, 0x1040b); /* '𐐳' -> '𐐋' */
	set( 0x10434, 0x1040c); /* '𐐴' -> '𐐌' */
	set( 0x10435, 0x1040d); /* '𐐵' -> '𐐍' */
	set( 0x10436, 0x1040e); /* '𐐶' -> '𐐎' */
	set( 0x10437, 0x1040f); /* '𐐷' -> '𐐏' */
	set( 0x10438, 0x10410); /* '𐐸' -> '𐐐' */
	set( 0x10439, 0x10411); /* '𐐹' -> '𐐑' */
	set( 0x1043a, 0x10412); /* '𐐺' -> '𐐒' */
	set( 0x1043b, 0x10413); /* '𐐻' -> '𐐓' */
	set( 0x1043c, 0x10414); /* '𐐼' -> '𐐔' */
	set( 0x1043d, 0x10415); /* '𐐽' -> '𐐕' */
	set( 0x1043e, 0x10416); /* '𐐾' -> '𐐖' */
	set( 0x1043f, 0x10417); /* '𐐿' -> '𐐗' */
	set( 0x10440, 0x10418); /* '𐑀' -> '𐐘' */
	set( 0x10441, 0x10419); /* '𐑁' -> '𐐙' */
	set( 0x10442, 0x1041a); /* '𐑂' -> '𐐚' */
	set( 0x10443, 0x1041b); /* '𐑃' -> '𐐛' */
	set( 0x10444, 0x1041c); /* '𐑄' -> '𐐜' */
	set( 0x10445, 0x1041d); /* '𐑅' -> '𐐝' */
	set( 0x10446, 0x1041e); /* '𐑆' -> '𐐞' */
	set( 0x10447, 0x1041f); /* '𐑇' -> '𐐟' */
	set( 0x10448, 0x10420); /* '𐑈' -> '𐐠' */
	set( 0x10449, 0x10421); /* '𐑉' -> '𐐡' */
	set( 0x1044a, 0x10422); /* '𐑊' -> '𐐢' */
	set( 0x1044b, 0x10423); /* '𐑋' -> '𐐣' */
	set( 0x1044c, 0x10424); /* '𐑌' -> '𐐤' */
	set( 0x1044d, 0x10425); /* '𐑍' -> '𐐥' */
	set( 0x1044e, 0x10426); /* '𐑎' -> '𐐦' */
	set( 0x1044f, 0x10427); /* '𐑏' -> '𐐧' */
}

void CharMap::load( ConvType type)
{
	switch (type)
	{
		case Diacritical:
			buildMapDiacritical();
			break;
		case Lowercase:
			buildMapTolower();
			break;
		case Uppercase:
			buildMapToupper();
			break;
	}
}

std::string CharMap::rewrite( const char* src, std::size_t srcsize, CharMap::ExceptionsF exceptions, ErrorBufferInterface* errorhnd) const
{
	try
	{
		std::string rt;
		textwolf::charset::UTF8 utf8;
		char buf[16];
		unsigned int bufpos;
		textwolf::CStringIterator itr( src, srcsize);

		while (*itr)
		{
			bufpos = 0;
			textwolf::UChar value = utf8.value( buf, bufpos, itr);
			if (exceptions)
			{
				const char* res = (*exceptions)(value);
				if (res)
				{
					rt.append(res);
					continue;
				}
			}
			std::map<unsigned int,std::size_t>::const_iterator mi = m_map.find( value);
			if (mi == m_map.end())
			{
				if (value == textwolf::charset::UTF8::MaxChar)
				{
					throw strus::runtime_error( "%s", _TXT( "illegal UTF-8 character in input"));
				}
				rt.append( buf, bufpos);
			}
			else
			{
				rt.append( m_strings.c_str() + mi->second);
			}
		}
		return rt;
	}
	CATCH_ERROR_MAP_RETURN( _TXT("error in normalizer: %s"), *errorhnd, std::string());
}

NormalizerFunctionInstanceInterface* LowercaseNormalizerFunction::createInstance( const std::vector<std::string>& args, const TextProcessorInterface*) const
{
	if (args.size())
	{
		m_errorhnd->report( _TXT("unexpected arguments passed to normalizer '%s'"), "lc");
		return 0;
	}
	try
	{
		return new CharMapNormalizerInstance( CharMap::Lowercase, 0, m_errorhnd);
	}
	catch (const std::bad_alloc&)
	{
		m_errorhnd->report( _TXT("out of memory in normalizer '%s'"), "lc");
		return 0;
	}
}

NormalizerFunctionInstanceInterface* UppercaseNormalizerFunction::createInstance( const std::vector<std::string>& args, const TextProcessorInterface*) const
{
	if (args.size())
	{
		m_errorhnd->report( _TXT("unexpected arguments passed to normalizer '%s'"), "uc");
		return 0;
	}
	try
	{
		return new CharMapNormalizerInstance( CharMap::Uppercase, 0, m_errorhnd);
	}
	catch (const std::bad_alloc&)
	{
		m_errorhnd->report( _TXT("out of memory in normalizer '%s'"), "uc");
		return 0;
	}
}

static const char* germanUmlautExceptions( unsigned int chr)
{
	if (chr >= 0x7F && chr <= 0xFF)
	{
		if (chr == 0xe4) return "ae";
		if (chr == 0xc4) return "Ae";
		if (chr == 0xdc) return "Ue";
		if (chr == 0xfc) return "ue";
		if (chr == 0xd6) return "Oe";
		if (chr == 0xf6) return "oe";
	}
	return 0;
}

NormalizerFunctionInstanceInterface* DiacriticalNormalizerFunction::createInstance( const std::vector<std::string>& args, const TextProcessorInterface*) const
{
	try
	{
		if (args.size() > 1)
		{
			m_errorhnd->report( _TXT("too many arguments passed to normalizer '%s'"), "convdia");
			return 0;
		}
		if (args.size() == 0)
		{
			return new CharMapNormalizerInstance( CharMap::Diacritical, 0, m_errorhnd);
		}
		else
		{
			std::string language_lo = string_conv::tolower( args[0]);
			if (language_lo == "de")
			{
				return new CharMapNormalizerInstance( CharMap::Diacritical, germanUmlautExceptions, m_errorhnd);
			}
			else
			{
				return new CharMapNormalizerInstance( CharMap::Diacritical, 0, m_errorhnd);
			}
		}
	}
	catch (const std::bad_alloc&)
	{
		m_errorhnd->report( "out of memory");
		return 0;
	}
}


class CharSet
{
public:
	CharSet( const std::vector<std::string>& setnames)
	{
		std::vector<std::string>::const_iterator si = setnames.begin(), se = setnames.end();
		for (; si != se; ++si)
		{
			std::string name = string_conv::tolower( *si);
			if (name == "punct_en" || name == "punct_de")
			{
				addRange( 44, 47);
				addRange( 58, 59);
				addRange( 33, 34);
				addRange( 39, 41);
				addRange( 63, 63);
			}
			else if (name == "digits")
			{
				addRange( '0', '9');
			}
			else if (name == "alpha_latin")
			{
				addRange( 'a', 'z');
				addRange( 'A', 'Z');
				addRange( 0xC0, 0x24F);
			}
			else if (name == "alpha_cyrillic")
			{
				addRange( 0x400, 0x4FF);
			}
			else if (name == "alpha_greek")
			{
				addRange( 0x370, 0x3FF);
				addRange( 0x500, 0x52F);
			}
			else if (name == "alpha_armenian")
			{
				addRange( 0x530, 0x58F);
			}
			else if (name == "alpha_hebrew")
			{
				addRange( 0x590, 0x5FF);
			}
			else if (name == "alpha_arabic")
			{
				addRange( 0x600, 0x6FF);
			}
			else if (name == "alpha_syriac")
			{
				addRange( 0x700, 0x74F);
			}
			else if (name == "alpha_ascii")
			{
				addRange( 'a', 'z');
				addRange( 'A', 'Z');
			}
			else
			{
				throw strus::runtime_error(_TXT("unknown character set name '%s' specified"), name.c_str());
			}
		}
	}

	CharSet(){}
	~CharSet(){}
	bool isMember( uint32_t chr) const
	{
		std::vector<Range>::const_iterator ri = m_ranges.begin(), re = m_ranges.end();
		for (; ri != re; ++ri)
		{
			if (chr >= ri->first && chr <= ri->second) return true;
		}
		return false;
	}

	void addRange( uint32_t first, uint32_t second)
	{
		if (first > second) throw strus::runtime_error(_TXT("illegal character range defined: %u..%u"), first, second);
		std::vector<Range>::const_iterator ri = m_ranges.begin(), re = m_ranges.end();
		for (; ri != re; ++ri)
		{
			if (first >= ri->first && second <= ri->second) return;
		}
		m_ranges.push_back( Range( first,second));
	}

private:
	typedef std::pair<uint32_t,uint32_t> Range;
	std::vector<Range> m_ranges;
};

class CharSelectNormalizerInstance
	:public NormalizerFunctionInstanceInterface
{
public:
	CharSelectNormalizerInstance( const std::vector<std::string>& setnames, ErrorBufferInterface* errorhnd)
		:m_set(setnames),m_errorhnd(errorhnd){}

	virtual std::string normalize(
			const char* src,
			std::size_t srcsize) const
	{
		try
		{
			std::string rt;
			textwolf::charset::UTF8 utf8;
			char buf[16];
			unsigned int bufpos;
			textwolf::CStringIterator itr( src, srcsize);
			bool selected = false;
	
			while (*itr)
			{
				bufpos = 0;
				textwolf::UChar value = utf8.value( buf, bufpos, itr);
				if (m_set.isMember( value))
				{
					rt.append( buf, bufpos);
					selected = true;
				}
				else
				{
					if (selected)
					{
						rt.push_back( ' ');
					}
					selected = false;
				}
			}
			return rt;
		}
		CATCH_ERROR_MAP_ARG1_RETURN( _TXT("error in '%s' normalizer: %s"), "charselect", *m_errorhnd, 0);
	}

private:
	CharSet m_set;
	ErrorBufferInterface* m_errorhnd;
};

NormalizerFunctionInstanceInterface* CharSelectNormalizerFunction::createInstance( const std::vector<std::string>& args, const TextProcessorInterface*) const
{
	try
	{
		return new CharSelectNormalizerInstance( args, m_errorhnd);
	}
	CATCH_ERROR_MAP_ARG1_RETURN( _TXT("error creating '%s' normalizer instance: %s"), "charselect", *m_errorhnd, 0);
}




